/* Copyright (c) 2004 ActiveState Corp, a division of Sophos Plc.
   See the file LICENSE.txt for licensing information. */

/* JavaScript-side control of the Code Intelligence system in Komodo
 * (code browsing, autocomplete and calltip triggering).
 *
 * Usage:
 * - Komodo's startup should call CodeIntel_Initialize() and shutdown should
 *   call CodeIntel_Finalize().
 * - All code calling the koICodeIntelService (defined on gCodeIntelSvc)
 *   or the Code Browser manager (defined on gCodeBrowserMgr) should first
 *   check that the system is active by checking the boolean
 *   gCodeIntelActive.
 * - CodeIntel_UpdateCodeBrowserVisibility() should be called whenever the
 *   Code Browser's showing/hidden state may have changed.
 */
// TODO:
//
// Feature Requests:
// - Currently <Enter> jumps to the current symbol in the Code Browser.
//   How about having <Shift+Enter> select the symbol... or just the line
//   (which is easier)? Could wait until/if the "line" attribute is
//   replaced with a more powerful "position" or similar. What about
//   <Ctrl+Enter>?


//---- globals

// gCodeIntelActive is true iff the Code Intel system is enabled,
// initialized, and active. This means it is safe to use gCodeIntelSvc and
// gCodeBrowserMgr.
var gCodeIntelActive = false;
var gCodeIntelSvc = null;
var gCodeBrowserMgr = null;
var gCodeBrowserCtrl = null; // commands available when cb focused
var gCodeIntelCtrl = null; // commands available always

var _gCodeIntel_log = getLoggingMgr().getLogger("codeintel_js");
//_gCodeIntel_log.setLevel(LOG_DEBUG);

var _gCodeIntel_observer = null;
var _gCodeIntel_prefObserver = null;


//---- internal support routines

// Return true iff the two file paths are equal.
function _fequal(a, b) {



    return a == b;

}


//---- internal observers

function _CodeIntelPrefObserver()
{
    try {
        this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                       getService(Components.interfaces.koIPrefService);
        this.prefSvc.prefs.prefObserverService.addObserver(this, "codeintel_enabled", 0);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
};
_CodeIntelPrefObserver.prototype.constructor = _CodeIntelPrefObserver;

_CodeIntelPrefObserver.prototype.finalize = function()
{
    try {
        this.prefSvc.prefs.prefObserverService.removeObserver(this, "codeintel_enabled");
        this.prefSvc = null;
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

_CodeIntelPrefObserver.prototype.observe = function(prefSet, prefName, prefSetID)
{
    _gCodeIntel_log.debug("observe pref '"+prefName+"': prefSet='"+prefSet+
                          "', prefSetID='"+prefSetID);
    try {
        switch (prefName) {
        case "codeintel_enabled":
            var enabled = prefSet.getBooleanPref("codeintel_enabled");
            if (enabled) {
                _CodeIntel_Activate();
            } else {
                _CodeIntel_Deactivate();
            }
            break;
        default:
            _gCodeIntel_log.error("unexpected pref name is "+
                                  "_CodeIntelPrefObserver: '"+prefName+"'\n");
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
};


function _CodeIntelObserver()
{
    try {
        this.obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                      getService(Components.interfaces.nsIObserverService);
        this.obsSvc.addObserver(this, "current_view_changed", false);
        this.obsSvc.addObserver(this, "current_view_language_changed", false);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}
_CodeIntelObserver.prototype.constructor = _CodeIntelObserver;

_CodeIntelObserver.prototype.finalize = function()
{
    try {
        this.obsSvc.removeObserver(this, "current_view_changed");
        this.obsSvc.removeObserver(this, "current_view_language_changed");
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

_CodeIntelObserver.prototype.observe = function(subject, topic, data)
{
    //_gCodeIntel_log.debug("observe(subject='"+subject+"', topic='"+topic+"', data='"+data+"')");
    try {
        var view;
        switch (topic) {
        // 'current_view_changed' is sent when a new tab in the editor pane
        // is made current. Annoyances:
        // - This is sent for _every_ file restored on startup, rather than
        //   (ideally) just the last one, i.e. the one that becomes current.
        //   Solution: The Code Intel system is only initialized after files
        //   have been opened on startup. It should gather the starter list of
        //   files and projects on its own.
        // - When the user has two "Tab Groups" and switches from a tab in
        //   one group to the other, _two_ "current_view_changed"
        //   notifications are sent for the newly selected tab. Solution:
        //   XXX Not sure what, if anything, need be done about this.
        case "current_view_changed":
            if (!gViewMgr.batchMode) { // batch edit or Ctrl+Tab'ing
                view = subject;
                //XXX Check 'view' because sometimes "current_view_changed"
                //    is sent with the subject set to NULL. To repro:
                //    - open two files in Komodo, one in each "Tab Group"
                //    - select the one in the second tab group and then
                //      close it with the "X" icon
                if (view && view.getAttribute("type") == "editor" &&
                    view.document) {
                    gCodeIntelSvc.ideEvent("switched_current_document", null,
                                           view.document);
                }
                window.setTimeout("window.updateCommands('codebrowser');", 0);
            }
            break;
        case "current_view_language_changed":
            view = subject;
            if (view.document) {
                gCodeIntelSvc.ideEvent("switched_current_language", null,
                                       view.document);
            }
            window.setTimeout("window.updateCommands('codebrowser');", 0);
            break;
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


//---- the controller for the code browser commands

function CodeBrowserController() {
}
// The following two lines ensure proper inheritance (see Flanagan, p. 144).
CodeBrowserController.prototype = new Controller();
CodeBrowserController.prototype.constructor = CodeBrowserController;

CodeBrowserController.prototype.destructor = function() {
}

CodeBrowserController.prototype._isRealNodeSelected = function() {
    if (gCodeBrowserMgr && gCodeBrowserMgr.view.selection.count != 0) {
        var type = gCodeBrowserMgr.view.currentNodeType;
        if (type && type != "pseudo-module" && type != "container") {
            return true;
        }
    }
    return false;
}

CodeBrowserController.prototype.is_cmd_copy_enabled = CodeBrowserController.prototype._isRealNodeSelected;
CodeBrowserController.prototype.do_cmd_copy = function() {
    gCodeBrowserMgr.copySelection();
}


function CodeIntelController() {
}

// The following two lines ensure proper inheritance (see Flanagan, p. 144).
CodeIntelController.prototype = new Controller();
CodeIntelController.prototype.constructor = CodeIntelController;

CodeIntelController.prototype.destructor = function() {
}

CodeIntelController.prototype._isRealNodeSelected = function() {
    if (gCodeBrowserMgr && gCodeBrowserMgr.view.selection.count != 0) {
        var type = gCodeBrowserMgr.view.currentNodeType;
        if (type && type != "pseudo-module" && type != "container") {
            return true;
        }
    }
    return false;
}

CodeIntelController.prototype.is_cmd_codeBrowserGoToDefinition_enabled = CodeIntelController.prototype._isRealNodeSelected;
CodeIntelController.prototype.do_cmd_codeBrowserGoToDefinition = function() {
    gCodeBrowserMgr._jumpToSymbol();
}


CodeIntelController.prototype.is_cmd_codeBrowserFindSymbol_enabled = CodeIntelController.prototype._isRealNodeSelected;
CodeIntelController.prototype.do_cmd_codeBrowserFindSymbol = function() {
    try {
        var v = gCodeBrowserMgr.view;
        var currentIndex = v.selection.currentIndex;
        if (currentIndex == -1) {
            return;
        }
        CodeIntel_FindSymbol(v.currentNodeSymbolPattern,
                             v.currentNodeModulePattern,
                             false,
                             v.currentNodeLanguage);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}



CodeIntelController.prototype.is_cmd_codeBrowserLocate_enabled = function() {
    if (!gCodeIntelActive) return false;
    var view = gViewMgr.currentView;
    return (view && view.scimoz && view.document &&
            view.codeIntelEnabled == true);
}

CodeIntelController.prototype.do_cmd_codeBrowserLocate = function() {
    CodeIntel_LocateCurrentScope();
}

//---- the UI Manager for the Code Browser (koICodeBrowserManager)

function CodeBrowserManager(treeId)
{
    _gCodeIntel_log.debug("CodeBrowserManager(treeId='"+treeId+"')");
    try {
        this.widgets = new Object();
        this.widgets.sortByButton = document.getElementById("codebrowser-sort-by-button");
        this.widgets.filterIcon = document.getElementById("codebrowser-filter-icon");
        this.widgets.filterTextbox = document.getElementById("codebrowser-filter-textbox");
        this.widgets.tree = document.getElementById("codebrowser-tree");
        this.widgets.description = document.getElementById("codebrowser-description");


        // Attach the nsITreeView implementation to the <tree/>.
        //XXX Do we have to delay this until we are visible for the first
        //    time? If so then make the .visible setter a method and change
        //    the call in CodeIntel_UpdateCodeBrowserVisibility().
        this.view = Components.classes["@activestate.com/koCodeBrowserTreeView;1"]
                              .createInstance();
        this.boxObject = this.widgets.tree.treeBoxObject;
        this.boxObject.view = this.view;

        // Restore preferences.
        this.view.restorePrefs();

        // Restore sort order.
        var key = null;
        if (gPrefs.hasStringPref("codeintel_sort_codebrowser_by")) {
            key = gPrefs.getStringPref("codeintel_sort_codebrowser_by");
        }
        if (! this.sortBy(key)) {
            var fallback = "file-order";
            _gCodeIntel_log.warn("invalid Code Browser sort order key from "+
                                 "prefs ('"+key+"'): falling back to '"+
                                 fallback+"'");
            this.sortBy(fallback);
        }

        this.visible = null; // is the "Code" tab visible?

        this._mouseMoveTimer = null;
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}
CodeBrowserManager.prototype.constructor = CodeBrowserManager;

CodeBrowserManager.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.koICodeBrowserManager) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}

CodeBrowserManager.prototype.finalize = function() {
    _gCodeIntel_log.debug("CodeBrowserManager.finalize()");
    try {
        this.view.savePrefs();
        gPrefs.setStringPref("codeintel_sort_codebrowser_by",
                             this.widgets.sortByButton.getAttribute("sortBy"));
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

CodeBrowserManager.prototype.onSelect = function(event) {
    try {
        gCodeBrowserMgr.updateCommands();
        gCodeBrowserMgr.updateDescription();
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

CodeBrowserManager.prototype.onFilterKeypress = function(event) {
    try {
        if (event.keyCode == event.DOM_VK_TAB && !event.ctrlKey) {
            event.cancelBubble = true;
            event.stopPropagation();
            event.preventDefault();
            gCodeBrowserMgr.widgets.tree.focus();
            return;
        } else if (event.keyCode == event.DOM_VK_ESCAPE) {
            if (this.widgets.filterTextbox.value != '') {
                this.widgets.filterTextbox.value = '';
                gCodeBrowserMgr.updateFilter();
                event.cancelBubble = true;
                event.stopPropagation();
                event.preventDefault();
            }
            return;
        } else if (event.ctrlKey && event.charCode == 'z'.charCodeAt(0)) {
            // ctrl-z doesn't do the right thing w.r.t. calling 'oninput',
            // so we do it ourselves.
            window.setTimeout("gCodeBrowserMgr.updateFilter();", 0)
        }
    } catch (e) {
        _gCodeIntel_log.exception(e);
    }
}

CodeBrowserManager.prototype.updateDescription = function() {
    try {
        var description = this.widgets.description;
        while (description.firstChild) {
            description.removeChild(description.firstChild);
        }
        if (this.view.selection.currentIndex < this.view.rowCount) {
            this.view.fillDescription(this.view.selection.currentIndex, document, description);
        }
    } catch (e) {
        _gCodeIntel_log.exception(e);
    }
}

CodeBrowserManager.prototype.updateCommands = function() {
    // update commands when the selection changes.
    _command_updateCommand('cmd_copy', gCodeBrowserCtrl);
    _command_updateCommand('cmd_codeBrowserGoToDefinition', gCodeIntelCtrl);
    _command_updateCommand('cmd_codeBrowserFindSymbol', gCodeIntelCtrl);
}

CodeBrowserManager.prototype.focus = function() {
    _gCodeIntel_log.debug("CodeBrowserManager.focus()");
    try {
        // Pass the focus onto the <tree/>.
        // It is important to do this because this focus handling is used to
        // get uilayout_toggleTab() to work properly.
        this.widgets.tree.focus();
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

CodeBrowserManager.prototype._jumpToSymbol = function() {
    _gCodeIntel_log.debug("CodeBrowserManager._jumpToSymbol()");
    try {
        var type = this.view.currentNodeType;
        if (! type) {
            return;
        }
        var path = this.view.currentNodeFilePath;
        var line = null;
        if (type != "pseudo-module" && type != "container") {
            line = this.view.currentNodeLine;
            if (type == "module" && line == 1) {
                // Special case: if this is a module node and the line is 1,
                // then don't specify to jump to line 1, this might be
                // changing the current line if the module is already open. A
                // module line of 1 usually indicates that this is a language
                // where file===module, i.e. the line number is essentially
                // meaningless.
                line = null;
            }
        }

        if (path.slice(0, 9) == "<Unsaved>") {
            //XXX Komodo's view system does not currently provide a simple
            //    method to open or get a handle on the view for a given
            //    untitled document. We therefore have to hack it ourselves.
            path = path.slice(10);
            var views = gViewMgr.topView.getViews(true);
            var v;
            for (var i = 0; i < views.length; i++) {
                v = views[i];
                if (v.getAttribute("type") == "editor"
                    && v.document.isUntitled
                    && _fequal(v.document.baseName, path))
                {
                    v.makeCurrent();
                    if (line) v.currentLine = line;
                    break;
                }
            }
        } else {
            var uri = uriparse_pathToURI(path);
            if (line) {
                gViewMgr.doFileOpenAtLine(uri, line);
            } else {
                gViewMgr.doFileOpen(uri);
            }
        }

        // Bug 32471: scroll line to a consistent position: 30% down from top.
        // (This is a SciMoz recipe to scroll a given line to 30% down from
        // the top of the viewport, properly accounting for folding and line
        // wrapping.)
        if (line) {
            var scimoz = gViewMgr.currentView.scimoz;
            scimoz.ensureVisible(line);
            var want2BNLinesFromTop = Math.round(scimoz.linesOnScreen*0.30);
            var wantTopLine2B = Math.max(0,
                scimoz.visibleFromDocLine(line) - want2BNLinesFromTop);
            var scrollBy = wantTopLine2B - scimoz.firstVisibleLine;
            //dump("- Want line "+line+" (considering folding: "
            //     +scimoz.visibleFromDocLine(line)+") to be "
            //     +want2BNLinesFromTop+" lines (30%) from top.\n"
            //     +"  I.e. want top line to be "+
            //     scimoz.docLineFromVisible(wantTopLine2B)
            //     +" (considering folding: "+wantTopLine2B+"),\n"
            //     +"  therefore need to scroll "+scrollBy+" lines.\n");
            scimoz.lineScroll(0, scrollBy);
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


CodeBrowserManager.prototype.onTreeKeyPress = function(event) {
    _gCodeIntel_log.debug("CodeBrowserManager.onTreeKeyPress(event)");
    try {
        if (event.keyCode == event.DOM_VK_TAB && !event.ctrlKey)
        {
            event.cancelBubble = true;
            event.stopPropagation();
            event.preventDefault();
            window.setTimeout("gCodeBrowserMgr.focusFilterBox();", 0);
            return;
        }
        if (event.keyCode == event.DOM_VK_ENTER
            || event.keyCode == event.DOM_VK_RETURN)
        {
            event.cancelBubble = true;
            event.stopPropagation();
            event.preventDefault();
            gCodeBrowserMgr._jumpToSymbol();
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

CodeBrowserManager.prototype.focusFilterBox = function() {
    this.widgets.filterTextbox.focus();
}


// Create instance of component given contractId and iid (as string).
CodeBrowserManager.prototype._createInstance = function(contractId, iidName) {
        var iid = Components.interfaces[iidName];
        return Components.classes[contractId].createInstance(iid);
}
// Get service given contractId and iid (as string).
CodeBrowserManager.prototype._getService = function(contractId, iidName) {
        var iid = Components.interfaces[iidName];
        return Components.classes[contractId].getService(iid);
}

CodeBrowserManager.prototype._copyToClipboard = function(text) {
    // Copy link/image url to clipboard.
    // Get clipboard.
    var clipboard = this._getService("@mozilla.org/widget/clipboard;1",
                                     "nsIClipboard");

    // Create tranferable that will transfer the text.
    var transferable = this._createInstance("@mozilla.org/widget/transferable;1",
                                            "nsITransferable");

    if (clipboard && transferable) {
        transferable.addDataFlavor("text/unicode");
        // Create wrapper for text.
        var data = this._createInstance("@mozilla.org/supports-string;1",
                                        "nsISupportsString");
        if (data) {
            data.data = text;
            transferable.setTransferData("text/unicode", data, text.length * 2);
            // Put on clipboard.
            clipboard.setData(transferable, null,
                              Components.interfaces.nsIClipboard.kGlobalClipboard);
        }
    }
}

CodeBrowserManager.prototype.copySelection = function() {
    try {
        var currentIndex = this.view.selection.currentIndex;
        if (currentIndex == -1) {
            return;
        }
        var column;

        column = this.boxObject.columns.getNamedColumn("codebrowser-tree-node");



        var text = this.view.getCellText(currentIndex, column);
        this._copyToClipboard(text);
    } catch (e) {
        log.exception(e);
    }
}


CodeBrowserManager.prototype.onDblClick = function(event) {
    _gCodeIntel_log.debug("CodeBrowserManager.onDblClick(event)");
    try {
        this._jumpToSymbol();
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

// Let the Code Browser tree view know if the mouse is over a specific
// node.  The view can use this info to show more information
// for that node.
CodeBrowserManager.prototype._handleMouseDwell = function(x, y) {
    try {
        var rowObj = new Object();
        var colObj = new Object();
        var childElementObj = new Object();
        this.boxObject.getCellAt(x, y, rowObj, colObj, childElementObj);
        var row = rowObj.value;
        var col = colObj.value;
        var childElement = childElementObj.value;

        // assert col == "codebrowser-tree-node"
        gCodeBrowserMgr.view.mouseOverNode(row, childElement);

        this._mouseMoveTimer.free();
        this._mouseMoveTimer = null;
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}
CodeBrowserManager.prototype.onMouseMove = function(event) {
    _gCodeIntel_log.debug("CodeBrowserManager.onMouseMove(event)");
    try {
        if (event.originalTarget.localName == "treechildren") {
            if (this._mouseMoveTimer) {
                this._mouseMoveTimer.stopTimeout();
                this._mouseMoveTimer.free();
                this._mouseMoveTimer = null;
            }
            this._mouseMoveTimer = new objectTimer(this, this._handleMouseDwell,
                                                   [event.clientX, event.clientY]);
            this._mouseMoveTimer.startTimeout(100);
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}
CodeBrowserManager.prototype.onMouseOut = function(event) {
    _gCodeIntel_log.debug("CodeBrowserManager.onMouseOut(event)");
    try {
        window.setTimeout("gCodeBrowserMgr.view.mouseOverNode(-1, '');", 50);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

// Set a new sort order for the code browser.
// Returns true iff successful.
CodeBrowserManager.prototype.sortBy = function(key) {
    _gCodeIntel_log.debug("CodeBrowserManager.sortBy(key='"+key+"')");
    try {
        this._validSortByKeys = ["file-order",
                                 "name"];
        var valid = false;
        for (var i = 0; i < this._validSortByKeys.length; ++i) {
            if (key == this._validSortByKeys[i]) {
                valid = true;
                break;
            }
        }

        if (valid) {
            this.view.sortBy(key);
            this.widgets.sortByButton.setAttribute("sortBy", key);
        }
        return valid;
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
        return false;
    }
}


CodeBrowserManager.prototype.haveFilesInWS = function(value) {
    _gCodeIntel_log.debug("CodeBrowserManager.haveFilesInWS(value='"+value+"')");
    try {
        if (value) {
            if (this.widgets.sortByButton.hasAttribute("disabled"))
                this.widgets.sortByButton.removeAttribute("disabled");
            if (this.widgets.filterIcon.hasAttribute("disabled"))
                this.widgets.filterIcon.removeAttribute("disabled");
            if (this.widgets.filterTextbox.hasAttribute("disabled"))
                this.widgets.filterTextbox.removeAttribute("disabled");
            // Ensure that there is a selection in the tree.
            if (!this.view.selection.count) { // there is no selection
                this.view.selection.select(0);
            }
        } else {
            this.widgets.sortByButton.setAttribute("disabled", "true");
            this.widgets.filterIcon.setAttribute("disabled", "true");
            this.widgets.filterTextbox.setAttribute("disabled", "true");
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


CodeBrowserManager.prototype.updateFilter = function() {
    _gCodeIntel_log.debug("CodeBrowserManager.updateFilter()");
    try {
        var filterString = this.widgets.filterTextbox.value;
        this.view.setFilter(filterString);
        if (this.view.rowCount) {
            this.view.selection.select(0);
        }
        this.updateCommands();
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

// Locate the appropriate node for the current scope in the Code Browser.
CodeBrowserManager.prototype.locateScope = function(scimoz, position) {
    _gCodeIntel_log.debug("CodeBrowserManager.locateScope(scimoz, position)");
    try {
        var located = this.view.locateScope(scimoz, position);
        if (located) {
            var filter = document.getElementById("codebrowser-filter-textbox");
            filter.value = "";
            uilayout_ensureTabShown("codebrowser_tab", true);
        }
        return located;
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
        return 0;
    }
}



//---- the UI manager for completions (autocomplete/calltip) in a view

function CodeIntelCompletionUIHandler(path, scimoz, language,
                                      ciCompletionLangSvc)
{
    _gCodeIntel_log.debug("CodeIntelCompletionUIHandler(path, scimoz, language, ciCompletionLangSvc)");
    try {
        this.path = path;
        this.scimoz = scimoz;
        this.language = language;
        this.ciCompletionLangSvc = ciCompletionLangSvc;
        this.completionFillups = ciCompletionLangSvc.completionFillups;

        scimoz.autoCSeparator = ciCompletionLangSvc.completionSeparatorOrd;
        scimoz.autoCSetFillUps(ciCompletionLangSvc.completionFillups);
        // Don't hide when there is no match: may just be mistyped character.
        scimoz.autoCAutoHide = false;
        // Cancel when moving before trigger point, NOT when moving before
        // current pos when the completion list was shown.
        scimoz.autoCCancelAtStart = false;
        // Order members as if all uppercase: "One result of this is that the
        // list should be sorted with the punctuation characters '[', '\',
        // ']', '^', '_', and '`' sorted after letters."
        scimoz.autoCIgnoreCase = true; 

        // Register images for autocomplete lists. Currently only Perl
        // autocomplete supports images.
        var iface = Components.interfaces.koICodeIntelCompletionUIHandler;
        if (this.language == "Perl") {
            scimoz.registerImage(iface.ACIID_CLASS,            markers_getPixmap("chrome://jaguar/skin/images/ac_class.xpm"));
            scimoz.registerImage(iface.ACIID_FUNCTION,         markers_getPixmap("chrome://jaguar/skin/images/ac_function.xpm"));
            scimoz.registerImage(iface.ACIID_MODULE,           markers_getPixmap("chrome://jaguar/skin/images/ac_module.xpm"));
            scimoz.registerImage(iface.ACIID_VARIABLE,         markers_getPixmap("chrome://jaguar/skin/images/ac_variable.xpm"));
            scimoz.registerImage(iface.ACIID_VARIABLE_SCALAR,  markers_getPixmap("chrome://jaguar/skin/images/ac_variable_scalar.xpm"));
            scimoz.registerImage(iface.ACIID_VARIABLE_ARRAY,   markers_getPixmap("chrome://jaguar/skin/images/ac_variable_array.xpm"));
            scimoz.registerImage(iface.ACIID_VARIABLE_HASH,    markers_getPixmap("chrome://jaguar/skin/images/ac_variable_hash.xpm"));
            scimoz.registerImage(iface.ACIID_INTERFACE,        markers_getPixmap("chrome://jaguar/skin/images/ac_interface.xpm"));
            scimoz.registerImage(iface.ACIID_NAMESPACE,        markers_getPixmap("chrome://jaguar/skin/images/ac_namespace.xpm"));
        } else if (this.language == "Python") {
            scimoz.registerImage(iface.ACIID_CLASS,            markers_getPixmap("chrome://jaguar/skin/images/ac_class.xpm"));
            scimoz.registerImage(iface.ACIID_FUNCTION,         markers_getPixmap("chrome://jaguar/skin/images/ac_function.xpm"));
            scimoz.registerImage(iface.ACIID_MODULE,           markers_getPixmap("chrome://jaguar/skin/images/ac_module.xpm"));
            scimoz.registerImage(iface.ACIID_VARIABLE,         markers_getPixmap("chrome://jaguar/skin/images/ac_variable.xpm"));
        } else if (this.language == "Ruby") {
            scimoz.registerImage(iface.ACIID_CLASS,            markers_getPixmap("chrome://jaguar/skin/images/ac_class.xpm"));
            scimoz.registerImage(iface.ACIID_FUNCTION,         markers_getPixmap("chrome://jaguar/skin/images/ac_function.xpm"));
            scimoz.registerImage(iface.ACIID_MODULE,           markers_getPixmap("chrome://jaguar/skin/images/ac_module.xpm"));
            scimoz.registerImage(iface.ACIID_VARIABLE,         markers_getPixmap("chrome://jaguar/skin/images/ac_variable.xpm"));
            //XXX These two should change to a "direcotory" one and something
            //    better than the crap namespace icon for Ruby modules
            //    (different than what CodeIntel calls a "module").
            scimoz.registerImage(iface.ACIID_DIRECTORY,        markers_getPixmap("chrome://jaguar/skin/images/ac_directory.xpm"));
            scimoz.registerImage(iface.ACIID_NAMESPACE,        markers_getPixmap("chrome://jaguar/skin/images/ac_namespace.xpm"));
        }

        this.callTipStack = [];
        // Can't use scimoz.{autoC|callTip}PosStart() for this because (1)
        // there is a bug in .callTipPosStart();
        //      http://mailman.lyra.org/pipermail/scintilla-interest/2004-April/004272.html
        // and (2) the calltip display position might not be the trigger point
        // if the call region is multi-line.
        this._lastCompletionTriggerPos = null;
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}
CodeIntelCompletionUIHandler.prototype.constructor = CodeIntelCompletionUIHandler;

CodeIntelCompletionUIHandler.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.koICodeIntelCompletionUIHandler) &&
        !iid.equals(Components.interfaces.nsISupports)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
}

CodeIntelCompletionUIHandler.prototype.triggerPrecedingCompletion = function()
{
    _gCodeIntel_log.debug("CodeIntelCompletionUIHandler."+
                          "triggerPrecedingCompletion()");
    try {
        // Determine start position.
        var startPos = null;
        if (this.scimoz.callTipActive() || this.scimoz.autoCActive()) {
            startPos = this._lastCompletionTriggerPos - 1;
        } else {
            startPos = this.scimoz.currentPos;
        }

        // Hand off to language service to find and display.
        var errmsg = this.ciCompletionLangSvc.triggerPrecedingCompletionUI(
            this.path, this.scimoz, startPos, this);
        if (errmsg && typeof(StatusBar_AddMessage) != "undefined") {
            StatusBar_AddMessage("Code Intelligence: "+errmsg,
                                 "codeintel", 3000, false);
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }

}

CodeIntelCompletionUIHandler.prototype.setAutoCompleteInfo = function(
    completions, triggerPos)
{
    _gCodeIntel_log.debug("CodeIntelCompletionUIHandler.setAutoCompleteInfo("+
                          "completions, triggerPos)");
    try {
        // If the trigger is no longer relevant, then drop the completions.
        // - if the current position is before the trigger pos
        var curPos = this.scimoz.currentPos;
        if (curPos < triggerPos) {
            _gCodeIntel_log.info("aborting autocomplete at "+triggerPos+
                                 ": cursor is before trigger position");
            return;
        }
        // - if the line changed
        var curLine = this.scimoz.lineFromPosition(curPos);
        var triggerLine = this.scimoz.lineFromPosition(triggerPos);
        if (curLine != triggerLine) {
            _gCodeIntel_log.debug("aborting autocomplete at "+triggerPos+
                                  ": current line number changed");
            return;
        }
        //XXX Should also abort if the trigger character was changed, e.g.
        //       foo.<BS>bar
        //    Will need to pass a "context" argument for this.

        // abort if any of the "fillups" have been typed since
        numTypedAlready = curPos - triggerPos;
        var ch;
        for (var i = triggerPos; i < curPos; i++) {
            ch = this.scimoz.getWCharAt(i);
            if (this.completionFillups.indexOf(ch) != -1) {
                _gCodeIntel_log.debug("aborting autocomplete at "+triggerPos+
                                      ": fillup character typed: '"+ch+"'");
                return;
            }
        }

        // Show the completions UI.
        this._lastCompletionTriggerPos = triggerPos;
        this.scimoz.autoCShow(numTypedAlready, completions);
        if (numTypedAlready > 0) {
            var typedAlready = this.scimoz.getTextRange(triggerPos, curPos);
            this.scimoz.autoCSelect(typedAlready);
        }

    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

CodeIntelCompletionUIHandler.prototype.setCallTipInfo = function(
    calltip, triggerPos, explicit)
{
    _gCodeIntel_log.debug("CodeIntelCompletionUIHandler.setCallTipInfo"+
                          "(calltip, triggerPos="+triggerPos+
                          ", explicit="+explicit+")");
    try {
        if (!explicit) {
            // If the trigger is no longer relevant, then drop the calltip.
            // - if the current position is before the trigger pos
            var curPos = this.scimoz.currentPos;
            if (curPos < triggerPos) {
                _gCodeIntel_log.info("aborting calltip at "+triggerPos+
                                     ": cursor is before trigger position");
                return;
            }
            // - if the current position is outside the call region
            //   c.f. http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html#user-interface-autocomplete-and-calltips
            var region = this.scimoz.getTextRange(triggerPos, curPos);
            var hltStartObj = new Object();
            var hltEndObj = new Object();
            this.ciCompletionLangSvc.getCallTipHighlightSpan(
                region, calltip, hltStartObj, hltEndObj);
            var hltStart = hltStartObj.value;
            var hltEnd = hltEndObj.value;
            if (hltStart == -1) {
                _gCodeIntel_log.info("aborting calltip at "+triggerPos+
                                     ": cursor is outside call region");
                return;
            }
        }

        // Show the callip.
        if (this.scimoz.callTipActive()) {
            this.scimoz.callTipCancel();
        }
        this._lastCompletionTriggerPos = triggerPos;
        this.scimoz.callTipShow(triggerPos, calltip);
        this.scimoz.callTipSetHlt(hltStart, hltEnd);
        callTipItem = {"triggerPos": triggerPos, "calltip": calltip};
        this.callTipStack.push(callTipItem);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}

CodeIntelCompletionUIHandler.prototype.updateCallTip = function() {
    _gCodeIntel_log.debug("CodeIntelCompletionUIHandler.updateCallTip()");
    try {
        if (! this.scimoz.callTipActive()) {
            // The calltip may get cancelled in various other places so
            // we have to make sure that the callTipStack here doesn't
            // grow unboundedly.
            this.callTipStack = [];
            return;
        }

        var curPos = this.scimoz.currentPos;
        var curLine = this.scimoz.lineFromPosition(curPos);
        var callTipItem = this.callTipStack[this.callTipStack.length-1];
        var triggerPos = callTipItem["triggerPos"];
        var calltip = callTipItem["calltip"];
        var callTipPos, triggerColumn;

        // Determine if we should cancel the calltip.
        var cancel = curPos < triggerPos; // cancel if cursor before trigger
        var region, hltStart, hltEnd;
        if (!cancel) {
            region = this.scimoz.getTextRange(triggerPos, curPos);
            var hltStartObj = new Object();
            var hltEndObj = new Object();
            this.ciCompletionLangSvc.getCallTipHighlightSpan(
                region, calltip, hltStartObj, hltEndObj);
            hltStart = hltStartObj.value;
            hltEnd = hltEndObj.value;
            cancel = hltStart == -1;  // cancel if cursor out of call region
        }

        // Cancel if required and fallback to previous calltip, if any.
        if (cancel) {
            // Cancel the current call tip.
            this.scimoz.callTipCancel();
            this.callTipStack.pop();

            // Start the calltip one up in the stack, if there is one.
            if (this.callTipStack.length) {
                callTipItem = this.callTipStack[this.callTipStack.length-1];
                triggerPos = callTipItem["triggerPos"];
                calltip = callTipItem["calltip"];
                if (curPos >= triggerPos) {
                    triggerColumn = this.scimoz.getColumn(triggerPos);
                    callTipPos = this.scimoz.positionAtColumn(
                            curLine, triggerColumn);
                    this._lastCompletionTriggerPos = triggerPos;
                    this.scimoz.callTipShow(callTipPos, calltip);
                    this.updateCallTip();
                }
            }
            return;
        }

        // If the cursor is on a different line from the current display
        // point then we need to move the calltip up or down.
        callTipPos = this.scimoz.callTipPosStart();
        var callTipLine = this.scimoz.lineFromPosition(callTipPos);
        if (callTipLine != curLine) {
            this.scimoz.callTipCancel();
            triggerColumn = this.scimoz.getColumn(triggerPos);
            var newCallTipPos = this.scimoz.positionAtColumn(curLine, triggerColumn);
            this._lastCompletionTriggerPos = triggerPos;
            this.scimoz.callTipShow(newCallTipPos, calltip);
            //dump("XXX moved the calltip to "+newCallTipPos+
            //     ", now it is at "+this.scimoz.callTipPosStart()+"\n");
        }

        // Update the highlighting.
        this.scimoz.callTipSetHlt(hltStart, hltEnd);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


//---- public routines

function CodeIntel_Initialize()
{
    _gCodeIntel_log.debug("CodeIntel_Initialize()");
    try {
        // Setup the pref observer to watch for Code Intel system enabling.
        _gCodeIntel_prefObserver = new _CodeIntelPrefObserver();

        if (gPrefs.getBooleanPref("codeintel_enabled")) {
            _CodeIntel_Activate();
        } else {
            _CodeIntel_Deactivate();
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


function CodeIntel_Finalize()
{
    _gCodeIntel_log.debug("CodeIntel_Finalize()");
    try {
        _CodeIntel_Deactivate();
        _gCodeIntel_prefObserver.finalize();
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


function _CodeIntel_UpgradeCIDBIfNecessary()
{
    _gCodeIntel_log.debug("_CodeIntel_UpgradeCIDBIfNecessary()");
    var stateObj = new Object();
    var detailsObj = new Object();
    var needToUpgrade = null;
    try {
        needToUpgrade = gCodeIntelSvc.needToUpgradeCIDB();
    } catch(ex) {
        var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                            getService(Components.interfaces.koILastErrorService);
        var err = lastErrorSvc.getLastErrorMessage();
        if (!err) {
            err = "<no error message: see 'pystderr.log' error log in your Komodo user data dir>";
        }
        dialog_alert("Could not upgrade your Code Intelligence Database "+
                     "because: "+err+". Your database will be backed up "+
                     "and a new empty database will be created.", null,
                     "Code Intelligence Database");
        gCodeIntelSvc.resetCIDB();
        return;
    }        

    if (needToUpgrade) {
        _gCodeIntel_log.debug("need to upgrade CIDB: starting upgrade wizard");
        var args = new Object();
        args.mode = "upgrade";
        window.openDialog("chrome://jaguar/content/codeintel/buildcidb.xul",
                          "komodo_buildcidb",
                          "chrome,modal=yes,close=yes,resizable",
                          args);
    }
}


function _CodeIntel_Activate()
{
    _gCodeIntel_log.debug("_CodeIntel_Activate()");
    try {
        // Show the Code Browser tab.
        var tab = document.getElementById("codebrowser_tab");
        if (tab.hasAttribute("collapsed"))
            tab.removeAttribute("collapsed");
        // <tabbox> Ctrl+Tab handling uses 'hidden' attribute.
        if (tab.hasAttribute("hidden"))
            tab.removeAttribute("hidden");

        // Setup services.
        gCodeBrowserMgr = new CodeBrowserManager("codebrowser-tree");
        gCodeIntelSvc = Components.classes["@activestate.com/koCodeIntelService;1"]
                              .getService(Components.interfaces.koICodeIntelService);
        try {
            _CodeIntel_UpgradeCIDBIfNecessary();
            gCodeIntelSvc.activateBackEnd();
        } catch(ex) {
            var lastErrorSvc = Components.classes["@activestate.com/koLastErrorService;1"].
                                getService(Components.interfaces.koILastErrorService);
            var err = lastErrorSvc.getLastErrorMessage();
            dialog_internalError(err, err);
            _CodeIntel_Deactivate();
            return;
        }
        gCodeIntelSvc.activateUI(gCodeBrowserMgr);

        gCodeBrowserCtrl = new CodeBrowserController();
        gCodeBrowserMgr.widgets.tree.controllers.insertControllerAt(0, gCodeBrowserCtrl);
        gCodeIntelCtrl = new CodeIntelController();
        window.controllers.insertControllerAt(0, gCodeIntelCtrl);

        // Notify the Code Intel engine of currently open files and projects.
        window.setTimeout(
            function() {
                var i;
                var editorViews = gViewMgr.topView.getViewsByType(true, "editor");
                for (i = 0; i < editorViews.length; i++) {
                    gCodeIntelSvc.ideEvent("opened_document",
                                           editorViews[i].document.file.URI,
                                           editorViews[i].document);
                }
                var projects = gProjectManager._projects;
                for (i = 0; i < projects.length; i++) {
                    gCodeIntelSvc.ideEvent("opened_project", projects[i].part.url,
                                           projects[i].part);
                }
                // Ensure that the backend knows which is the current view.
                var v = gViewMgr.currentView;
                if (v) {
                    gCodeIntelSvc.ideEvent("switched_current_document",
                                           v.document.file.URI, v.document);
                }
            },
            2000
        );

        // Activate the Code Browser.
        CodeIntel_UpdateCodeBrowserVisibility();

        // Observe some other file and project changes.
        _gCodeIntel_observer = new _CodeIntelObserver();

        gCodeIntelActive = true;
        var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                     getService(Components.interfaces.nsIObserverService);
        obsSvc.notifyObservers(null, "codeintel_activated", null);
        window.updateCommands('codebrowser');

        // Offer to prepare the DB for the various languages if it has not
        // been done for the current database version.
        if (!gCodeIntelSvc.haveRunBuildCIDBWizard()) {
            var args = new Object();
            args.launchedAutomatically = true;
            window.openDialog("chrome://jaguar/content/codeintel/buildcidb.xul",
                              "komodo_buildcidb",
                              "chrome,close=yes,resizable",
                              args);
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


function _CodeIntel_Deactivate()
{
    _gCodeIntel_log.debug("_CodeIntel_Deactivate()");
    try {
        gCodeIntelActive = false;

        // Shutdown the various services if they are up and running.
        if (gCodeIntelSvc) {
            gCodeIntelSvc.deactivate();
            gCodeIntelSvc = null;
            gCodeBrowserMgr.finalize();
            if (gCodeBrowserCtrl) {
                gCodeBrowserMgr.widgets.tree.controllers.removeController(gCodeBrowserCtrl);
                window.controllers.removeController(gCodeIntelCtrl);
            }
            gCodeBrowserMgr = null;
            if (_gCodeIntel_observer) {
                _gCodeIntel_observer.finalize();
                _gCodeIntel_observer = null;
            }
        }

        // Hide the Code Browser tab.
        var tabbox = document.getElementById("leftTabBox");
        var tab = document.getElementById("codebrowser_tab");
        if (tabbox.selectedTab == tab) {
            var alternate = tab.previousSibling;
            if (!alternate)
                alternate = tab.nextSibling;
            if (alternate)
                tabbox.selectedTab = alternate;
        }
        tab.setAttribute("collapsed", "true");
        // <tabbox> Ctrl+Tab handling uses 'hidden' attribute.
        tab.setAttribute("hidden", "true");

        var obsSvc = Components.classes["@mozilla.org/observer-service;1"].
                     getService(Components.interfaces.nsIObserverService);
        obsSvc.notifyObservers(null, "codeintel_deactivated", null);

        window.updateCommands('codebrowser');
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


// Called by Komodo's UI handling to notify the Code Intel system that
// the visibility of the Code Browser may have changed.
function CodeIntel_UpdateCodeBrowserVisibility()
{
    _gCodeIntel_log.debug("CodeIntel_UpdateCodeBrowserVisibility()");
    try {
        if (!gCodeIntelActive) return;
        if (uilayout_isCodeBrowserTabShown()) {
            gCodeBrowserMgr.visible = true;
            _gCodeIntel_log.info("Code Browser tab is showing");
        } else {
            gCodeBrowserMgr.visible = false;
            _gCodeIntel_log.info("Code Browser tab is hidden");
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


// Locate the node in the Code Browser corresponding to the current scope.
function CodeIntel_LocateCurrentScope()
{
    _gCodeIntel_log.debug("CodeIntel_LocateCurrentScope()");
    try {
        if (!gCodeIntelActive) return;
        var view = gViewMgr.currentView;
        if (view && view.scimoz && view.document) {
            var scimoz = view.scimoz;
            var located = gCodeBrowserMgr.locateScope(scimoz,
                                                      scimoz.currentPos);
            if (!located) {
                var line = scimoz.lineFromPosition(scimoz.currentPos)+1;
                dialog_alert("Could not find a node corresponding to the "+
                             "scope at "+view.document.baseName+":"+line+
                             " in the Code Browser.");
            }
        }
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


// Search for the given symbol/module in the Object Browser.
//
//  "symbol" is the symbol search pattern.
//  "module" is the module search pattern.
//      If neither "symbol" nor "module" is specified then the symbol pattern
//      is taken as the selection or word under the cursor in the current
//      buffer.
//  "exact" specifies whether search for the exact pattern.
//  "language" is the language to which to limit the search. If not specified
//      (or null) the current buffer language is used (if it is one of the
//      language supported by the Code Intel system. Specify the empty
//      string to search all languages.
//
function CodeIntel_FindSymbol(symbol /*=null*/, module /*=null*/,
                              exact /*=false*/, language /*=null*/)
{
    _gCodeIntel_log.debug("CodeIntel_FindSymbol(symbol="+symbol+
                          ", module="+module+", exact="+exact+
                          ", language="+language+")");
    try {
        if (!gCodeIntelActive) return;

        if ((typeof(symbol) == "undefined" || symbol == null)
            && (typeof(module) == "undefined" || module == null))
        {
            var scimoz = gViewMgr.currentView.scimoz;
            symbol = scimoz.selText;
            if (! symbol) {
                symbol = Interpolate_getWordUnderCursor(scimoz);
            }
        }
        if (typeof(exact) == "undefined" || exact == null) {
            exact = false;
        }
        if (typeof(language) == "undefined" || language == null) {
            language = gViewMgr.currentView.document.language;
            if (! gCodeIntelSvc.isSupportedLanguage(language)) {
                language = "";
            }
        }

        launch_objectBrowser(symbol, module, exact, language);
    } catch(ex) {
        _gCodeIntel_log.exception(ex);
    }
}


