/* Copyright (c) 2000-2003 ActiveState Corp.
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: JavaScript; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Generic OK/Cancel dialog from which the user can select from a list of
 * items.
 *
 * Usage:
 *  All dialog interaction is done via an object passed in and out as the first
 *  window argument: window.arguments[0]. All these arguments, except "items"
 *  are optional. See dialog_selectFromList() documentation for details.
 *      .title          the dialog title
 *      .prompt         the question to ask
 *      .items          is the list of items from which the user can choose
 *      .selectionCondition is a string defining how selection behaves
 *      .stringifier    a function that, if non-null, will be called to
 *                      stringify each item for display.
 *      .doNotAskUI     show the "Don't ask me again" UI
 *      .yesNoCancel    is a boolean to make this a Yes/No/Cancel dialog
 *  On return window.arguments[0] has:
 *      .retval         "OK" or "Cancel"
 *      .selected       (only iff .retval=="OK") a list of selected items
 *      .doNotAsk       (iff .doNotAskUI) a boolean indicating if this question
 *                      need be asked again.
 *
 */

var log = getLoggingMgr().getLogger("dialogs.selectFromList");
//log.setLevel(LOG_DEBUG);

var gItems = null;          // List of items from which user is to select.
var gSelectionCondition = null;
var gDoNotAskUI = false;    // true iff "Don't ask me again" UI is being used.
var gStringifier = null;


var gItemsTreeView = {  // The nsITreeView object for the list of items.
    getCellText : function(row, column) {
        if (gStringifier) {
            return gStringifier(gItems[row]);
        } else if (typeof(gItems[row]) == 'string') {
            return gItems[row];
        } else {
            return gItems[row].text;
        }
    },
    rowCount : function() { return gItems.length; },

    // Everything below here is just to satisfy the interface, and not all of it may
    // be required.
    tree : null,
    isSeparator : function(index) {return false;},
    isContainer : function(index) {return false;},
    setTree : function(out) { this.tree = out; },
    getRowProperties : function(row,prop){},
    getColumnProperties : function(column,prop){},
    getCellProperties : function(row,prop){},
    isContainer : function(row) {return false;},
    cycleCell: function(row, colId) {},
    setRowCount : function(rowCount) {
        this.rowCount = rowCount;
        this.tree.beginUpdateBatch();
        this.tree.rowCountChanged(0, this.rowCount);
        this.tree.invalidate();
        this.tree.endUpdateBatch();
    },
    selectionChanged : function() {},
    performAction : function(action) {},
    isSorted : function() {return true;},
    getImageSrc : function() {return null;},
    cycleHeader : function() {},
    /* our own defined methods */
    sort: function() {}
};


//---- interface routines for XUL

function OnLoad()
{
    var dialog = document.getElementById("dialog_selectFromList")
    var acceptButton = dialog.getButton("accept");
    var noButton = dialog.getButton("extra1");
    var cancelButton = dialog.getButton("cancel");
    var itemsTree = document.getElementById("items");
    var msg;
    // .title
    if (typeof window.arguments[0].title != "undefined" &&
        window.arguments[0].title != null) {
        window.title = window.arguments[0].title;
    } else {
        window.title = "Komodo";
    }

    // .prompt
    var descWidget = document.getElementById("prompt");
    var desc = window.arguments[0].prompt;
    if (typeof desc != "undefined" && desc != null) {
        var textNode = document.createTextNode(desc);
        descWidget.appendChild(textNode);
    } else {
        descWidget.setAttribute("collapsed", "true");
    }

    // .items
    gItems = window.arguments[0].items;
    if (typeof gItems == "undefined" || gItems == null) {
        //XXX Is this the kind of error handling we want to do in onload
        //    handlers?
        msg = "Internal Error: illegal 'items' value for "
                  +"Select From List dialog: '"+gItems+"'.";
        log.error(msg);
        alert(msg);
        window.close();
    }

    itemsTree.treeBoxObject.view = gItemsTreeView;
    gItemsTreeView.setRowCount(gItems.length);

    // .selectionCondition
    if (typeof window.arguments[0].selectionCondition == "undefined" ||
        window.arguments[0].selectionCondition == null) {
        gSelectionCondition = "one-or-more";
    } else {
        gSelectionCondition = window.arguments[0].selectionCondition;
    }
    switch (gSelectionCondition) {
    case "one-or-more":
        SelectAll();
        break;
    case "zero-or-more":
        SelectAll();
        break;
    case "one":
        itemsTree.setAttribute("seltype", "single");
        gItemsTreeView.selection.select(0);
        document.getElementById("select-all").setAttribute("collapsed", true);
        document.getElementById("clear-all") .setAttribute("collapsed", true);
        break;
    default:
        //XXX Is this the kind of error handling we want to do in onload
        //    handlers?
        msg = "Internal Error: illegal selection condition value for "
                  +"Select From List dialog: '"+gSelectionCondition+"'.";
        log.error(msg);
        alert(msg);
        window.close();
    }

    // .stringifier
    if (typeof window.arguments[0].stringifier != "undefined" &&
        window.arguments[0].stringifier != null) {
        gStringifier = window.arguments[0].stringifier;
    }

    // .doNotAskUI
    if (typeof window.arguments[0].doNotAskUI != "undefined" &&
        window.arguments[0].doNotAskUI != null) {
        gDoNotAskUI = window.arguments[0].doNotAskUI;
    }
    if (gDoNotAskUI) {
        document.getElementById("doNotAsk-checkbox")
                .removeAttribute("collapsed");
    }

    // .yesNoCancel
    var yesNoCancel = window.arguments[0].yesNoCancel;
    if (typeof yesNoCancel == "undefined" || yesNoCancel == null) {
        yesNoCancel = false;
    }
    if (yesNoCancel) {
        if (gSelectionCondition != "zero-or-more") {
            msg = "Internal Error: Attempted to use yesNoCancel=true "+
                      "and selectionCondition='"+gSelectionCondition+
                      "' for a Select From List dialog. This does not make "+
                      "sense and is illegal.";
            log.error(msg);
            alert(msg);
            window.close();
        }
        acceptButton.setAttribute("label", "Yes");
        acceptButton.setAttribute("accesskey", "y");
        noButton.setAttribute("label", "No");
        noButton.setAttribute("accesskey", "n");
        cancelButton.setAttribute("accesskey", "c");
    } else {
        acceptButton.setAttribute("accesskey", "o");
        noButton.setAttribute("collapsed", "true");
        cancelButton.setAttribute("accesskey", "c");
    }

}


function OnSelectionChange()
{
    var dialog = document.getElementById("dialog_selectFromList");
    var acceptButton = dialog.getButton("accept");
    var noButton = dialog.getButton("extra1");

    if (gSelectionCondition == "one-or-more") {
        if (gItemsTreeView.selection.count == 0) {
            acceptButton.setAttribute("disabled", "true");
            noButton.setAttribute("disabled", "true");
        } else if (acceptButton.hasAttribute("disabled")) {
            acceptButton.removeAttribute("disabled");
            noButton.removeAttribute("disabled");
        }
    }
    if (gDoNotAskUI) {
        var checkbox = document.getElementById("doNotAsk-checkbox");
        if (gSelectionCondition == "one-or-more"
            && gItemsTreeView.selection.count == 0) {
            checkbox.setAttribute("disabled", "true");
            checkbox.setAttribute("checked", "false");
        } else {
            if (checkbox.hasAttribute("disabled"))
                checkbox.removeAttribute("disabled");
        }
    }
}

// The onclick handler on the <tree> is to allow on to double-click on one of
// the selection options to select it and close the dialog. This is only
// allowed for gSelectionCondition=="one".
function OnClick(event)
{
    if (gSelectionCondition != "one") return;

    // c.f. mozilla/mailnews/base/resources/content/threadPane.js
    var t = event.originalTarget;

    // single-click on a column
    if (t.localName == "treecol") return;

    // double-click in the tree body
    if (event.detail == 2 && t.localName == "treechildren") {
        if (Accept()) {
            window.close();
        }
    }
}


function SelectAll()
{
    gItemsTreeView.selection.selectAll();
    OnSelectionChange();
    document.getElementById("items").focus();
}


function ClearAll()
{
    gItemsTreeView.selection.clearSelection();
    OnSelectionChange();
    document.getElementById("items").focus();
}


function Accept()
{
    var acceptButton = document.getElementById("dialog_selectFromList").getButton("accept");
    window.arguments[0].retval = acceptButton.getAttribute("label");

    if (gDoNotAskUI) {
        var checkbox = document.getElementById("doNotAsk-checkbox");
        if (! checkbox.getAttribute("disabled")) {
            window.arguments[0].doNotAsk = checkbox.checked;
        } else {
            window.arguments[0].doNotAsk = false;
        }
    }

    // Get the list of selected items.
    var selected = new Array();
    var i, item;
    for (i = 0; i < gItems.length; ++i) {
        if (gItemsTreeView.selection.isSelected(i)) {
            item = gItems[i];
            selected.push(item);
        }
    }
    window.arguments[0].selected = selected;

    return true;
}


function No()
{
    var noButton = document.getElementById("dialog_selectFromList").getButton("extra1");
    window.arguments[0].retval = noButton.getAttribute("label");

    ClearAll(); // do this first because it might affect the checkbox state
    if (gDoNotAskUI) {
        var checkbox = document.getElementById("doNotAsk-checkbox");
        if (! checkbox.getAttribute("disabled")) {
            window.arguments[0].doNotAsk = checkbox.checked;
        } else {
            window.arguments[0].doNotAsk = false;
        }
    }

    window.arguments[0].selected = [];

    window.close();
    return true;
}


function Cancel()
{
    window.arguments[0].retval = "Cancel";
    if (gDoNotAskUI) {
        // Don't skip this dialog next time if it was cancelled this time.
        window.arguments[0].doNotAsk = false;
    }
    return true;
}

