/* Copyright (c) 2000-2003 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

/* -*- Mode: JavaScript; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * Functions for accessing common and relatively generic dialogs.
 *
 * Summary of Dialogs:
 *  dialog_alert            Display an alert message with optional additional
 *                          text. (Use the plain alert() if you can.)
 *  dialog_prompt           Ask the user for a string.
 *  dialog_yesNoCancel      Ask the user a Yes/No/Cancel question.
 *  dialog_yesNo            Ask the user a Yes/No question.
 *  dialog_okCancel         State a proposed action and ask the user to verify
 *                          with OK/Cancel.
 *  dialog_selectFromList   Ask the user to select from a list of strings.
 *  dialog_customButtons    Show the user some prompt and request one of a
 *                          number of responses.
 *  dialog_editEnvVar       Edit an environment variable.
 *  dialog_internalError    Notify the user of an unexpected internal error.
 *  dialog_authenticate     Username/password authentication dialog.
 *
 * Notes on common "doNotAskPref" option:
 *      Many of the dialog_*() methods in this module have a "doNotAskPref"
 *      optional argument that will result in
 *          [ ] Don't ask me again.
 *      functionality on the dialog. If the argument is null or left empty no
 *      such UI or functionality will be shown. Otherwise it is uses one or two
 *      prefs to lookup and store relevant data. Every usage will use a boolean
 *      pref of the name:
 *          donotask_<doNotAskPref>
 *      Some dialogs will also require/use a _string_ pref of the name:
 *          donotask_action_<doNotAskPref>
 *      If the boolean pref is true the dialog is NOT shown and dialog_*()
 *      will return as it did the last time the dialog was shown. See each
 *      method's documentation for details.
 *
 *      Using a new "doNotAskPref" value, say "FOO", in your code will require
 *      you to add the first or both of the following lines to
 *      default-prefs.xml:
 *          <boolean id="donotask_FOO">0</boolean>
 *          <string id="donotask_action_FOO"></string>
 *
 *      Legacy Note: Some older implementations used boolean prefs with the
 *      "dontask_" prefix (notice the slight spelling difference). The name has
 *      been changed to avoid collision.
 *
 * NOTE:
 *  Only purely generic stuff should go in here at all, since it is utilized by
 *  many XUL files!
 */


//---- globals

var dialogsLog = getLoggingMgr().getLogger("dialogs");

var _dialogs_gPrefs = Components.classes['@activestate.com/koPrefService;1'].
                      getService(Components.interfaces.koIPrefService).prefs;



//---- public methods

// Ask the user a Yes/No/Cancel question.
//
// All arguments can be left blank or specified as null to get a default value.
//  "prompt" is question to ask.
//  "response" is the default response. This button is shown as the default.
//      Must be one of "Yes", "No", or "Cancel". If left empty of null the
//      default response is "Yes".
//  "text" allows you to specify a string of text that will be display in a
//      non-edittable selectable text box. If "text" is null or no specified
//      then this textbox will no be shown.
//  "title" is the dialog title.
//  "doNotAskPref", uses/requires the following two prefs:
//      boolean donotask_<doNotAskPref>: whether to not show the dialog
//      string donotask_action_<doNotAskPref>: "Yes" or "No"
//
// Returns "Yes", "No" or "Cancel", whichever the user selected.
//
function dialog_yesNoCancel(prompt, response, text, title, doNotAskPref)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(response) == 'undefined') response = "Yes";
    if (typeof(text) == 'undefined') text = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(doNotAskPref) == 'undefined') doNotAskPref = null;

    // Break out early if "doNotAskPref" prefs so direct.
    var bpref = null, spref = null;
    if (doNotAskPref) {
        bpref = "donotask_"+doNotAskPref;
        spref = "donotask_action_"+doNotAskPref;
        if (_dialogs_gPrefs.getBooleanPref(bpref)) {
            var action = _dialogs_gPrefs.getStringPref(spref);
            if (action == "No" || action == "Yes") {
                return action;
            } else {
                dialogsLog.error("illegal action for Yes/No/Cancel dialog in '" +
                                 spref + "' preference: '" + action + "'");
                // Reset the boolean pref.
                _dialogs_gPrefs.setBooleanPref(bpref, false);
            }
        }
    }

    // Show the dialog.
    var obj = new Object();
    obj.prompt = prompt;
    obj.response = response;
    obj.text = text;
    obj.title = title;
    obj.doNotAskUI = doNotAskPref != null;
    window.openDialog("chrome://jaguar/content/dialogs/yesNoCancel.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);

    if (doNotAskPref && obj.doNotAsk) {
        _dialogs_gPrefs.setBooleanPref(bpref, true);
        _dialogs_gPrefs.setStringPref(spref, obj.response);
    }
    return obj.response;
}


// Ask the user a Yes/No question.
//
// All arguments can be left blank or specified as null to get a default value.
//  "prompt" is question to ask.
//  "response" is the default response. This button is shown as the default.
//      Must be one of "Yes" or "No". If left empty or null the default
//      response is "Yes".
//  "text" allows you to specify a string of text that will be display in a
//      non-edittable selectable text box. If "text" is null or no specified
//      then this textbox will no be shown.
//  "title" is the dialog title.
//  "doNotAskPref", uses/requires the following two prefs:
//      boolean donotask_<doNotAskPref>: whether to not show the dialog
//      string donotask_action_<doNotAskPref>: "Yes" or "No"
//
// Returns "Yes" or "No", whichever the user selected. Note that cancelling the
// dialog via <Esc> (or equivalent) is taken to mean a "No" answer from the
// user; though there is a subtle difference in that the possible "Don't ask me
// again." checkbox setting is NOT honoured if the dialog is cancelled.
//
function dialog_yesNo(prompt, response, text, title, doNotAskPref)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(response) == 'undefined') response = null;
    if (typeof(text) == 'undefined') text = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(doNotAskPref) == 'undefined') doNotAskPref = null;

    // Break out early if "doNotAskPref" prefs so direct.
    var bpref = null, spref = null;
    if (doNotAskPref) {
        bpref = "donotask_"+doNotAskPref;
        spref = "donotask_action_"+doNotAskPref;
        if (_dialogs_gPrefs.getBooleanPref(bpref)) {
            var action = _dialogs_gPrefs.getStringPref(spref);
            if (action == "No" || action == "Yes") {
                return action;
            } else {
                dialogsLog.error("illegal action for Yes/No/Cancel dialog in '" +
                                 spref + "' preference: '" + action + "'");
                // Reset the boolean pref.
                _dialogs_gPrefs.setBooleanPref(bpref, false);
            }
        }
    }

    // Show the dialog.
    var obj = new Object();
    obj.prompt = prompt;
    obj.response = response;
    obj.text = text;
    obj.title = title;
    obj.doNotAskUI = doNotAskPref != null;
    window.openDialog("chrome://jaguar/content/dialogs/yesNo.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);

    if (doNotAskPref && obj.doNotAsk) {
        _dialogs_gPrefs.setBooleanPref(bpref, true);
        _dialogs_gPrefs.setStringPref(spref, obj.response);
    }
    return obj.response;
}


// Ask the user a OK/Cancel question.
//
// All arguments can be left blank or specified as null to get a default value.
//  "prompt" is message to show.
//  "response" is the default response. This button is shown as the default.
//      Must be one of "OK" or "Cancel". If left empty or null the default
//      response is "OK".
//  "text" allows you to specify a string of text that will be display in a
//      non-edittable selectable text box. If "text" is null or no specified
//      then this textbox will no be shown.
//  "title" is the dialog title.
//  "doNotAskPref", uses/requires the following pref:
//      boolean donotask_<doNotAskPref>: whether to not show the dialog
//
// Returns "OK" or "Cancel", whichever the user selected.
//
function dialog_okCancel(prompt, response, text, title, doNotAskPref)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(response) == 'undefined') response = null;
    if (typeof(text) == 'undefined') text = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(doNotAskPref) == 'undefined') doNotAskPref = null;

    // Break out early if "doNotAskPref" prefs so direct.
    var bpref = null;
    if (doNotAskPref) {
        bpref = "donotask_"+doNotAskPref;
        if (_dialogs_gPrefs.getBooleanPref(bpref)) {
            return "OK";
        }
    }

    // Show the dialog.
    var obj = new Object();
    obj.prompt = prompt;
    obj.response = response;
    obj.text = text;
    obj.title = title;
    obj.doNotAskUI = doNotAskPref != null;
    window.openDialog("chrome://jaguar/content/dialogs/okCancel.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);

    if (doNotAskPref && obj.doNotAsk) {
        _dialogs_gPrefs.setBooleanPref(bpref, true);
    }
    return obj.response;
}

// Show the user some prompt and request one of a number of responses.
//
// Arguments:
//  "prompt" is message to show.
//  "buttons" is a list of strings, each a label of a button to show.
//      Currently this is limited to three buttons, plus an optional "Cancel"
//      button. For example to mimic (mostly) dialog_yesNo use ["Yes", "No"]
//      and to mimic dialog_yesNoCancel use ["Yes", "No", "Cancel"].
//  "response" is the default response. This button is shown as the default.
//      It must be one of the strings in "buttons" or empty, in which case the
//      first button is the default.
//  "text" allows you to specify a string of text that will be display in a
//      non-edittable selectable text box. If "text" is null or no specified
//      then this textbox will no be shown.
//  "title" is the dialog title.
//  "doNotAskPref", uses/requires the following two prefs:
//      boolean donotask_<doNotAskPref>: whether to not show the dialog
//      string donotask_action_<doNotAskPref>: the name of the button pressed
//
// Returns the name of the button pressed, i.e. one of the strings in
// "buttons", or "Cancel" if the dialog was cancelled (it is possible the
// cancel a dialog without a cancel button.
//
//TODO:
// - Add accesskey support to the custom buttons: if a char is preceded by an
//   ampersand that use that for the access key. See
//   http://www.mozilla.org/projects/ui/accessibility/accesskey.html
//
function dialog_customButtons(prompt, buttons, response, text, title,
                              doNotAskPref)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(buttons) == 'undefined') buttons = null;
    if (typeof(response) == 'undefined') response = null;
    if (typeof(text) == 'undefined') text = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(doNotAskPref) == 'undefined') doNotAskPref = null;

    // Break out early if "doNotAskPref" prefs so direct.
    var bpref = null, spref = null;
    if (doNotAskPref) {
        bpref = "donotask_"+doNotAskPref;
        spref = "donotask_action_"+doNotAskPref;
        if (_dialogs_gPrefs.getBooleanPref(bpref)) {
            return _dialogs_gPrefs.getStringPref(spref);
        }
    }

    // Show the dialog.
    var obj = new Object();
    obj.prompt = prompt;
    obj.buttons = buttons;
    obj.response = response;
    obj.text = text;
    obj.title = title;
    obj.doNotAskUI = doNotAskPref != null;
    window.openDialog("chrome://jaguar/content/dialogs/customButtons.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);

    if (doNotAskPref && obj.doNotAsk) {
        _dialogs_gPrefs.setBooleanPref(bpref, true);
        _dialogs_gPrefs.setStringPref(spref, obj.response);
    }
    return obj.response;
}

// Display an alert.
// NOTE: If the standard alert() meets your needs, then USE IT. Basically this
//       means that you only need the "prompt" argument. There is no point in
//       using our custom dialog when a standard one will do.
//
// All arguments can be left blank or specified as null to get a default value.
//  "prompt" is message to show.
//  "text" allows you to specify a string of text that will be display in a
//      non-edittable selectable text box. If "text" is null or no specified
//      then this textbox will no be shown.
//  "title" is the dialog title.
//  "doNotAskPref", uses/requires the following pref:
//      boolean donotask_<doNotAskPref>: whether to not show the dialog
//
// This function does not return anything.
//
function dialog_alert(prompt, text, title, doNotAskPref)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(text) == 'undefined') text = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(doNotAskPref) == 'undefined') doNotAskPref = null;

    // Break out early if "doNotAskPref" prefs so direct.
    var bpref = null;
    if (doNotAskPref) {
        bpref = "donotask_"+doNotAskPref;
        if (_dialogs_gPrefs.getBooleanPref(bpref)) {
            return;
        }
    }

    // Show the dialog.
    var obj = new Object();
    obj.prompt = prompt;
    obj.text = text;
    obj.title = title;
    obj.doNotAskUI = doNotAskPref != null;
    window.openDialog("chrome://jaguar/content/dialogs/alert.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);
    if (doNotAskPref && obj.doNotAsk) {
        _dialogs_gPrefs.setBooleanPref(bpref, true);
    }
}


// A dialog to query the user for a string in a textbox.
//
// All arguments can be left blank or specified as null to get a default value.
//  "prompt" is text to place before the input textbox.
//  "label" is a label to place on the textbox.
//  "value" is a default value for the textbox.
//  "title" is the dialog title.
//  "mruName" can be specified (a string) to get an MRU on the text box. The
//      value of the string is the namespace for the MRU.
//  "validator" is a callable object to validate the current value when the
//      user presses "OK". It is called with the current value as the only
//      argument.  If the function returns false the "OK" is ignored.
//  "multiline" is a boolean indicating that the textbox should be multiline.
//      "mruName" and "multiline" are mutually exclusive.
//  "screenX", "screenY" allow one to specify a dialog position other than
//      the alert position.
//
// It returns null if the dialog was cancelled. Otherwise it returns the value
// entered by the user.
//
// Dev Note: Not adding "doNotAsk" functionality until there is a proven need
// for it. What should the pref store: (1) a hard coded value to use?, (2) a
// special "<default>" string specifying that the given default "value" should
// just be returned?, (3) both?
//
function dialog_prompt(prompt, label, value, title, mruName, validator,
                       multiline, screenX, screenY)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(label) == 'undefined') label = null;
    if (typeof(value) == 'undefined') value = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(mruName) == 'undefined') mruName = null;
    if (typeof(validator) == 'undefined') validator = null;
    if (typeof(multiline) == 'undefined') multiline = null;
    if (typeof(screenX) == 'undefined') screenX = null;
    if (typeof(screenY) == 'undefined') screenY = null;
    if (mruName && multiline) {
        log.warn("Cannot use both 'mruName' and 'multiline' on prompt "+
                 "dialogs. 'mruName' will be ignored.");
        mruName = null;
    }

    var obj = new Object();
    obj.prompt = prompt;
    obj.label = label;
    obj.value = value;
    obj.title = title;
    obj.mruName = mruName;
    obj.validator = validator;
    obj.multiline = multiline;
    obj.screenX = screenX;
    obj.screenY = screenY;
    window.openDialog("chrome://jaguar/content/dialogs/prompt.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);
    if (obj.retval == "OK") {
        return obj.value;
    } else {
        return null;
    }
}


// This dialog has the same API as the old authDialog. However it is a
// stub for dialog_authenticate2(). We should move to the later. I
// want to change the API slightly to conform more with the rest of the
// dialogs in this file.
//
// DEPRECATED. We should move to dialog_authenticate2().
function dialog_authenticate(title, message, loginname, allowAnonymous,
                             allowPersist)
{
    var prompt = message;
    var username = loginname;
    return dialog_authenticate2(message, // prompt
                                       null, // server
                                       loginname, // username
                                       allowAnonymous,
                                       title, // title
                                       null); // login
}


// A dialog to get a user's username and password.
//
// All arguments can be left blank or specified as null to get a
// default value:
//  "prompt" is text to place before the input textbox.
//  "server" is the server URI to which the user is logging in.
//  "username" is a default username to start with.
//  "allowAnonymous" is a boolean indicating that the user may login
//      anonymously and to provide UI to assist with that.
//  "title" is the dialog title. Defaults to "Login As".
//  "login" is a callback to do the login. It is called when the user
//      pressed "Login". The call signature is:
//          login(server, username, password)
//      If the function returns false the "Login" is ignored.
//
// It returns null if the user cancels the dialog. Otherwise it returns
// an object with the following attributes:
//  .server        the original server passed in
//  .username   the username entered by the user
//  .password   the password entered by the user
//
function dialog_authenticate2(prompt, server, username, allowAnonymous,
                              title, login)
{
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(server) == 'undefined') server = null;
    if (typeof(username) == 'undefined') username = null;
    if (typeof(allowAnonymous) == 'undefined') allowAnonymous = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(login) == 'undefined') login = null;

    var obj = new Object();
    obj.prompt = prompt;
    obj.server = server;
    obj.username = username;
    obj.allowAnonymous = allowAnonymous;
    obj.title = title;
    obj.login = login;
    window.openDialog("chrome://jaguar/content/dialogs/authenticate.xul",
                      "_blank",
                      "chrome,modal,titlebar",
                      obj);
    if (obj.retval == "Cancel") {
        return null;
    } else {
        return obj;
    }
}



// Ask the user to select from a list of strings (often a list of files/URLs).
//
// Arguments (all except "items" are optional):
//  "title" is the dialog title. If not specified (or null) a default title
//      will be used.
//  "prompt" is text to place before the listbox.
//  "items" is the list of items from which the user can choose. An "item" can
//      either simply be a string (which itself will be shown in the presented
//      listbox, or it can be a JavaScript object. If the latter, two
//      attributes are significant:
//          item.text   The string to be displayed in the listbox. (The
//                      "stringifier" argument, if used, overrides this.)
//          item.id     A string that is used to identify that particular item.
//                      This id is used in prefs when the "doNotAskPref"
//                      feature is being used.
//  "selectionCondition" is a string describing how many items can/must be
//      selected for the OK button to be enabled. The following values are
//      understood:
//          "one-or-more" (the default) requires that one or more items be
//              selected. By default all items will be selected.
//          "zero-or-more" allows any number of items to be selected. By
//              default all items will be selected.
//          "one" only allows exactly one to be selected. By default the first
//              item will be selected.
//  "stringifier" is a function that, if non-null, will be called to stringify
//      each item for display. When your items are URIs it is common to use
//      uriparse_displayPath for this argument.
//  "doNotAskPref", uses/requires the following two prefs:
//          boolean donotask_<doNotAskPref>: whether to not show the dialog
//          string donotask_action_<doNotAskPref>: "All" or "None" or item id
//      This prefs are only set in the following circumstances:
//          - selectionCondition="one-or-more" and all of the items are
//            selected (string pref set to "All")
//          - selectionCondition="zero-or-more" and all or none of the items
//            are selected (string pref set to "All" or "None")
//          - selectionCondition="one" and an item is selected (string pref set
//            to the item id or the item's string representation, see "items"
//            discussion above)
//  "yesNoCancel" is a boolean indicating that the dialog should use
//      Yes/No/Cancel buttons instead of the usual OK/Cancel. The "No" button,
//      the new one, returns an empty list for the selected items. By default
//      this is false. Note: yesNoCancel=true does not make sense if
//      selectionCondition!="zero-or-more", and an error will be raised in this
//      condition.
//
// It returns null if the dialog was cancelled. Otherwise it returns the list
// of selected items.
//
// Dev Note: This is getting pretty heavily burdened. Perhaps it is time for
// dialog_selectOneFromList() and dialog_selectMultiFromList().
//
function dialog_selectFromList(title, prompt, items, selectionCondition,
                               stringifier, doNotAskPref, yesNoCancel)
{
    if (typeof(title) == 'undefined') title = null;
    if (typeof(prompt) == 'undefined') prompt = null;
    if (typeof(items) == 'undefined') items = null;
    if (typeof(selectionCondition) == 'undefined') selectionCondition = "one-or-more";
    if (typeof(stringifier) == 'undefined') stringifier = null;
    if (typeof(doNotAskPref) == 'undefined') doNotAskPref = null;
    if (typeof(doNotAskPref) == 'undefined') yesNoCancel = false;

    // Break out early if "doNotAskPref" prefs so direct.
    var bpref = null, spref = null;
    if (doNotAskPref) {
        bpref = "donotask_"+doNotAskPref;
        spref = "donotask_action_"+doNotAskPref;
        if (_dialogs_gPrefs.getBooleanPref(bpref)) {
            switch (selectionCondition) {
            case "one-or-more":
            case "zero-or-more":
                var action = _dialogs_gPrefs.getStringPref(spref);
                if (action == "All") {
                    return items;
                } else if (action == "None") {
                    return [];
                } else {
                    dialogsLog.error("illegal action for Select From List "+
                                     "dialog in '"+spref+"' preference: '"+
                                     action+"'");
                    // Reset the prefs.
                    _dialogs_gPrefs.setBooleanPref(bpref, false);
                    _dialogs_gPrefs.setStringPref(spref, "");
                }
                break;
            case "one":
                var id = _dialogs_gPrefs.getStringPref(spref);
                for (var i = 0; i < items.length; ++i) {
                    var itemId = null;
                    if (typeof(items[i]) == 'string') {
                        itemId = items[i];
                    } else {
                        itemId = items[i].id;
                    }
                    if (id == itemId) {
                        return [items[i]];
                    }
                }
                dialogsLog.warn("identified item, '"+id+"', for Select From "+
                                "List dialog not found in items: pref='" +
                                spref+"'")
                // Reset the prefs.
                _dialogs_gPrefs.setBooleanPref(bpref, false);
                _dialogs_gPrefs.setStringPref(spref, "");
                break;
            default:
                throw("illegal selection condition: '"+selectionCondition+"'");
            }
        }
    }

    // Show the dialog.
    var obj = new Object();
    obj.title = title;
    obj.prompt = prompt;
    obj.items = items;
    obj.selectionCondition = selectionCondition;
    obj.stringifier = stringifier;
    obj.doNotAskUI = doNotAskPref != null;
    obj.yesNoCancel = yesNoCancel;
    window.openDialog("chrome://jaguar/content/dialogs/selectFromList.xul",
                      "_blank",
                      "chrome,dialog=no,modal,titlebar,resizable=yes",
                      obj);

    if (doNotAskPref && obj.doNotAsk) {
        switch (selectionCondition) {
        case "one-or-more":
        case "zero-or-more":
            if (obj.selected.length == obj.items.length) {
                _dialogs_gPrefs.setBooleanPref(bpref, true);
                _dialogs_gPrefs.setStringPref(spref, "All");
            } else if (obj.selected.length == 0) {
                _dialogs_gPrefs.setBooleanPref(bpref, true);
                _dialogs_gPrefs.setStringPref(spref, "None");
            } else {
                dialogsLog.error("unexpected selected list, it was not all "+
                                 "or none of the items as is required to "+
                                 "set the doNotAsk prefs");
            }
            break;
        case "one":
            _dialogs_gPrefs.setBooleanPref(bpref, true);
            var selectedId = null;
            if (typeof(obj.selected[0]) == 'string') {
                selectedId = obj.selected[0];
            } else {
                selectedId = obj.selected[0].id;
            }
            _dialogs_gPrefs.setStringPref(spref, selectedId);
            break;
        default:
            throw("illegal selection condition: '"+selectionCondition+"'");
        }
    }
    if (obj.retval != "Cancel") {
        return obj.selected;
    } else {
        return null;
    }
}



// A dialog to query the user for a string in a textbox.
//
// All arguments can be left blank or specified as null to get a default value.
//  "name" is a default env. var. name.
//  "value" is a default env. var. value.
//  "title" is the dialog title.
//  "mruName" can be specified (a string) to get an MRU on the text box. The
//      value of the string is the namespace for the MRU. If not specified it
//      defaults to "dialog_editEnvVar". (Why not have an env var MRU be
//      default?)
//  "interpolateValues" boolean to uncollapse the interpolation shortcut menu
//
// It returns null if the dialog was cancelled. Otherwise it returns an
// object with the following attributes:
//  .name       the name entered by the user
//  .value      the value entered by the user
//
// Dev Note: a "validator" argument, a la dialog_prompt(), could easily be
// added if useful.
function dialog_editEnvVar(name, value, title, mruName /* dialog_editEnvVar */,
                           interpolateValues)
{
    if (typeof(name) == 'undefined') name = null;
    if (typeof(value) == 'undefined') value = null;
    if (typeof(title) == 'undefined') title = null;
    if (typeof(mruName) == 'undefined' || mruName == null) mruName = "dialog_editEnvVar";
    if (typeof(interpolateValues) == 'undefined') interpolateValues = false;

    var obj = new Object();
    obj.name = name;
    obj.value = value;
    obj.title = title;
    obj.mruName = mruName;
    obj.interpolateValues = interpolateValues;
    window.openDialog("chrome://jaguar/content/dialogs/editEnvVar.xul",
                      "_blank",
                      "chrome,modal,titlebar,resizable=yes",
                      obj);
    if (obj.retval == "OK") {
        return obj;
    } else {
        return null;
    }
}


// Display an internal error (and encourage the user to report it).
//
//  "error" is a short description of the error.
//  "text" is some text the user can quote in their bug report. This should
//      include the details that you would like in a bug report.
//  "bnamespace" (optional) is a namespace that supports
//      browse_OpenUrlInDefaultBrowser(). Defaults to the current namespace.
//
// This function does not return anything.
//
function dialog_internalError(error, text, bnamespace)
{
    if (typeof(error) == 'undefined' || error == null)
        throw("Must specify 'error' argument to dialog_internalError().");
    if (typeof(text) == 'undefined' || text == null)
        throw("Must specify 'text' argument to dialog_internalError().");
    if (typeof(bnamespace) == 'undefined')
        bnamespace = null;

    // Show the dialog.
    var obj = new Object();
    obj.error = error;
    obj.text = text;
    if (bnamespace != null) {
        obj.bnamespace = bnamespace;
    }
    window.openDialog("chrome://jaguar/content/dialogs/internalError.xul",
                      "_blank",
                      "chrome,modal,titlebar,dialog=no",
                      obj);
}



// This is commonly used in dialogs that don't use the <dialog> XBL (for
// whatever reason) like so:
//    <keyset id="<some id>">
//        <key keycode="VK_ESCAPE" modifiers="" oncommand="window.close();"/>
//        <key keycode="VK_ENTER" oncommand="dialogs_handleEnterKey();"/>
//        <key keycode="VK_RETURN" oncommand="dialogs_handleEnterKey();"/>
//    </keyset>
// Preferably, each such dialgo should be converted to use the <dialog> XBL,
// but that is not always practical.
//
function dialogs_handleEnterKey() {
    // 1) If button has focus, execute 'oncommand' handler
    // 2) If no button is focused, do default button (if not disabled).
    // 3) If no button is focused, and the default button is disabled, do nothing.
    // NOTE: This presumes that if a button has something to do it *is*
    // its oncommand handler (i.e. 'onclick' is no good)
    var element = document.commandDispatcher.focusedElement;
    var command = null;
    if (element && element.nodeName == "button") {
        // a button has the focus:
        // - Don't do anything, the button's oncommand handler will get
        //   called.
    } else {
        // no button has the focus:
        // - do default button's oncommand if there is one
        var buttons = document.getElementsByTagName("button");
        for (var i=0; i<buttons.length; i++) {
            var button = buttons[i];
            if (!button.disabled
                && button.hasAttribute("default")
                && buttons[i].getAttribute("default"))
            {
                command = buttons[i].getAttribute('oncommand');
            }
        }
    }
    if (command) {
        eval(command);
    }
}



// The following window positioning methods should all die. Any dialogs using
// them should consider being refactored as <dialog/>'s which can then use the
// built in CenterWindowOnScreen() and MoveToAlertPosition().

function koCenterWindowOnScreen() {
    dialogsLog.error('koCenterWindowOnScreen DEPRECATED');
    dialog_centerWindowOnScreen();
}

function dialog_centerWindowOnScreen() {
  window.sizeToContent();
  var w = screen.availWidth;
  var h = screen.availHeight;
  var popW = window.outerWidth;
  var popH = window.outerHeight;
  var topPos = (w-popW)/2;
  var leftPos = (h-popH) * (1.0-0.62);
  window.moveTo(topPos, leftPos);
}

function SetWindowLocation()
{
   var gLocation = document.getElementById("location");
   if (gLocation)
   {
     window.screenX = Math.max(0, Math.min(window.opener.screenX + Number(gLocation.getAttribute("offsetX")),
                                           screen.availWidth - window.outerWidth));
     window.screenY = Math.max(0, Math.min(window.opener.screenY + Number(gLocation.getAttribute("offsetY")),
                                           screen.availHeight - window.outerHeight));
  }
}

function SaveWindowLocation()
{
   var gLocation = document.getElementById("location");
   if (gLocation)
   {
     var newOffsetX = window.screenX - window.opener.screenX;
     var newOffsetY = window.screenY - window.opener.screenY;
     gLocation.setAttribute("offsetX", window.screenX - window.opener.screenX);
     gLocation.setAttribute("offsetY", window.screenY - window.opener.screenY);
   }
}


function dialog_pickIcon()
{
    var obj = new Object();
    window.openDialog("chrome://jaguar/content/dialogs/iconpicker.xul",
                      "_blank",
                      "chrome,dialog=no,modal,titlebar,resizable=yes",
                      obj);
    if (obj.retval == "OK") {
        return obj.value;
    } else {
        return null;
    }
}

