/* Copyright (c) 2000-2002 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// The New Plan:
//    - This module is reponsible for MRUs (examples: recent files, recent
//      projects, recent find/replace patterns).
//    - It has the smarts to tie an MRU to a koIOrderedPreference so that
//      it is serialized.
//    - It has the smarts to observe changes in relevant prefs which affect
//      the MRUs, e.g. mruProjectSize.
//      XXX IMO, This is the wrong place to do this. Ideally there would be a
//          koIMRU interface with separate implementations for the various
//          MRUs (and possibly subclassed interfaces for special behaviour).
//    - As required, other modules will call this module's MRU_* methods
//      to note changes in particular MRUs.
//    - This module will send out a "mru_changed" notification whenever an
//      MRU changes. It is up to the various UIs (e.g. File->Recent Files,
//      Getting Started) to observe this notification and act accordingly.
//

//---- globals

try {
    var gMRUPrefObserver = null;
    var mrulog = getLoggingMgr().getLogger('mru');
    //mrulog.setLevel(LOG_DEBUG);
} catch (e) {
    alert(e);
}

//---- internal support routines

function _notifyOfMRUChange(prefName)
{
    try {
        var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                    getService(Components.interfaces.nsIObserverService);
        observerSvc.notifyObservers(null, 'mru_changed', prefName);
    } catch(e) {
        // Raises exception if no observers are registered. Just ignore that.
    }
}


function _MRUPrefObserver()
{
    this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    this.prefSvc.prefs.addObserver(this);
};
_MRUPrefObserver.prototype.destroy = function()
{
    this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    this.prefSvc.prefs.removeObserver(this);
    this.prefSvc = null;
}
_MRUPrefObserver.prototype.observe = function(prefSet, prefSetID, prefName)
{
    // Adjust the size of MRUs if the size preference for that MRU changes.
    // XXX Shouldn't really be hardcoding specific MRU names here. This module
    //     is _supposed_ to be generalized.
    if (prefName != "mruProjectSize" && prefName != "mruFileSize"
        && prefName != "mruTemplateSize") {
        return;
    }

    mrulog.info("observed pref '" + prefName + "' change (prefSet=" +
                prefSet + ", prefSetID=" + prefSetID + ")");
    var listPrefName = null;
    if (prefName == "mruProjectSize") {
        listPrefName = "mruProjectList";
    } else if (prefName == "mruFileSize") {
        listPrefName = "mruFileList";
    } else if (prefName == "mruTemplateSize") {
        listPrefName = "mruTemplateList";
    }
    var mruList = null;
    if (this.prefSvc.prefs.hasPref(listPrefName)) {
        mruList = this.prefSvc.prefs.getPref(listPrefName);
    }
    if (mruList) {
        // Chop the list to the correct size.
        var maxEntries = MRU_maxEntries(listPrefName);
        while (mruList.length > maxEntries) {
            mruList.deletePref(mruList.length - 1);
        }
        _notifyOfMRUChange(listPrefName);
    }
};


//---- public methods

function MRU_initialize()
{
    mrulog.info("MRU_initialize()");
    gMRUPrefObserver = new _MRUPrefObserver();
    KomodoRegisterOnUnloadHandler(MRU_finalize);
}


function MRU_finalize()
{
    mrulog.info("MRU_finalize()");
    gMRUPrefObserver.destroy();
    gMRUPrefObserver = null;
}


function MRU_maxEntries(prefName)
{
    mrulog.info("MRU_maxEntries(prefName="+prefName+")");
    // XXX:HACK This should be an attribute on difference instances of a
    //          koIMRU interface.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    var maxEntries;
    if (prefName == 'mruFileList')  {
        maxEntries = prefSvc.prefs.getLongPref('mruFileSize');
    } else if (prefName == 'mruProjectList')  {
        maxEntries = prefSvc.prefs.getLongPref('mruProjectSize');
    } else if (prefName == 'mruTemplateList')  {
        maxEntries = prefSvc.prefs.getLongPref('mruTemplateSize');
    } else {
        maxEntries = 50; // Bigger?
    }
    return maxEntries;
}


function MRU_add(prefName, entry, caseSensitive)
{
    mrulog.info("MRU_add(prefName="+prefName+", entry="+entry+
                ", caseSensitive="+caseSensitive+")");

    // Add the given "entry" (a string) to the given MRU (indentified by
    // the name of pref, "prefName", with which it is associated).
    // "caseSensitive" is a boolean indicating how to match "entry" with
    // existing MRU entries.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);

    // Validate arguments.
    var errmsg;
    if (!prefName) {
        errmsg = "MRU_add: invalid argument: prefName='"+prefName+"'";
        mrulog.error(errmsg);
        throw(errmsg);
    }
    if (!entry) {
        errmsg = "MRU_add: warning: no entry: prefName='"+prefName+
                 "', entry='"+entry+"'";
        mrulog.warn(errmsg)
    }

    var mruList = null;
    if (prefSvc.prefs.hasPref(prefName)) {
        mruList = prefSvc.prefs.getPref(prefName);
    } else {
        // Create a preference-set with this name.
        mruList = Components.classes["@activestate.com/koOrderedPreference;1"]
                  .createInstance();
        mruList.id = prefName;
        // Add the MRU list to the global preference set.
        prefSvc.prefs.setPref(prefName, mruList);
    }

    // If the mru list already contains this entry, first remove it so
    // that it will be reinserted at the top of the list again.
    for (var i = 0; i < mruList.length; i++) {
        var existingEntry = mruList.getStringPref(i);
        if (caseSensitive && entry == existingEntry) {
            mruList.deletePref(i);
            break;
        } else if (entry.toLowerCase() == existingEntry.toLowerCase()) {
            mruList.deletePref(i);
            break;
        }
    }

    // Also: keep the list constrained to the correct size.
    var maxEntries = MRU_maxEntries(prefName);
    while (mruList.length >= maxEntries && mruList.length > 0) {
        mruList.deletePref(mruList.length - 1);
    }

    // "Push" the entry onto the mru "stack".
    if (maxEntries > 0) {
        mruList.insertStringPref(0, entry);
    }

    _notifyOfMRUChange(prefName);
}


function MRU_addURL(prefName, url)
{
    mrulog.info("MRU_addURL(prefName="+prefName+", url="+url+")");

    // Never want to add kodebugger:// URLs to our MRUs.
    // XXX:HACK This should be a special case of a sub-classed implementation
    //          of koIMRU for the recent files and projects MRUs.
    if (url.indexOf("kodebugger://") == 0) {
        return;
    }

    //XXX Might want to cache or preprocess this.
    var infoService = Components.classes['@activestate.com/koInfoService;1'].
                      getService(Components.interfaces.koIInfoService);
    // On *nix filesystems, files with same name different case are recognised
    // as a separate files they are the same file on win32.
    var caseSensitive = infoService.platform.substring(0,3) != "win";

    // max entries should itself come from a preference!
    MRU_add(prefName, url, caseSensitive);
}


function MRU_addFromACTextbox(widget)
{
    var searchType = widget.getAttribute("autocompletesearch");
    var searchParam = widget.searchParam;
    var prefName;
    switch (searchType) {
    case "mru":
        prefName = searchParam;
        break
    case "mru_and_filepath":
    case "mru_and_dirpath":
        prefName = stringutils_getSubAttr(searchParam, "mru");
        break
    default:
        errmsg = "MRU_addFromACTextbox: invalid argument: autocompletesearch='"+searchType+"'";
        mrulog.error(errmsg);
        throw(errmsg);
    }
    if (!prefName) {
        errmsg = "MRU_addFromACTextbox: warning: could not determine "+
                 "pref name from search param: '"+searchParam+"'";
        mrulog.warn(errmsg);
        return;
    }

    mrulog.info("MRU_addFromACTextbox(widget): widget.value='"+widget.value+
                "', prefName="+prefName);
    MRU_add(prefName, widget.value, true);
}


function MRU_get(prefName, index /* =0 */)
{
    if (typeof(index) == "undefined" || index == null) index = 0;
    mrulog.info("MRU_get(prefName="+prefName+", index="+index+")");

    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    var retval = null;
    if (!prefSvc.prefs.hasPref(prefName)) {
        retval = "";
    } else {
        var mruList = prefSvc.prefs.getPref(prefName);
        if (index >= mruList.length) {
            retval = "";
        } else {
            retval = mruList.getStringPref(index);
        }
    }
    mrulog.info("Returning: " + retval);
    return retval;
}

function MRU_del(prefName, index)
{
    mrulog.info("MRU_del(prefName="+prefName+", index="+index+")");
    // Remove the identified entry from the MRU list.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    if (prefSvc.prefs.hasPref(prefName)) {
        var mruList = prefSvc.prefs.getPref(prefName);
        mruList.deletePref(index);
        _notifyOfMRUChange(prefName);
    }
}

function MRU_reset(prefName)
{
    mrulog.info("MRU_reset(prefName="+prefName+")");
    // Reset (a.k.a remove all elements) from the given MRU list.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    if (prefSvc.prefs.hasPref(prefName)) {
        var mruList = prefSvc.prefs.getPref(prefName);
        mruList.reset();
        _notifyOfMRUChange(prefName);
    }
}



