#!python
# Copyright (c) 2004 Sophos Ltd.
# See the file LICENSE.txt for licensing information.

"""The glue between the Komodo-independent 'codeintel' Python package
and Komodo's Code Intelligence functionality.
"""
#TODO:
# - get PySQLite into our silo'd Python and make it easy for dev builders
#   on all platforms
# - what mechanism to use to change code browser fileset?
# - What about cleaning out info from the DB for deleted files?
#   Perhaps that can be done lazily when looking up symbols? I.e.
#   when about to report a symbol that comes from a file, check
#   that that file still exists. That might be too slow. Perhaps
#   the logic could be: check that the file still exists if we
#   haven't done so in the last set amount of time (minute? hour?
#   day? week?). Would probably also want some background process
#   to keep up-to-date. Ideally this would, at Komodo runtime,
#   involve a file-system watcher.
#

import os
import sys
import string
import re
import threading
import logging
import types
import time
import pprint

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
from xpcom.server import UnwrapObject
from koTreeView import TreeView
import uriparse

import codeintel
from codeintel.database import Database
from codeintel.common import parseAttributes, canonicalizePath,\
                             isUnsavedPath, CodeIntelError, symbolType2Name,\
                             ST_FUNCTION, ST_CLASS, ST_INTERFACE,\
                             ST_VARIABLE, ST_ARGUMENT
from codeintel.citdl import S_ID, S_FILE_ID, S_SCOPE, S_NAME, S_LINE, \
                            S_LINEEND, S_TYPE, S_ATTRIBUTES, \
                            M_ID, M_FILE_ID, M_LINE, M_NAME
import cb
import apsw


#---- globals

log = logging.getLogger("koCodeIntel")
#log.setLevel(logging.INFO)

descriptionCSS = """
body {
    font-size: 70%;
    margin:0px;
}
body, p, h1, h2, h3, table, td, th, ul, ol, textarea, input {
    font-family: verdana,helvetica,arial,sans-serif;
}"""



#---- component implementations

class KoCodeBrowserTreeView(TreeView):
    _com_interfaces_ = [components.interfaces.koICodeBrowserTreeView,
                        components.interfaces.nsITreeView]
    _reg_clsid_ = "{E9E7E7EB-2DED-479B-AA1C-D8AF0E1347EA}"
    _reg_contractid_ = "@activestate.com/koCodeBrowserTreeView;1"
    _reg_desc_ = "Komodo Code Browser Tree View"

    def __init__(self):
        TreeView.__init__(self) #, debug="koCodeBrowserTreeView")
        self._tree = None
        self._sortedBy = "name"
        self._root = None
        self._rows = []
        self._manager = None
        self._dataLock = threading.RLock()

        self._filterString = ''
        self._wsFiles = { # the working set of files
            # <file>: <language>
        }
        # Mapping of node "scoped-name" to a boolean indicating if that
        # node is open (or should be) in the Code Browser.
        self._nodeIsOpen = {}
        self._ignoreNextToggleOpenState = 0
        # Data for controlling the display of node detail when the mouse is
        # hovering over that row's icon.
        self._showDetailForRow = -1
        atomSvc = components.classes["@mozilla.org/atom-service;1"].\
                  getService(components.interfaces.nsIAtomService)
        self._showDetailAtom = atomSvc.getAtom("showDetail")

    def restorePrefs(self):
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                            .getService().prefs # global prefs
        self._nodeIsOpen = {}
        if prefSvc.hasStringPref("codeintel_open_codebrowser_nodes"):
            nioStr = prefSvc.getStringPref("codeintel_open_codebrowser_nodes")
            try:
                self._nodeIsOpen = eval(nioStr)
            except SyntaxError, ex:
                pass

    def savePrefs(self):
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                            .getService().prefs # global prefs
        prefSvc.setStringPref("codeintel_open_codebrowser_nodes",
                              repr(self._nodeIsOpen))

    def activate(self, manager, codeBrowserMgr):
        self._manager = manager
        self._codeBrowserMgr = codeBrowserMgr
        self._generateTree()

    def sortBy(self, key):
        changed = 0

        self._dataLock.acquire()
        try:
            if self._sortedBy != key:
                self._sortedBy = key
                if self._root:
                    rows = []
                    self._root.generateRows(rows, self._nodeIsOpen,
                                            self._sortedBy, 
                                            filterString=self._filterString)
                    self._rows = rows
                    changed = 1
        finally:
            self._dataLock.release()

        if self._tree and changed:
            self._tree.beginUpdateBatch()
            self._tree.invalidate()
            self._tree.endUpdateBatch()

    def setFilter(self, filterString):
        self._filterString = filterString
        self._dataLock.acquire()
        try:
            if self._root:
                rows = []
                self._root.generateRows(rows, self._nodeIsOpen,
                                        self._sortedBy, 
                                        filterString=self._filterString)
                self._rows = rows
        finally:
            self._dataLock.release()

        if self._tree:
            self._tree.beginUpdateBatch()
            self._tree.invalidate()
            self._tree.endUpdateBatch()

            self._tree.view.selection.select(self._tree.view.selection.currentIndex)




    def locateScope(self, scimoz, position):
        #XXX If we stop using getAdjustedCurrentScope() and instead get the
        #    scope from path/line the CodeIntelCompletionUIHandler will have
        #    to be updated to use the proper CI-path so that CIDB lookups
        #    work.
        file_id, type, id, row = self._manager.getAdjustedCurrentScope(scimoz, position)
        if type is None:
            return 0

        # Algorithm: walk through the node hierarchy to find the identified
        # node. Set nodeIsOpen to true for all its parents. Re-generate the
        # rows and set that row as the current one.
        self._filterString = None  # Don't want filtering to get in the way.
        locatedAt = None
        self._dataLock.acquire()
        try:
            if self._root:
                rows = []
                locatedAt = self._root.generateRows(
                    rows, self._nodeIsOpen, self._sortedBy,
                    nodeToLocate=(type,id))
                if locatedAt is not None:
                    self._rows = rows
        finally:
            self._dataLock.release()

        if self._tree and locatedAt is not None:
            self._tree.beginUpdateBatch()
            self._tree.invalidate()
            self._tree.endUpdateBatch()

            self._tree.view.selection.select(locatedAt)



            self._tree.ensureRowIsVisible(locatedAt)

        return locatedAt is not None

    def wsDump(self):
        if 0: #XXX Turn this off for now.
            print "-"*50, "Code Browser working set:"
            pprint.pprint(self._wsFiles)
            print "-"*70

    def wsAddFile(self, filename, language):
        log.debug("KoCodeBrowserTreeView.wsAddFile('%s', '%s')", filename,
                  language)
        self._dataLock.acquire()
        try:
            cpath = canonicalizePath(filename)
            if cpath in self._wsFiles:
                # This is possible because there is a (not serious) race
                # condition for "opened_document" ideEvents on startup.
                return 
            assert cpath not in self._wsFiles
            if not self._wsFiles:
                addedFirstFile = 1
            else:
                addedFirstFile = 0
            self._wsFiles[cpath] = language
            if log.isEnabledFor(logging.DEBUG):
                self.wsDump()
            self._updateTree("add", cpath, language)
            if addedFirstFile:
                self._codeBrowserMgr.haveFilesInWS(1)
        finally:
            self._dataLock.release()

    def wsAddFiles(self, fileinfo):
        log.debug("KoCodeBrowserTreeView.wsAddFiles(%r)", fileinfo)
        for filename, language in fileinfo:
            self.wsAddFile(filename, language)

    def wsRemoveFile(self, filename):
        log.debug("KoCodeBrowserTreeView.wsRemoveFile('%s')", filename)
        self._dataLock.acquire()
        try:
            cpath = canonicalizePath(filename)
            if cpath not in self._wsFiles:
                # Don't strictly require a removed filename to have been in
                # the working set. This allows the caller to not bother
                # filtering calls to this method based on whether its language
                # is supported. Such filtering _is_ done for adding files to
                # the working set, though.
                return
            del self._wsFiles[cpath]
            if log.isEnabledFor(logging.DEBUG):
                self.wsDump()
            self._updateTree("remove", cpath)
            if not self._wsFiles:
                self._codeBrowserMgr.haveFilesInWS(0)
        finally:
            self._dataLock.release()
    def wsRemoveFiles(self, filenames):
        log.debug("KoCodeBrowserTreeView.wsRemoveFiles(%r)", filenames)
        for filename in filenames:
            self.wsRemoveFile(filename)

    def updateFile(self, filename, language):
        cpath = canonicalizePath(filename)
        self._updateTree("update", cpath, language)
        if cpath == self.get_currentNodeFilePath():
            self._codeBrowserMgr.updateDescription()

    def _updateTree(self, action, path, language=None):
        """Update the Code Browser data for a relevant change.

            "action" is one of 'update', 'add', or 'remove'.
            "path" is the full path to the file.
            "language" is the file's language, or None if this is not
                available or relevant for the given action.

        The idea is to try to be more efficient than just re-generating
        the whole tree for any change. As well, invalidating only <tree/>
        rows whose data has actually changed can enable cleaner UI
        refreshing.
        """
        self._dataLock.acquire()
        try:
            cx = self._manager.getDatabaseConnection()
            cu = cx.cursor()
            self._manager.acquireCIDBReadLock()
            try:
                try:
                    func = {"update": self._root.updatePath,
                            "add": self._root.addPath,
                            "remove": self._root.removePath}[action]
                    func(cu, path, language)
    
                    #PERF: If invalidating the whole tree becomes a
                    #      performance problem then look into limiting
                    #      the invalidation to only those rows that have
                    #      actually changed.
                    oldRowCount = len(self._rows)
                    rows = []
                    self._root.generateRows(rows, self._nodeIsOpen,
                                            self._sortedBy,
                                            filterString=self._filterString,
                                            cx=cx, cu=cu)
                    self._rows = rows
                    if self._tree:
                        self._tree.beginUpdateBatch()
                        self._tree.rowCountChanged(0, len(self._rows)-oldRowCount)
                        self._tree.invalidate()
                        self._tree.endUpdateBatch()
                except cb.CBError, ex:
                    #XXX logging exception for now to try to debug this
                    log.exception("problem updating path '%s' in Code "
                                  "Browser: action '%s': %s", path,
                                  action, ex)
                    # ignore this error for now
            finally:
                self._manager.releaseCIDBReadLock()
                cu.finalize()
                cx.close()

            if log.isEnabledFor(logging.DEBUG):
                self._root.pprint()
        finally:
            self._dataLock.release()

    def _generateTree(self):
        """Generate the code browsing AST and extract the display rows."""
        self._dataLock.acquire()
        log.debug("-------------- KoCodeBrowserTreeView._generateTree() ---------------")
        try:
            oldRowCount = len(self._rows)
            root = cb.CBRootNode()
            rows = []

            if not self._wsFiles:
                # Abort early if don't have any files in the working set.
                rows = []
            else:
                cx = self._manager.getDatabaseConnection()
                cu = cx.cursor()
                self._manager.acquireCIDBReadLock()
                try:
                    # Get a set of modules in all files in the working set.
                    for path, language in paths.copy().items():
                        try:
                            root.addPath(cu, path, language)
                        except cb.CBError, ex:
                            log.error("problem adding path '%s' to Code "
                                      "Browser: %s", ex)
                            continue # drop this error for now

                    self._root = root

                    rows = []
                    root.generateRows(rows, self._nodeIsOpen, self._sortedBy,
                                      filterString=self._filterString,
                                      cx=cx, cu=cu)
                finally:
                    self._manager.releaseCIDBReadLock()
                    cu.finalize()
                    cx.close()

            self._root = root
            if log.isEnabledFor(logging.DEBUG):
                self._root.pprint()
            self._rows = rows

            self._tree.beginUpdateBatch()
            self._tree.rowCountChanged(0, len(self._rows)-oldRowCount)
            self._tree.invalidate()
            self._tree.endUpdateBatch()
        finally:
            log.debug("-"*70)
            self._dataLock.release()

    def ignoreNextToggleOpenState(self):
        self._ignoreNextToggleOpenState = 1

    def mouseOverNode(self, row, element):
        invalids = [] # rows that need redrawing
        if element == "image" and self._showDetailForRow != row:
            # The mouse is now over a new node's image: show details for
            # that node.
            invalids = [i for i in (self._showDetailForRow, row)
                        if i != -1]
            self._showDetailForRow = row
        elif (self._showDetailForRow != -1 and element != "image"
              and self._showDetailForRow != row):
            # The mouse is over a row for which we are not showing detail:
            # we should remove any special detail UI for other rows.
            invalids = [self._showDetailForRow]
            self._showDetailForRow = -1

        if invalids:
            self._tree.beginUpdateBatch()
            for invalid in invalids:

                column = self._tree.columns.getNamedColumn("codebrowser-tree-node")



                self._tree.invalidateCell(invalid, column)
            self._tree.endUpdateBatch()

    #---- current node accessor attributes
    def _getCurrentNode(self):

        index = self._tree.view.selection.currentIndex



        if index == -1 or index >= len(self._rows):
            return None
        return self._rows[index]['node']

    def get_currentNodeType(self):
        node = self._getCurrentNode()
        if not node: return None
        return node.type

    def get_currentNodeFilePath(self):
        node = self._getCurrentNode()
        if not node: return None
        if node.type in ("file", "pseudo-module"):
            return node.cpath
        else:
            return node.file_path

    def get_currentNodeLine(self):
        node = self._getCurrentNode()
        if not node or node.line is None: return 0
        return node.line

    def get_currentNodeLanguage(self):
        node = self._getCurrentNode()
        if not node: return None
        return node.file_language

    def get_currentNodeSymbolPattern(self):
        node = self._getCurrentNode()
        if not node: return None
        if node.type == "import":
            if node.symbol in ("*", "**"): # special signifiers
                symbol = None
            else:
                symbol = node.symbol
        else:
            symbol = node.name
        return symbol

    def get_currentNodeModulePattern(self):
        node = self._getCurrentNode()
        if not node: return None
        if node.type == "import":
            return node.module
        else:
            return None

    #---- nsITreeView methods
    def get_rowCount(self):
        try:
            return len(self._rows)
        except TypeError: # self._rows is None
            return 0

    def getCellText(self, row, column):

        col = column.id



        cell = ""
        try:
            if (col == "codebrowser-tree-node"
                and row == self._showDetailForRow
                and "detail" in self._rows[row]):
                cell = self._rows[row]["detail"]
            else:
                cell = self._rows[row][col]
        except IndexError:
            log.error("no row %d" % row)
        except KeyError:
            log.error("unknown column: '%s'" % col)
        if type(cell) not in types.StringTypes:
            cell = str(cell)
        return cell

    def getCellProperties(self, row, column, properties):
        #assert col == "codebrowser-tree-node"
        if row == self._showDetailForRow:
            properties.AppendElement(self._showDetailAtom)

    def isContainer(self, row):
        try:
            node = self._rows[row]["node"]
            if node.type == 'module' and not node.linkedUp:
                return 1
            return len(node.children)
        except IndexError, ex:
            log.error("error checking if row %d is a container: %s", row, ex)
            if log.isEnabledFor(logging.ERROR):
                log.error("--------------- current rows -------------------")
                pprint.pprint(self._rows)
                log.error("--------------- end of rows --------------------")
                import traceback
                traceback.print_exception(*sys.exc_info())
            return 0
    def isContainerEmpty(self, row):
        return 0
    def isContainerOpen(self, row):
        node = self._rows[row]['node']
        if self._rows[row]["node-is-open"] and \
           node.type == 'module' and \
           not node.linkedUp:
            node.linkUp()
            self._dataLock.acquire()
            try:
                # Must recalculate the rows.
                oldRowCount = len(self._rows)
                self._rows = []
                rows = []
                self._root.generateRows(rows, self._nodeIsOpen, self._sortedBy,
                                        filterString=self._filterString)
                self._rows = rows
            finally:
                self._dataLock.release()

            self._tree.beginUpdateBatch()
            self._tree.rowCountChanged(0, len(self._rows)-oldRowCount)
            self._tree.invalidate()
            self._tree.endUpdateBatch()
        return self._rows[row]["node-is-open"]

    def toggleOpenState(self, row):
        node = self._rows[row]['node']
        if node.type == 'module' and not node.linkedUp:
            node.linkUp()

        if self._ignoreNextToggleOpenState:
            log.debug("ignoring this toggleOpenState(row=%r)", row)
            self._ignoreNextToggleOpenState = 0
            return

        self._dataLock.acquire()
        try:
            scopedName = self._rows[row]["scoped-name"]
            isOpen = self.isContainerOpen(row)
            self._nodeIsOpen[scopedName] = not isOpen

            # Must recalculate the rows.
            oldRowCount = len(self._rows)
            self._rows = []
            rows = []
            self._root.generateRows(rows, self._nodeIsOpen, self._sortedBy,
                                    filterString=self._filterString)
            self._rows = rows
        finally:
            self._dataLock.release()

        self._tree.beginUpdateBatch()
        self._tree.rowCountChanged(0, len(self._rows)-oldRowCount)
        self._tree.invalidate()
        self._tree.endUpdateBatch()
        self.selection.currentIndex = row
        self.selection.select(row);

    def getParentIndex(self, row):
        return self._rows[row]["parent-index"]

    def getLevel(self, row):
        return self._rows[row]["level"]

    def getImageSrc(self, row, column):

        col = column.id



        if col != "codebrowser-tree-node":
            image = ""
        else:
            image = self._rows[row]["node"].imageURL
        return image

    def hasNextSibling(self, rowIndex, afterIndex):
        """From the nsITreeView.idl docs:

        HasNextSibling is used to determine if the row at rowIndex has a
        nextSibling that occurs *after* the index specified by
        afterIndex.  Code that is forced to march down the view looking
        at levels can optimize the march by starting at afterIndex+1.
        """
        nextSiblingIndex = self._rows[rowIndex].get("next-sibling-index", -1)
        return nextSiblingIndex > afterIndex

    def fillDescription(self, index, document, description):
        if index < 0 or index >= len(self._rows): return
        self._rows[index]['node'].fillDescription(document, description)

    #XXX handle canDropOn(), canDropBeforeAfter(), drop()
    #XXX What about the performAction[on{Row|Cell}]() methods?

    if 0:
        # Turn this on to silence TreeView logging, if enabled,
        # for these specific methods.
        def isSorted(self):
            return 0
        def getRowProperties(self, row, properties):
            pass
        def getColumnProperties(self, 
                            column,



                            properties):
            pass


class _Manager(codeintel.Manager):
    """Subclass the Manager class:
    - to notify relevant parts of the Komodo UI when a certain scan requests
      complete
    - to add smarts to determine the current scope more efficiently
      (hopefully) -- by caching CIDB data on the current file -- and more
      correctly -- given recent edits and language-specific smarts.
    """
    def __init__(self, registry=None, coreLanguages=[]):
        codeintel.Manager.__init__(self, registry, coreLanguages)
        self._codeBrowserTree = None
        obsSvc = components.classes["@mozilla.org/observer-service;1"]\
                 .getService(components.interfaces.nsIObserverService)
        self._proxiedObsSvc = getProxyForObject(None,
            components.interfaces.nsIObserverService,
            obsSvc, PROXY_ALWAYS | PROXY_SYNC)
        self._phpInfo = components.classes["@activestate.com/koPHPInfoInstance;1"]\
                            .getService(components.interfaces.koIPHPInfoEx)
        self._csLock = threading.RLock()
        self._currFileName = None
        self._currLanguage = None
        self._flushCSCache()
        self._batchUpdateProgressUIHandler = None

    def activateUI(self, codeBrowserTree):
        self._codeBrowserTree = getProxyForObject(None,
            components.interfaces.koICodeBrowserTreeView,
            codeBrowserTree, PROXY_ALWAYS | PROXY_SYNC)

    def requestCompleted(self, request):
        if self._codeBrowserTree:
            self._codeBrowserTree.updateFile(request.path, request.language)
        if request.path == self._currFileName:
            self._flushCSCache()
            try:
                self._proxiedObsSvc.notifyObservers(
                    None, "current_view_codeintel_scan_completed", None)
            except COMException, ex:
                pass

    def getImportHandler(self, language):
        importer = codeintel.Manager.getImportHandler(self, language)
        if language == "Python":
            # Set the "environment path" using the _user's_ Python
            # environment settings, because Komodo messes with that
            # environment.
            import koprocessutils
            userenv = koprocessutils.getUserEnv()
            PYTHONPATH = userenv.get(importer.PATH_ENV_VAR, "")
            importer.setEnvPath(PYTHONPATH)
        elif language == "PHP":
            # Getting the PHP include_path is quite difficult and the
            # codeintel package has not yet learned how to do it.
            # Override the PHPImportHandler's .setCorePath()
            # with one smart enough to do it.
            _phpInfo = self._phpInfo
            def _setPHPIncludePath(self, compiler=None, extra=None):
                if compiler:
                    _phpInfo.executablePath = compiler
                if extra:
                    _phpInfo.cfg_file_path = extra
                self.corePath = _phpInfo.include_path.split(os.pathsep)
            importer.__class__.setCorePath = _setPHPIncludePath

        # Add the "Additional Perl/Python/Tcl Import Directories" (in that
        # language's preferences panel) to the import path.
        lang2extraPathsPref = {
            "Python": "pythonExtraPaths",
            "Perl": "perlExtraPaths",
            "Tcl": "tclExtraPaths",
            "Ruby": "rubyExtraPaths",
        }
        if language in lang2extraPathsPref:
            prefName = lang2extraPathsPref[language]
            prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                                .getService().prefs # global prefs
            if prefSvc.hasStringPref(prefName):
                extraPaths = prefSvc.getStringPref(prefName).strip().split(os.pathsep)
                if extraPaths:
                    importer.setCustomPath(extraPaths)

        return importer

    #---- batch update progress handling
    def batchUpdateStart(self, uiHandler):
        if uiHandler:
            self._batchUpdateProgressUIHandler = getProxyForObject(None,
                components.interfaces.koICodeIntelBatchUpdateProgressUIHandler,
                uiHandler, PROXY_ALWAYS | PROXY_ASYNC)
        codeintel.Manager.batchUpdateStart(self)

    def batchUpdateProgress(self, stage, obj):
        if not self._batchUpdateProgressUIHandler:
            return

        # "Upgrade" progress is handled differently.
        #   "stage" is a string summarizing the full upgrade
        #   "obj" is a 2-tuple: (<current-stage-message>, <percentage-done>)
        if self._batchScheduler.mode == "upgrade":
            msg = "%s: %s" % (stage, obj[0])
            self._batchUpdateProgressUIHandler.setStatusMessage(msg)
            self._batchUpdateProgressUIHandler.setProgressMeterValue(obj[1])
            return

        cache = self._batchProgressCache or {}
        remaining = eta = percent = None
        percentFromImporting = cache.setdefault("percentFromImporting", 0.00)
        percentFromGathering = cache.setdefault("percentFromGathering", 0.00)
        percentFromScanning = cache.setdefault("percentFromScanning", 0.00)
        if stage == "Importing":
            msg = "Importing: %s" % obj
            # just want at least 5% on first CIX file import
            percentFromImporting = 0.05
        elif stage == "Gathering files":
            remaining = self._batchScheduler.getNumFilesToProcess()+1
            msg = "Gathering files..."
            timesGatheringFiles = cache.setdefault("timesGatheringFiles", 0)
            timesGatheringFiles += 1
            percentFromGathering = min(0.01*timesGatheringFiles, 0.05)
            cache["timesGatheringFiles"] = timesGatheringFiles
        elif stage == "Scanning": # 'obj' is a request object
            remaining = self._batchScheduler.getNumFilesToProcess()+1
            total = cache.setdefault("totalFiles", remaining)
            # Progress meter
            percentFromScanning = float(total-remaining)/float(total)
            # Message
            # 79: display width; 25: other stuff, e.g. "Scanning", "ETA: ..."
            msg = "Scanning: %s" % obj.path
            # ETA
            FEW = 50
            now = time.time()
            try:
                lastFew = cache["lastFew"]
            except KeyError:
                cache["lastFew"] = [now]
            else:
                secsForOne = (now - lastFew[0]) / len(lastFew)
                lastFew.append(now)
                if len(lastFew) > FEW: del lastFew[0]
                cache["lastFew"] = lastFew

                remainSecs = secsForOne * float(remaining)
                if remainSecs < 60.0:
                    eta = "%d sec" % int(remainSecs)
                elif (remainSecs/60.0 < 60.0):
                    eta = "%d min" % int(remainSecs/60.0)
                else:
                    eta = "%d hr" % int(remainSecs/3600.0)

            # Adjust percent so it ranges over the remainder after the
            # accumulated "Gathering" and "Importing" percentages have been
            # factored out.
            suckedUp = percentFromImporting + percentFromGathering
            percentFromScanning = percentFromScanning * (1.00-suckedUp)
        else:
            if obj:
                msg = "%s: %s" % (stage, obj)
            else:
                msg = stage
        cache["percentFromImporting"] = percentFromImporting
        cache["percentFromGathering"] = percentFromGathering
        cache["percentFromScanning"] = percentFromScanning
        if percent is None:
            percent = percentFromImporting + percentFromGathering + percentFromScanning
        self._batchProgressCache = cache

        self._batchUpdateProgressUIHandler.setStatusMessage(msg)
        #if remaining is not None:
        if remaining is None: remaining = -1
        self._batchUpdateProgressUIHandler.setRemainingFiles(remaining)
        if percent is not None:
            self._batchUpdateProgressUIHandler.setProgressMeterValue(int(percent*100.0))
        #if eta is not None:
        self._batchUpdateProgressUIHandler.setETA(eta)

    def batchUpdateCompleted(self, reason):
        if not self._batchUpdateProgressUIHandler:
            return

        if reason == "completed":
            self._batchUpdateProgressUIHandler.completed()
        elif reason == "error":
            errors = self.batchUpdateGetErrors()
            try:
                self._batchUpdateProgressUIHandler.erroredout(
                    '\n'.join(errors))
            except:
                print
                import traceback
                traceback.print_exception(*sys.exc_info())
        else:
            self._batchUpdateProgressUIHandler.cancelled()
        self._batchUpdateProgressUIHandler = None

    def batchUpdateError(self, errmsg):
        pass

    #---- Current Scope (CS) smarts handling.
    def _flushCSCache(self):
        self._csLock.acquire()
        try:
            self._symbolRows = self._moduleRows = None
            self._edits = []
        finally:
            self._csLock.release()

    def _fillCSCache(self):
        cx = self.getDatabaseConnection()
        cu = cx.cursor()
        self.acquireCIDBReadLock()
        try:
            cpath = canonicalizePath(self._currFileName)
            cu.execute("SELECT id FROM file WHERE compare_path=? LIMIT 1", (cpath,))
            for row in cu:
                file_id = row[0]
                break
            else:
                self._symbolRows = []
                self._moduleRows = []
                return
            cu.execute("SELECT * FROM symbol WHERE file_id=? "
                       "AND type IN (?, ?, ?) ORDER BY line",
                       (file_id, ST_CLASS, ST_FUNCTION, ST_INTERFACE))
            self._symbolRows = tuple(cu)
            cu.execute("SELECT * FROM module WHERE file_id=%d ORDER BY line" % file_id)
            self._moduleRows = tuple(cu)
        finally:
            self.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    def setCurrentFile(self, filename, language):
        self._csLock.acquire()
        try:
            self._currLanguage = language
            if self._currFileName != filename:
                self._currFileName = filename
                self._flushCSCache()
        finally:
            self._csLock.release()

    def editedCurrentFile(self, scimoz, linesAdded):
        """Called to notify of line add/remove changes in the curr file."""
        self._csLock.acquire()
        try:
            currLine = scimoz.lineFromPosition(scimoz.currentPos)+1
            self._edits.insert(0, (currLine+linesAdded, -linesAdded))
            #XXX This could get *very* large if "enable re-scanning of the
            #    current file while editing" is disabled. How to handle?
            #    Does a scan get issued when switching away and back to the
            #    file?
        finally:
            self._csLock.release()

    def _getScopeLine(self, scimoz, position):
        currLine = scimoz.lineFromPosition(position)+1 # 1-based
        scopeLine = currLine
        if self._currLanguage == "Python":
            # The edits-tracking mechanism alone probably doesn't work
            # that well for Python because one can add content to a scope
            # starting from outside that scope's last-scanned
            # line:lineend by just tabing to the appropriate indent
            # level. Instead of applying edits to the current line we
            # look backwards for the appropriate def/class declaration
            # and start on that line.
            nChars, lineContent = scimoz.getLine(currLine-1)
            if re.search("^\s*(def|class) \w+", lineContent):
                # If on "def foo/class Foo" line of the scope, don't move up
                # to the containing scope even though, technically, this
                # is the correct scope for evaluation because:
                # - it doesn't hurt that much: the parent scope will be
                #   searched when a CITDL expression misses in the child
                # - showing the parent scope in the statusbar might
                #   confuse the user
                # - it will allow for the correct scope for one-liners, e.g:
                #       def foo(bar): return bar.spam
                pass
            else:
                indent = ""
                for ch in lineContent:
                    if ch in ' \t':
                        indent += ch
                    else:
                        break
                if ' ' in indent and '\t' in indent:
                    # Do the best we can with mixed tabs and spaces, but
                    # don't worry too much because this is poor form anyway.
                    pattern = r"^(\s{0,%d})(def|class)" % (len(indent)-1)
                elif '\t' in indent:
                    pattern = r"^(\t{0,%d})(def|class)" % (len(indent)-1)
                elif ' ' in indent:
                    pattern = r"^( {0,%d})(def|class)" % (len(indent)-1)
                else: # indent == ''
                    pattern = None
                    scopeLine = 1 # module-level scope
                if pattern:
                    import findlib
                    result = findlib.find(scimoz.text, pattern, position,
                                          patternType="regex-python",
                                          searchBackward=1)
                    if result:
                        scopeLine = scimoz.lineFromPosition(result.start)+1
                        #print "getAdjustedScope: found Python '%s' def'n on "\
                        #      "line %d" % (result.value, scopeLine)

        # The edits-tracking mechanism alone should work well (dogfooding
        # will tell) for brace-languages because one has to add a newline
        # from within the line:lineend range of an existing method to add
        # new content in the scope.
        #XXX If self._edits is long (pick some threshold) then fold those
        #    into cache. Threshold: perhaps a good value is when the
        #    number of edits exceeds half the number of scope elements.
        if len(self._edits) > 50:
            log.warn("Currently managing a large number of line edits (%d) "
                     "in 'current scope handling'. Should consider "
                     "refreshing status." % len(self._edits))
        #print "getAdjustedScope: apply edits: %r" % self._edits
        for line, linesAdded in self._edits:
            if line <= scopeLine:
                scopeLine += linesAdded
        #print "getAdjustedScope: scope line %d -> %d" % (currLine, scopeLine)
        return scopeLine

    def _getScopeFromCache(self, scopeLine):
        symbolRow = None
        for row in self._symbolRows:
            if (row[S_LINE] <= scopeLine and
                (row[S_LINEEND] is None or scopeLine <= row[S_LINEEND])):
                symbolRow = row
            elif row[S_LINE] > scopeLine:
                break # we've already gone past the given "scopeLine"
        if symbolRow:
            return (symbolRow[S_FILE_ID], "symbol", symbolRow[S_ID], symbolRow)

        if not self._moduleRows:
            raise CodeIntelError("there are no CIDB module entries for "
                                 "file '%s'" % self._currFileName)
        moduleRow = None
        for row in self._moduleRows:
            if row[M_LINE] <= scopeLine:
                moduleRow = row
            elif row[M_LINE] > scopeLine:
                break # we've gone past the given "scopeLine"
        if not moduleRow:
            raise CodeIntelError("unexpectedly, no module rows for file "
                                 "'%s' correspond to scopeLine %d"
                                 % (self._currFileName, scopeLine))
        return (moduleRow[M_FILE_ID], "module", moduleRow[M_ID], moduleRow)

    def getAdjustedCurrentScope(self, scimoz, position):
        """A getScopeForFileAndLine() adjusted for recent edits."""
        self._csLock.acquire()
        try:
            scopeLine = self._getScopeLine(scimoz, position)
            try:
                if self._moduleRows is None: self._fillCSCache()
                retval = self._getScopeFromCache(scopeLine)
            except (apsw.Error, CodeIntelError, ValueError), ex:
                # Note: Silently ignore errors from this. CodeIntelError
                # mostly just mean that the file was just opened and the
                # statusbar wants to know the current scope. We are probably
                # at the first line, so the default is fine. Database errors
                # may include:
                #   OperationalError: database is locked
                log.debug(str(ex))
                retval = (0, None, 0, None)
            return retval
        finally:
            self._csLock.release()

    def getAdjustedCurrentScopeInfo(self, scimoz, position):
        """A getScopeInfoForFileAndLine() adjusted for recent edits."""
        self._csLock.acquire()
        try:
            file_id, table, id, row = self.getAdjustedCurrentScope(scimoz, position)
            if table is None:
                retval = (None, None, None, None)
            elif table == "module":
                retval = ("module", row[M_NAME], None, None)
            else:
                type, id, scope, name, attrStr =\
                    row[S_TYPE], row[S_ID], row[S_SCOPE], row[S_NAME], row[S_ATTRIBUTES]
                attributes = parseAttributes(attrStr)
                typeName = symbolType2Name(type)
                imageURL = cb.getImageURLForSymbol(typeName, attributes)
                desc = cb.getDescForSymbol(typeName, name, attributes, scope,
                                           self._currLanguage)
                retval = (typeName, name, imageURL, desc)
            return retval
        finally:
            self._csLock.release()


class KoCodeIntelService:
    _com_interfaces_ = [components.interfaces.koICodeIntelService]
    _reg_clsid_ = "{CF1F65B6-25EC-4FB3-A2CB-241CB436E377}"
    _reg_contractid_ = "@activestate.com/koCodeIntelService;1"
    _reg_desc_ = "Komodo Code Intelligence Service"

    def __init__(self):
        self.isBackEndActive = self.isUIActive = 0
        self.cidbVersion = '.'.join(map(str, Database.CIDB_VERSION))
        koDirSvc = components.classes["@activestate.com/koDirs;1"].\
                   getService(components.interfaces.koIDirs)
        self.cidbPath = os.path.join(koDirSvc.hostUserDataDir, "codeintel.db")
        self.manager = None

        obsSvc = components.classes["@mozilla.org/observer-service;1"]\
                 .getService(components.interfaces.nsIObserverService)
        self._proxiedObsSvc = getProxyForObject(None,
            components.interfaces.nsIObserverService,
            obsSvc, PROXY_ALWAYS | PROXY_SYNC)
        self.fileSvc = components.classes["@activestate.com/koFileService;1"]\
            .getService(components.interfaces.koIFileService)
        self.langRegistrySvc = components.classes['@activestate.com/koLanguageRegistryService;1'].\
             getService(components.interfaces.koILanguageRegistryService)

        # XXX todo, add pref observer to trn on/off showing projects
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                            .getService().prefs # global prefs
        self._showProjects = prefSvc.getBooleanPref('codeintel_browser_show_projects')
        self._showAllProjects = prefSvc.getBooleanPref('codeintel_browser_show_all_projects')
        self._projects = []
        self._files = [] #files that were sent via opened_document

    def _createManager(self):
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                            .getService().prefs # global prefs
        completionLangs = prefSvc.getStringPref("codeintel_completion_languages").split()
        self.manager = _Manager(coreLanguages=completionLangs)

    def needToUpgradeCIDB(self):
        """Return true if the CIDB needs to be upgraded. Raise an
        exception and setLastError() if cannot upgrade or if CIDB looks
        inappropriate.
        """
        if 0: #XXX for debugging only: delete CIDB to force a new one
            if os.path.exists(self.cidbPath):
                print "DEBUGGING: deleting CIDB to force creation of a new one"
                os.remove(self.cidbPath)
                time.sleep(1)
        if not self.manager:
            self._createManager()
        try:
            state, details = self.manager.getCIDBUpgradeInfo(self.cidbPath)
        except (CodeIntelError, apsw.Error), ex:
            msg = "unexpected error getting CIDB upgrade info (see error log): %s" % ex
            log.exception(msg)
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(0, msg)
            raise ServerException(nsError.NS_ERROR_FAILURE, msg)
        if state == codeintel.CIDB_UPGRADE_NOT_NECESSARY:
            return False
        elif state == codeintel.CIDB_UPGRADE_NECESSARY:
            return True
        elif state == codeintel.CIDB_UPGRADE_NOT_POSSIBLE:
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(0, details)
            raise ServerException(nsError.NS_ERROR_FAILURE, details)

    def resetCIDB(self):
        if not self.manager:
            self._createManager()
        self.manager.resetCIDB(self.cidbPath, backup=True)

    def activateBackEnd(self):
        if self.isBackEndActive: return
        # Setup to use the 'codeintel' package.
        if not self.manager:
            self._createManager()
        cb.initialize(self.manager)
        try:
            self.manager.initialize(self.cidbPath)
        except (CodeIntelError, apsw.Error, EnvironmentError), ex:
            err = "Error activating Code Intelligence backend: "+str(ex)
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                           .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(0, err)
            raise ServerException(nsError.NS_ERROR_FAILURE, err)
        else:
            self.isBackEndActive = 1

    def activateUI(self, codeBrowserMgr):
        if self.isUIActive: return
        if not self.isBackEndActive:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                "must activate the back end before activating the UI")
        self.codeBrowserMgr = codeBrowserMgr
        self._codeBrowserTree = UnwrapObject(self.codeBrowserMgr.view)
        self.manager.activateUI(self.codeBrowserMgr.view)
        self._codeBrowserTree.activate(self.manager, self.codeBrowserMgr)
        self.isUIActive = 1

    def deactivate(self):
        self.isBackEndActive = self.isUIActive = 0
        self.manager.finalize()
        self.manager = None

    def haveRunBuildCIDBWizard(self):
        prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                            .getService().prefs # global prefs
        # For historical reasons we are still using the pref of this name.
        # With the addition of CIDB upgrading in Komodo 3.1 the
        # "_for_cidb_version" tracking is no longer necessary.
        value = prefSvc.getStringPref("codeintel_have_run_build_wizard_for_cidb_version")
        haveRun = bool(value.strip())
        return haveRun

    def _getCIPath(self, document):
        if document.isUntitled:
            cipath = os.path.join("<Unsaved>", document.displayPath)
        else:
            cipath = document.displayPath
        return cipath

    def ideEvent_EditedCurrentDocument(self, document, scimoz, linesAdded,
                                       rescan=0):
        import codeintel
        if not self.isUIActive:
            return

        language = document.language
        cipath = self._getCIPath(document)
        ## This event name can happen _very_ frequently so we don't
        ## even want to call the effective-no-op-"log.info()" in
        ## production use.
        #log.info("ideEvent edited_current_document: %s (%s)", cipath, language)

        if self.manager.isSupportedLanguage(language):
            if linesAdded:
                self.manager.editedCurrentFile(scimoz, linesAdded)
            if rescan:
                request = codeintel.ScanRequest(cipath,
                                                language,
                                                codeintel.PRIORITY_CURRENT,
                                                content=document.buffer)
                self.manager.stageRequest(request)

    def _addURIRequest(self, uri):
        # Determine the probable language from the file basename.
        f = self.fileSvc.getFileFromURI(uri)
        language = self.langRegistrySvc.suggestLanguageForFile(f.baseName)
        cipath = f.displayPath
        if cipath in self._files: return
        mtime = f.lastModifiedTime
        return self._addPathRequest(cipath, language, mtime=mtime)

    def _addPathRequest(self, cipath, language, name = "opened_document", document=None, mtime=None):
        if self.manager.isSupportedLanguage(language):
            self._codeBrowserTree.wsAddFile(cipath, language)
            if name == "opened_document":
                self.manager.setCurrentFile(cipath, language)
                if document:
                    mtime = document.file and document.file.lastModifiedTime or None
            content = None
            if document:
                content = document.buffer
            request = codeintel.ScanRequest(cipath,
                                            language,
                                            codeintel.PRIORITY_OPEN,
                                            content=content,
                                            mtime=mtime)
            self.manager.stageRequest(request)
            return 1
        return 0

    def ideEvent(self, name, sdata, odata):
        #log.info("ideEvent(name='%s', sdata=%r, odata=%r)", name, sdata, odata)
        import codeintel
        try:
            if not self.isUIActive:
                return

            if (name == "opened_document" or name == "changed_document"):
                url, document = sdata, odata
                language = document.language
                cipath = self._getCIPath(document)
                log.info("ideEvent %s: %s (%s)", name, cipath, language)
                if self.manager.isSupportedLanguage(language):
                    if name == "opened_document":
                        self._codeBrowserTree.wsAddFile(cipath, language)
                        self.manager.setCurrentFile(cipath, language)
                    mtime = document.file and document.file.lastModifiedTime or None
                    request = codeintel.ScanRequest(cipath,
                                                    language,
                                                    codeintel.PRIORITY_OPEN,
                                                    content=document.buffer,
                                                    mtime=mtime)
                    self.manager.stageRequest(request)
                    self._files.append(cipath)
            elif name == "closing_document":
                dummy, document = sdata, odata
                cipath = self._getCIPath(document)
                try:
                    self._files.remove(cipath)
                except ValueError, ex:
                    log.exception("(PLEASE ADD THIS ERROR MESSAGE TO KOMODO "
                                  "BUG 41280): %s: cipath=%r self._files=%r",
                                  ex, cipath, self._files)
                if self._showProjects and not document.isUntitled:
                    for p in self._projects:
                        urls = p.getAllContainedURLs()
                        if document.file.URI in urls:
                            return
                self._codeBrowserTree.wsRemoveFile(cipath)
                if isUnsavedPath(cipath):
                    self._removeUnsavedFile(cipath)
            #XXX TODO for project files in Code Browser:
            # - uriparse.displayPath() is what we want
            # - need to have added to and removed from project in
            #   added_file_to_project and removed_file_from_project events
            # - figure out a W.S. system for these that is efficient
            # - get a separate mode for PseudoModule in cb.py (perhaps
            #   should finally separate PseudoModule into parts: three
            #   sub-classes: ErrorNode, NotScannedModule, ScanningModule)
            # - consider adding a toggle button to turn off "files in
            #   projects"
            # - consider adding a "Scan" entry to context menu on modules
            #   in C.B.
            elif name == "current_project_changed":
                if self._showAllProjects or not self._showProjects:
                    return
                dummy, project = sdata, odata.QueryInterface(components.interfaces.koIProject)
                if self._projects and project == self._projects[0]:
                    return
                if self._projects:
                    urls = self._projects[0].getAllContainedURLs()
                    for url in urls:
                        f = self.fileSvc.getFileFromURI(url)
                        cipath = f.displayPath
                        if cipath not in self._files:
                            self._codeBrowserTree.wsRemoveFile(cipath)
                    self._projects = []
                if project:
                    self._projects.append(project)
                    urls = self._projects[0].getAllContainedURLs()
                    for url in urls:
                        self._addURIRequest(url)
            elif name == "closing_project":
                # remove all project files that are not open in the editor
                if not self._showProjects or not self._projects:
                    return
                dummy, project = sdata, odata.QueryInterface(components.interfaces.koIProject)
                if not self._showAllProjects and project != self._projects[0]:
                    return
                urls = project.getAllContainedURLs()
                self._projects.remove(project)
                for url in urls:
                    f = self.fileSvc.getFileFromURI(url)
                    cipath = f.displayPath
                    if cipath not in self._files:
                        self._codeBrowserTree.wsRemoveFile(cipath)
            #    print "XXX:TODO let CodeBrowserMgr know project '%s' is closing" % project.name
            elif name == "opened_project":
                if not self._showProjects or (not self._showAllProjects and self._projects):
                    return
                # add all project files that are in the project
                url, project = sdata, odata.QueryInterface(components.interfaces.koIProject)
                self._projects.append(project)
                log.info("ideEvent %s: project='%s'", project.name, os.path.basename(project.url))
                urls = project.getAllContainedURLs()
                for url in urls:
                    self._addURIRequest(url)
            #    pprint.pprint(urls)
            elif name == "added_file_to_project":
                if not self._showProjects:
                    return
                url, project = sdata, odata.QueryInterface(components.interfaces.koIProject)
                log.info("ideEvent %s: %s", name, os.path.basename(url))
                self._addURIRequest(url)
            elif name == "removed_file_from_project":
                if not self._showProjects:
                    return
                url, project = sdata, odata.QueryInterface(components.interfaces.koIProject)
                log.info("ideEvent %s: %s", name, os.path.basename(url))
                for p in self._projects:
                    urls = p.getAllContainedURLs()
                    if url in urls:
                        return
                f = self.fileSvc.getFileFromURI(url)
                cipath = f.displayPath
                if cipath not in self._files:
                    self._codeBrowserTree.wsRemoveFile(cipath)
            elif (name == "switched_current_language" or
                  name == "switched_current_document"):
                # Switched current editor pane view to a new document or
                # changed that document's language.
                # Note: It would probably be cleaner to pass in a koIView,
                #       but our XBL/koIView mixing doesn't play nicely with
                #       (Py)XPCOM.
                dummy, document = sdata, odata
                language = document.language
                cipath = self._getCIPath(document)
                log.info("ideEvent %s: %s (%s)", name, cipath, language)
                if name == "switched_current_language":
                    self._codeBrowserTree.wsRemoveFile(cipath)
                    if isUnsavedPath(cipath):
                        self._removeUnsavedFile(cipath)
                if self.manager.isSupportedLanguage(language):
                    self.manager.setCurrentFile(cipath, language)
                    mtime = document.file and document.file.lastModifiedTime or None
                    request = codeintel.ScanRequest(cipath,
                                                    language,
                                                    codeintel.PRIORITY_CURRENT,
                                                    content=document.buffer,
                                                    mtime=mtime)
                    self.manager.stageRequest(request)
                    if name == "switched_current_language":
                        self._codeBrowserTree.wsAddFile(cipath, language)
            else:
                raise NotImplementedError("unknown ideEvent name: '%s'" % name)
        except Exception, ex:
            #XXX Error trapping. Should just log and then drop any error... i.e.
            #    I don't think there is anything here that should be fatal.
            log.exception("Whoa! Unexpected failure handling ideEvent '%s'"
                          % name)

    def _removeUnsavedFile(self, cipath):
        db = Database(self.manager)
        db.removeFile(cipath)

    def batchUpdateRequest(self, type, path, language, extra):
        self.manager.batchUpdateRequest(type, path, language, extra)
    def batchUpdateStart(self, uiHandler):
        self.manager.batchUpdateStart(uiHandler)
    def batchUpdateStop(self):
        self.manager.batchUpdateStop()
    def batchUpdateWait(self):
        self.manager.batchUpdateWait()
    def batchUpdateGetErrors(self):
        return self.manager.batchUpdateGetErrors()

    def isSupportedLanguage(self, language):
        return self.manager.isSupportedLanguage(language)

    def getSupportedLanguages(self):
        return self.manager.getSupportedLanguages()

    def getScopeForFileAndLine(self, path, line):
        return self.manager.getScopeForFileAndLine(path, line)[:3]
    def getScopeInfoForFileAndLine(self, path, line, language):
        return self.manager.getScopeInfoForFileAndLine(path, line, language)
    def getAdjustedCurrentScope(self, scimoz, position):
        return self.manager.getAdjustedCurrentScope(scimoz, position)[:3]
    def getAdjustedCurrentScopeInfo(self, scimoz, position):
        return self.manager.getAdjustedCurrentScopeInfo(scimoz, position)

    def _sendStatusMessage(self, msg, highlight=0, timeout=5000):
        sm = components.classes["@activestate.com/koStatusMessage;1"]\
             .createInstance(components.interfaces.koIStatusMessage)
        sm.category = "codeintel"
        sm.msg = msg
        sm.timeout = timeout
        sm.highlight = highlight
        try:
            self._proxiedObsSvc.notifyObservers(sm, "status_message", None)
        except COMException, ex:
            pass

    def getMembers(self, language, path, line, citdl, explicit,
                   scopeFileId=0, scopeTable=None, scopeId=0, content=None):
        types, members = [], []
        try:
            typesAndMembers = self.manager.getMembers(language, path, line,
                citdl, scopeFileId, scopeTable, scopeId, content)
            if not typesAndMembers:
                self._sendStatusMessage(
                    "No AutoComplete members for '%s'" % citdl, explicit)
            else:
                for t,m in typesAndMembers:
                    types.append(t)
                    members.append(m)
        except codeintel.CodeIntelError, ex:
            self._sendStatusMessage("error determining members: "+str(ex),
                                    explicit)
        return types, members

    def getCallTips(self, language, path, line, citdl, explicit,
                    scopeFileId=0, scopeTable=None, scopeId=0, content=None):
        try:
            calltips = self.manager.getCallTips(language, path, line, citdl,
                scopeFileId, scopeTable, scopeId, content)
            if not calltips:
                self._sendStatusMessage("No CallTip for '%s'" % citdl,
                                        explicit)
        except codeintel.CodeIntelError, ex:
            self._sendStatusMessage("error determining CallTip: "+str(ex),
                                    explicit)
            calltips = []
        return calltips

    def getSubimports(self, language, module, cwd, explicit):
        try:
            subimports = self.manager.getSubimports(language, module, cwd,
                                                    explicit)
            if not subimports:
                self._sendStatusMessage("No available subimports for '%s'" % module,
                                        explicit)
        except codeintel.CodeIntelError, ex:
            self._sendStatusMessage("error determining subimports: "+str(ex),
                                    explicit)
            subimports = []
        return subimports


class KoCodeIntelObjectBrowserTreeView(TreeView):
    _com_interfaces_ = [components.interfaces.koICodeIntelObjectBrowserTreeView]
    _reg_clsid_ = "{25BEE466-F460-46D2-8FA5-082E134A1E3E}"
    _reg_contractid_ = "@activestate.com/koCodeIntelObjectBrowserTreeView;1"
    _reg_desc_ = "Komodo Object Browser Tree View"

    def __init__(self):
        TreeView.__init__(self)
        self._tree = None
        self._sortedBy = None
        self._root = None
        self._rows = []
        self._dataLock = threading.RLock()
        self._searchTerm = ''
        self._exact = 1
        self._browsing = 0
        self._collectionId = ''
        self._sortedBy = 'file_path'
        self._codeIntelSvc = components.classes["@activestate.com/koCodeIntelService;1"]\
                              .getService(components.interfaces.koICodeIntelService);
        self._manager = UnwrapObject(self._codeIntelSvc).manager
        self._languageIdCache = {}

    def _getLanguageId(self, cu, language):
        """CIDB language_id caching.
        
        Returns None if the given language is not in the CIDB.
        """
        if language not in self._languageIdCache:
            cu.execute("SELECT id FROM language WHERE name=? LIMIT 1", (language,))
            try:
                self._languageIdCache[language] = cu.next()[0]
            except StopIteration:
                return None
        return self._languageIdCache[language]

    def getModuleCountForLanguage(self, language):
        cx = self._manager.getDatabaseConnection()
        cu = cx.cursor()
        self._manager.acquireCIDBReadLock()
        try:
            if language:
                language_id = self._getLanguageId(cu, language)
                if language_id is None:
                    return 0
                cu.execute("SELECT count(module.id) FROM module, file "
                           "WHERE module.id=file.id "
                           "AND file.language_id=%d" % language_id)
            else:
                cu.execute("SELECT count(id) FROM module")
            return cu.next()[0]
        finally:
            self._manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    def getFileCountForLanguage(self, language):
        cx = self._manager.getDatabaseConnection()
        cu = cx.cursor()
        self._manager.acquireCIDBReadLock()
        try:
            if language:
                cu.execute("SELECT count(id) FROM file "
                           "WHERE file.language_id=(SELECT id FROM language WHERE name=?)",
                           (language,))
            else:
                cu.execute("SELECT count(id) FROM file")
            return cu.next()[0]
        finally:
            self._manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    def search(self, collectionId, symbol, module, language, exact,
               showImports):
        if not self._tree: return
        if not self._manager:
            return # we must not have been activated yet

        simpleSymbol = '_' not in symbol and '%' not in symbol
        simpleModule = '_' not in module and '%' not in module
        self._collectionId = collectionId
        self._exact = exact
        classes = {"class": cb.CBClassNode,
                   "function": cb.CBFunctionNode,
                   "variable": cb.CBVariableNode,
                   "interface": cb.CBInterfaceNode}

        cx = self._manager.getDatabaseConnection()
        cu = cx.cursor()
        self._manager.acquireCIDBReadLock()
        try:
            if language:
                language_id = self._getLanguageId(cu, language)
                if language_id is None:
                    log.exception("language '%s' not in the CIDB", language)
                    return
    
            if not symbol:
                # we're doing a module browse
                sql = ("SELECT module.id, module.file_id, module.name,"
                       "       file.compare_path, file.path, language.name "
                       "  FROM file,language,module "
                       " WHERE file.id=module.file_id "
                       "   AND language.id=file.language_id"
                       "   AND module.name LIKE ?")
                args = ['%'+module+'%']
                if language:
                    sql += " AND file.language_id=%d" % language_id
                sql += " ORDER BY module.name"
    
                self._nodeIsOpen = {}
                self._browsing = 1
                self._root = root = cb.CBRootNode()
                self._dataLock.acquire()
                try:
                    data = tuple( cu.execute(sql, args) )
    
                    for (module_id, file_id, module_name, file_cpath,
                         file_path, language_name) in data:
                        module = cb.CBModuleNode([module_id],
                                                 file_id,
                                                 module_name,
                                                 line=0,
                                                 attributes={},
                                                 doc='',
                                                 file_cpath=file_cpath,
                                                 file_path=file_path,
                                                 file_language=language_name,
                                                 scan_errors={})
                        try:
                            root.addModule(cu, module, True)
                        except cb.CBError, ex:
                            log.error("problem adding module to Object "
                                      "Browser: %s", ex)
                            continue # drop this error for now
    
                    oldRowCount = len(self._rows)
                    rows = []
                    self._root.generateRows(rows, self._nodeIsOpen,
                                            self._sortedBy, cx=cx, cu=cu)
                    self._rows = rows
                    self.sortBy('name', True)
    
                    if self._tree:
                        self._tree.beginUpdateBatch()
                        self._tree.rowCountChanged(0, len(self._rows)-oldRowCount)
                        self._tree.invalidate()
                        self._tree.endUpdateBatch()
                finally:
                    self._dataLock.release()
    
                self._tree.beginUpdateBatch()
                self._tree.invalidate()
                self._tree.endUpdateBatch()

                self._tree.view.selection.select(self._tree.view.selection.currentIndex)



                return
    
            else:
                # we're doing a 'symbol' search
                sql = ("SELECT symbol.id, symbol.file_id, symbol.module_id, "
                       "       symbol.scope_id, symbol.scope, symbol.name, "
                       "       symbol.signature, symbol.line, symbol.type, "
                       "       symbol.attributes, symbol.doc, module.name, "
                       "       file.compare_path, file.path, language.name "
                       "  FROM symbol, file, language, module"
                       " WHERE file.id=symbol.file_id"
                       "   AND language.id=file.language_id"
                       "   AND module.id=symbol.module_id")
                args = []
                if language:
                    sql += " AND file.language_id=%d" % language_id
                if module:
                    sql += " AND module.name LIKE ?"
                    args.append('%'+module+'%')
                if exact:
                    sql += " AND symbol.name=?"
                    args.append(symbol)
                elif simpleSymbol:
                    sql += " AND symbol.name LIKE ?"
                    args.append('%'+symbol+'%')
                else:
                    sql += " AND symbol.name LIKE ?"
                    args.append('%'+symbol.replace('%', '_')+'%')
                sql += ' ORDER BY symbol.name'
                
                results = {}
                self._browsing = 0
                self._dataLock.acquire()
                try:
                    data = tuple(cu.execute(sql, args))
    
                    rows = []
                    for (id, file_id, module_id, scope_id, scope, name,
                         signature, line, type, attrStr, doc,
                         module_name, file_cpath,
                         file_path, file_language) in data:
    
                        if not simpleSymbol:
                            # double-check -- we need to postprocess
                            # to make sure that _'s are really filtered
                            if ((exact and name != symbol) or
                                (not exact and name.find(symbol) == -1)):
                                continue
                        # add to current module
                        typeName = symbolType2Name(type)
                        kw = {'id': id,
                              'file_id': file_id,
                              'module_id': module_id,
                              'scope_id': scope_id,
                              'scope': scope,
                              'name': name,
                              'signature': signature,
                              'line': line,
                              'type': typeName,
                              'attributes': parseAttributes(attrStr),
                              'doc': doc,
                              'file_cpath': file_cpath,
                              'file_path': file_path,
                              'file_language': file_language,
                              'module_name': module_name}
                        if typeName in classes:
                            node = classes[typeName](**kw)
                            row = {'node': node, 'level': 0}
                            rows.append(row)
    
                    if showImports:
                        sql = ("SELECT import.id, import.file_id, "
                               "       import.module_id, import.scope_id, "
                               "       import.line, import.module, "
                               "       module.name, import.symbol, "
                               "       import.alias, file.compare_path, "
                               "       file.path, language.name "
                               "  FROM import, file, language, module "
                               " WHERE file.id=import.file_id "
                               "   AND language.id=file.language_id "
                               "   AND module.id=import.module_id")
                        args = []
                        if language:
                            sql += " AND file.language_id=%d" % language_id
                        if module:
                            sql += " AND module.name LIKE ?"
                            args.append('%'+module+'%')
                        if exact:
                            sql += " AND (import.symbol=? OR import.module=?)"
                            args += [symbol, symbol]
                        else:
                            cleaned = simpleSymbol and symbol or symbol.replace('%', '_')
                            sql += " AND (import.symbol LIKE ? OR import.module LIKE ?)"
                            args += ['%'+cleaned+'%', '%'+cleaned+'%']
                        sql += " ORDER BY import.module_id"
    
                        for (id, file_id, module_id, scope_id, line,
                             imported_module_name,
                             module_name, name, alias, file_cpath, file_path,
                             file_language) in cu.execute(sql, args):
    
                            if not simpleSymbol:
                                # double-check -- we need to postprocess
                                # to make sure that _'s are really filtered
                                if ((exact and imported_module_name != symbol) or
                                    (not exact and imported_module_name.find(symbol) == -1)):
                                    continue
                            # add to current module
                            typeName = 'import'
                            kw = {'id': id,
                                  'file_id': file_id,
                                  'module_id': module_id,
                                  'scope_id': scope_id,
                                  'line': line,
                                  'module': imported_module_name,
                                  'module_name': module_name,
                                  'symbol': name,
                                  'alias': alias,
                                  'file_cpath': file_cpath,
                                  'file_path': file_path,
                                  'file_language': file_language}
                            node = cb.CBImportNode(**kw)
                            row = {'node': node, 'level': 0}
                            rows.append(row)
        
                    self._rows = rows
                finally:
                    self._dataLock.release()
        
                if self._tree:
                    self._tree.beginUpdateBatch()
                    self._tree.invalidate()
                    self._tree.endUpdateBatch()

                    self._tree.view.selection.select(self._tree.view.selection.currentIndex)



        finally:
            self._manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    #---- current node accessor attributes
    def _getCurrentNode(self):

        index = self._tree.view.selection.currentIndex



        if index == -1 or index >= len(self._rows):
            return None
        return self._rows[index]['node']

    def get_currentNodeType(self):
        node = self._getCurrentNode()
        if not node: return None
        return node.type

    def get_currentNodeFilePath(self):
        node = self._getCurrentNode()
        if not node: return None
        if node.type in ("file", "pseudo-module"):
            return node.cpath
        else:
            return node.file_path

    def get_currentNodeLine(self):
        node = self._getCurrentNode()
        if node.line is None:
            return 0
        return node.line

    def get_currentNodeLanguage(self):
        node = self._getCurrentNode()
        if not node: return None
        return node.file_language

    #---- nsITreeView methods
    def get_rowCount(self):
        try:
            return len(self._rows)
        except TypeError: # self._rows is None
            return 0

    def sortBy(self, key, initial=False):
        if not self._tree:
            return
        self._dataLock.acquire()
        try:
            if initial or self._sortedBy != key:
                self._sortedBy = key
                self._doSort()
            else:
                self._rows.reverse()
        finally:
            self._dataLock.release()

        self._tree.beginUpdateBatch()
        self._tree.invalidate()
        self._tree.endUpdateBatch()

    def _doSort(self):
        key = self._sortedBy
        if key == 'line': # do a numeric sort for line #
            self._rows.sort(lambda a,b: cmp(getattr(a['node'], key, ''),
                                            getattr(b['node'], key, '')))
        elif key == 'name':
            self._rows.sort(lambda a,b: cmp(_nodeToName(a['node'], self._browsing).lower(),
                                            _nodeToName(b['node'], self._browsing).lower()))
        elif key == 'module':
            self._rows.sort(lambda a,b: cmp(_nodeToModuleName(a['node']).lower(),
                                            _nodeToModuleName(b['node']).lower()))
        else:
            # deal with "None" values as well through "or ''" trick.
            self._rows.sort(lambda a,b: cmp((getattr(a['node'], key, '') or ''),
                                            (getattr(b['node'], key, '') or '')))

    def getCellText(self, row, column):

        col = column.id



        cell = ""
        try:
            node = self._rows[row]['node']
            if col == "name":
                return _nodeToName(node, self._browsing)
            elif col == "module":
                return _nodeToModuleName(node)
            elif col == "file_path":
                if self._browsing:
                    if node.type == 'module':
                        retval = node.file_path
                    else:
                        retval = ''
                else:
                    retval = node.file_path
                if retval.endswith('.cix'):
                    retval = "(Precompiled Data)"
                return retval
            elif col == "line":
                if (node.type != 'module' and
                    node.line >= 1):
                        return str(node.line)
                return ''
            else:
                cell = getattr(node, col)
                if type(cell) not in types.StringTypes:
                    cell = str(cell)
                return cell
        except IndexError:
            log.error("no row %d" % row)
        except KeyError:
            log.error("unknown column: '%s'" % col)
        if type(cell) not in types.StringTypes:
            cell = str(cell)
        return cell

    def getCellProperties(self, row, col, properties):
        return
        #assert col == "codebrowser-tree-node"
        if row == self._showDetailForRow:
            properties.AppendElement(self._showDetailAtom)

    def isContainer(self, row):
        try:
            node = self._rows[row]["node"]
            if node.type == "module":
                if not node.linkedUp:
                    return 1
            return len(node.children)
        except IndexError, ex:
            log.error("error checking if row %d is a container: %s", row, ex)
            if log.isEnabledFor(logging.ERROR):
                log.error("--------------- current rows -------------------")
                pprint.pprint(self._rows)
                log.error("--------------- end of rows --------------------")
                import traceback
                traceback.print_exception(*sys.exc_info())
            return 0
    def isContainerEmpty(self, row):
        return 0
    def isContainerOpen(self, row):
        return self._rows[row]["node-is-open"]

    def toggleOpenState(self, row):
        node = self._rows[row]["node"]
        if node.type == 'module' and not node.linkedUp:
            node.linkUp()

        self._dataLock.acquire()
        try:
            scopedName = self._rows[row]["scoped-name"]
            isOpen = self.isContainerOpen(row)
            self._nodeIsOpen[scopedName] = not isOpen

            # Must recalculate the rows.
            oldRowCount = len(self._rows)
            self._rows = []
            rows = []
            self._root.generateRows(rows, self._nodeIsOpen, self._sortedBy)
            self._rows = rows
        finally:
            self._dataLock.release()

        self._tree.beginUpdateBatch()
        self._tree.rowCountChanged(0, len(self._rows)-oldRowCount)
        self._tree.invalidate()
        self._tree.endUpdateBatch()
        self.selection.currentIndex = row
        self.selection.select(row);

    def getParentIndex(self, row):
        return self._rows[row]["parent-index"]

    def getLevel(self, row):
        return self._rows[row]["level"]

    def getImageSrc(self, row, column):

        col = column.id



        node = self._rows[row]["node"]
        if col == "name":
            return node.imageURL
        elif col == "file_path" and \
             (node.type == 'module' or not self._browsing):
                language = node.file_language
                return cb.getImageURLForModule(language)
        return ''

    def hasNextSibling(self, rowIndex, afterIndex):
        """From the nsITreeView.idl docs:

        HasNextSibling is used to determine if the row at rowIndex has a
        nextSibling that occurs *after* the index specified by
        afterIndex.  Code that is forced to march down the view looking
        at levels can optimize the march by starting at afterIndex+1.
        """
        nextSiblingIndex = self._rows[rowIndex].get("next-sibling-index", -1)
        return nextSiblingIndex > afterIndex

def _nodeToName(node, browsing):
    if node.type == 'module':
        return node.name
    if node.type == 'container':
        return node.title
    if not browsing and node.type != 'import' and node.scope:
        separators = {
            "Python": ".",
            "Perl": "::",
            "PHP": "::",
            #XXX Though it isn't really this easy for Ruby, I don't think.
            "Ruby": "::",
        }
        separator = separators.get(node.file_language, ".")
        return separator.join([node.scope, node.name])        
    return node.name

def _nodeToModuleName(node):
    if node.type == 'module':
        return node.name
    else:
        return node.module_name
