#!/usr/bin/env python
# Copyright (c) 2005 ActiveState Corp, a division of Sophos Plc.
# See the file LICENSE.txt for licensing information.

"""A module to hold all Komodo's <textbox type="autocomplete"/> autocomplete
search types.

In XUL, one identifies the autocomplete source with the
"autocompletesearch" attribute:

    <textbox type="autocomplete"
             autocompletesearch="history"
             .../>

This instanciates a component with the contract ID:
    @mozilla.org/autocomplete/search;1?name=<name>
(in this example "@mozilla.org/autocomplete/search;1?name=history")
implementing the nsIAutoCompleteSearch interface.

See class KoTACDemoSearch for a sample search implementation to get you
started.

Currently implemented textbox autocomplete types:
    mru                 complete Komodo MRU preference list
    filepath            complete local paths
    mru_and_filepath    complete both an MRU pref and local paths
    dirpath             complete local directories
    mru_and_dirpath     complete both an MRU pref and local dirs
"""

import os
import logging

from xpcom import components, COMException, ServerException, nsError

log = logging.getLogger("koTAC")
#log.setLevel(logging.DEBUG)


#---- helper stuff for autocomplete implementations below

class KoTACResult:
    """A base implementation of nsIAutoCompleteResult.
    
    This component can be used directly or may be subclassed for
    specific textbox autocomplete search types if extra functionality is
    required. If subclassed, all the _reg_*_ attributes MUST be
    overriden.

    Typical usage for creating a search result:

    1. For a successful search:
            result = KoTACResult(searchString)
            for match in matches:
                result.addMatch(value, ...)
       Note that if there are no matches found this usage will properly
       set the searchResult to RESULT_NOMATCH.
    2. For an invalid search string:
            result = KoTACResult(searchString,
                searchResult=nsIAutoCompleteResult.RESULT_IGNORED)
    3. For a failed search:
            result = KoTACResult(searchString,
                searchResult=nsIAutoCompleteResult.RESULT_FAILURE,
                errorDescription="...")

    http://lxr.mozilla.org/seamonkey/source/toolkit/components/autocomplete/public/nsIAutoCompleteResult.idl
    """
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteResult]
    _reg_clsid_ = "{F29911D8-F3C3-11D9-A507-000D934CF260}"
    _reg_contractid_ = "@activestate.com/autocomplete/result;1"
    _reg_desc_ = "Komodo textbox autocomplete search result"

    def __init__(self, searchString, searchResult=None, defaultIndex=0,
                 errorDescription=None, matches=None):
        """Create a new nsIAutoCompleteResult instance.

            "searchString" is the original search string.

        All other arguments are optional. See the discussion of typical
        usage in the class docstring.
            "searchResult" is one of the nsIAutoCompleteResult.RESULT_*
                constants. The default is RESULT_NOMATCH -- i.e. a
                successful search is presumed and the search result is
                automatically changed to RESULT_SUCCESS when the first
                match is added with .addMatch().
            "defaultIndex" is 0 by default. Normally this is set with
                the "isDefault" argument to .addMatch(), if at all.
            "errorDescription" is an error string. Only relevant if the
                searchResult is RESULT_FAILURE.
            "matches" is a list of 3-tuples:
                    (<value>, <comment>, <style-hint>)
                If not specified it starts out empty. Results are
                typically added with .addMatch().
        """
        nsIAutoCompleteResult = components.interfaces.nsIAutoCompleteResult
        self.searchString = searchString
        if searchResult is None:
            self.searchResult = nsIAutoCompleteResult.RESULT_NOMATCH
        else:
            self.searchResult = searchResult
        self.defaultIndex = defaultIndex
        self.errorDescription = errorDescription
        if matches is None:
            self.matches = []
        else:
            self.matches = matches
    
    def addMatch(self, value, comment=None, style=None, isDefault=False):
        """Add an AutoComplete search match/hit

            "value" is the (string) value of the match.
            "comment" is an optional comment string (which may or may
                not be shown in the autocomplete textbox UI depending on
                the "showcommentcolumn" XUL attribute).
            "style" is an optional style hint string to be used on this
                match row in the autocomplete textbox UI. Basically,
                this is a CSS class name and the usage here is equivalent to 
                the "showDetail" CSS class used in the Komodo Code
                Browser tree view for highlighting the hovered item.
            "isDefault" (optional, default False) is a boolean
                indicating if this match should be the "defaultIndex".
                The default index is used if the "completedefaultindex"
                XUL attribute is true. By default the zeroth index is
                the default. See:
                http://xulplanet.com/references/elemref/ref_textboxFirefoxAutoComplete.html#attr_completedefaultindex
        """
        nsIAutoCompleteResult = components.interfaces.nsIAutoCompleteResult
        if not self.matches \
           and self.searchResult == nsIAutoCompleteResult.RESULT_NOMATCH:
            self.searchResult = nsIAutoCompleteResult.RESULT_SUCCESS
        self.matches.append((value, comment, style))
        if isDefault:
            self.defaultIndex = len(self.matches) - 1

    #---- nsIAutoCompleteResult implementation
    def get_matchCount(self):
        return len(self.matches)
    def getValueAt(self, index):
        return self.matches[index][0]
    def getCommentAt(self, index):
        return self.matches[index][1]
    def getStyleAt(self, index):
        return self.matches[index][2]
    def removeValueAt(self, rowIndex, removeFromDb=False):
        del self.matches[rowIndex]



#---- component implementations

class KoTACDemoSearch:
    """This is a demostration nsIAutoCompleteSearch implementation to show how
    textbox autocomplete can be implemented in Komodo. There are two methods
    to implement:
        startSearch     Every implementation must write this one. See the
                        docstring below for details.
        stopSearch      Probably only useful if you implement an
                        asynchronous search.
    """
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteSearch]
    _reg_clsid_ = "{9A4DEF8A-F3CD-11D9-AA2A-000D934CF260}"
    _reg_contractid_ = "@mozilla.org/autocomplete/search;1?name=demo"
    _reg_desc_ = "demo textbox autocomplete search: completes filenames in ~/tmp"

    def __init__(self):
        self.tmpdir = os.path.expanduser(os.path.join("~", "tmp"))

    def startSearch(self, searchString, searchParam, previousResult, listener):
        """Synchronously or asynchronously search for the given search string.

            searchString (str)
            searchParam (str)
            previousResult (nsIAutoCompleteResult)
            listener (nsIAutoCompleteObserver)

        The result of the search should be reported via the "listener":
            void onSearchResult(in nsIAutoCompleteSearch search, in nsIAutoCompleteResult result);

        AutoComplete search best practices:
        - If possible limit the search to just the set in
          "previousResult". I.e. if the last search was for "fo" and
          this one is for "foo", then searching within the previous
          result might be faster. If so, then don't need to create a new
          nsIAutoCompleteResult: just pare down the "previousResult" and
          pass it back.
        """
        nsIAutoCompleteResult = components.interfaces.nsIAutoCompleteResult
        log.debug("startSearch(searchString=%r, searchParam=%r, "
                  "previousResult=%r, listener)", searchString, 
                  searchParam, previousResult)
        try:
            filenames = os.listdir(self.tmpdir)
        except EnvironmentError, ex:
            err = "cannot list tmp dir: %s" % ex
            result = KoTACResult(searchString,
                searchResult=nsIAutoCompleteResult.RESULT_FAILURE,
                errorDescription=err)
            listener.onSearchResult(self, result)
            return

        result = KoTACResult(searchString)
        for name in [f for f in filenames if f.startswith(searchString)]:
            path = os.path.join(self.tmpdir, name)
            comment = "%s bytes" % os.stat(path).st_size
            style = None
            # If uncommenting this, then must be sure to include the
            # relevant CSS in the XUL with the autocomplete textbox.
            # (For "showDetail" this is codeintel.css.)
            #if name.endswith(".py"):
            #    style = "showDetail"
            result.addMatch(name, comment, style)
        listener.onSearchResult(self, result)

    def stopSearch(self):
        """This is sent by the autocomplete controller to stop a
        possible previous asynchronous search.
        """
        pass


class KoTACFilePathSearch:
    """Textbox AutoComplete search for local filepaths."""
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteSearch]
    _reg_clsid_ = "{681A2200-F3E9-11D9-8F02-000D934CF260}"
    _reg_contractid_ = "@mozilla.org/autocomplete/search;1?name=filepath"
    _reg_desc_ = "local file path textbox autocomplete search"

    dirsOnly = False

    def startSearch(self, pattern, cwd, previousResult, listener):
        """Synchronously complete the given path pattern. If the path is
        relative, then treat it relative to the given "cwd". If "cwd" is
        empty then there is not commplete on relative paths.
        """
        #XXX Disable this optimization because
        #    - It screws up with non-normalized paths, e.g. 'foo//bar'.
        if False and (previousResult and previousResult.searchString
            and pattern.startswith(previousResult.searchString)
            and os.sep not in pattern[len(previousResult.searchString):]):
            # OPTIMIZATION: Use the previous result too more efficiently
            # pare down the list (instead of stat'ing again).
            result = KoTACResult(pattern)
            for i in range(previousResult.matchCount):
                value = previousResult.getValueAt(i)
                if value.startswith(pattern):
                    result.addMatch(value)
        else:
            result = KoTACResult(pattern)
            #XXX For mounted filesystems this could take a long time.
            #    Ideally this should be asynchronous with the result being
            #    dropped if get stopSearch() before finished.
            for path in _genPathCompletions(pattern, cwd, self.dirsOnly):
                result.addMatch(path)

        listener.onSearchResult(self, result)

    def stopSearch(self):
        pass


class KoTACDirPathSearch(KoTACFilePathSearch):
    """Textbox AutoComplete search for local directory paths."""
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteSearch]
    _reg_clsid_ = "{A6725BFC-FB38-11D9-977C-000D93B826C2}"
    _reg_contractid_ = "@mozilla.org/autocomplete/search;1?name=dirpath"
    _reg_desc_ = "local directory path textbox autocomplete search"

    dirsOnly = True # argument to path completion generator


class KoTACMruSearch:
    """Textbox AutoComplete search for an MRU (most recently used list)
    in Komodo preferences.
    """
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteSearch]
    _reg_clsid_ = "{686ED0A6-F4AC-11D9-9048-000D934CF260}"
    _reg_contractid_ = "@mozilla.org/autocomplete/search;1?name=mru"
    _reg_desc_ = "Komodo MRU pref textbox autocomplete search"

    def __init__(self):
        self.prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                .getService().prefs # global prefs

    def startSearch(self, searchString, prefName, previousResult, listener):
        """Synchronously complete the given string against the Komodo
        MRU pref list.
        """
        DEBUG = False
        nsIAutoCompleteResult = components.interfaces.nsIAutoCompleteResult
        log.debug("startSearch(searchString=%r, prefName=%r, "
                  "previousResult=%r, listener)", searchString, 
                  prefName, previousResult)

        result = KoTACResult(searchString)
        if DEBUG: allitems = []
        if self.prefSvc.hasPref(prefName):
            mru = self.prefSvc.getPref(prefName)
            for i in range(mru.length):
                item = mru.getStringPref(i)
                if DEBUG: allitems.append(item)
                if item.startswith(searchString):
                    result.addMatch(item)
        if DEBUG: print "All '%s' items: %r" % (prefName, allitems)
        listener.onSearchResult(self, result)

    def stopSearch(self):
        pass


class KoTACMruAndFilePathSearch:
    """Textbox AutoComplete search combining elements (as best as
    possible) of the "filepath" and "mru" autocomplete backends.
    Sometimes you want both, e.g.: the 'folders' part of the Find
    toolbar.
    """
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteSearch]
    _reg_clsid_ = "{6064FA68-F5A3-11D9-B732-000D934CF260}"
    _reg_contractid_ = "@mozilla.org/autocomplete/search;1?name=mru_and_filepath"
    _reg_desc_ = "Komodo MRU+filepath textbox autocomplete search"

    dirsOnly = False

    def __init__(self):
        self.prefSvc = components.classes["@activestate.com/koPrefService;1"]\
                .getService().prefs # global prefs

    def startSearch(self, searchString, searchParam, previousResult, listener):
        r"""Synchronously complete the given string against the Komodo
        MRU pref list *and* the filepath.
        
        Here is how it could (and does currently) work:
        - limit the number of MRU hits to, say, 5 (something small)
        - put the MRU hits first
        - style the MRU and path hits slightly differently (icons?)
        - perhaps make the *default index* the first path hit (so really
          it completes paths by default if the textbox set
          completedefaultindex="true")

        The searchParam (i.e. the 'autocompletesearchparam' attribute)
        syntax is like specifying CSS style attributes (the key/value
        pairs can be in any order):

            cwd: ...; mru: ...; maxmru: ...

        Where:
            'cwd' is the current working directory. If not given then
                there won't be any completions for relative paths.
            'mru' is an MRU list preference name and is require to get
                MRU results.
            'maxmru' is a maximum number of MRU results to return.
                Something like '5' is typical. If not specified then all
                MRU results are returned.
        """
        import re
        DEBUG = False
        nsIAutoCompleteResult = components.interfaces.nsIAutoCompleteResult
        log.debug("startSearch(searchString=%r, searchParam=%r, "
                  "previousResult=%r, listener)", searchString, 
                  searchParam, previousResult)

        # Parse the searchParam
        SEARCH_PARAM_RE = re.compile(r"(?P<key>\w+):\s*(?P<name>[^\s;]*)\s*;?")
        cwd = prefName = mruLimit = None
        if DEBUG: print "startSearch: parse searchParam: %r" % searchParam
        for key, value in SEARCH_PARAM_RE.findall(searchParam):
            if key == "cwd":
                cwd = value
            elif key == "mru":
                prefName = value
            elif key == "maxmru":
                try:
                    mruLimit = int(value)
                except ValueError:
                    log.warn("invalid searchParam 'maxmru' value, skipping: %r" % value)
            else:
                raise ServerException(nsError.NS_ERROR_FAILURE,
                        "illegal searchParam: unknown key: %r" % key)
        if not cwd and not prefName:
            log.warn("potentially bogus autocompletesearchparam for "
                     "mru_and_filepath search: %r" % searchParam)
        if DEBUG:
            print "startSearch: prefName=%r" % prefName
            print "startSearch: cwd=%r" % cwd
            print "startSearch: mruLimit=%r" % mruLimit

        result = KoTACResult(searchString)

        # Pref hits first (limited number)
        if prefName:
            num = 0
            if self.prefSvc.hasPref(prefName):
                mru = self.prefSvc.getPref(prefName)
                for i in range(mru.length):
                    item = mru.getStringPref(i)
                    if item.startswith(searchString):
                        if DEBUG:
                            print "startSearch: mru hit: %r" % item
                        result.addMatch(item, style="ac-mru")
                        num += 1
                        if mruLimit is not None and num >= mruLimit:
                            break
                    else:
                        if DEBUG:
                            print "startSearch: skip mru item: %r" % item

        # Now filepath hits.
        if DEBUG:
            print "startSearch: complete '%s' path in '%s'" % (searchString, cwd)
        num = 0
        for path in _genPathCompletions(searchString, cwd, self.dirsOnly):
            if DEBUG:
                print "startSearch: path hit: '%s'" % path
            isDefault = num == 0
            result.addMatch(path, style="ac-path", isDefault=isDefault)
            num += 1

        listener.onSearchResult(self, result)

    def stopSearch(self):
        pass


class KoTACMruAndDirPathSearch(KoTACMruAndFilePathSearch):
    _com_interfaces_ = [components.interfaces.nsIAutoCompleteSearch]
    _reg_clsid_ = "{0D80BBBF-FB39-11D9-8CCB-000D93B826C2}"
    _reg_contractid_ = "@mozilla.org/autocomplete/search;1?name=mru_and_dirpath"
    _reg_desc_ = "Komodo MRU+dirpath textbox autocomplete search"

    dirsOnly = True



#---- internal support routines

def _genPathCompletions(pattern, cwd, dirsOnly=False):
    import sys, glob
    from os.path import isabs, join, isdir, ismount

    RELATIVE, ABSOLUTE, TILDE = range(3)
    if isabs(pattern):
        type = ABSOLUTE
        abspattern = pattern
    elif pattern and pattern[0] == '~':
        type = TILDE
        abspattern = os.path.expanduser(pattern)
        tilde_part = pattern.split(os.sep, 1)[0]
        home_part = os.path.expanduser(tilde_part)
    else:
        type = RELATIVE
        if cwd:
            abspattern = join(cwd, pattern)
        else:
            raise StopIteration

    for path in glob.glob(abspattern+"*"):
        if isdir(path) or (sys.platform == "win32" and ismount(path)):
            path += os.sep
        elif dirsOnly:
            continue
        if type == RELATIVE:
            if cwd:
                path = path[len(cwd)+len(os.sep):]
        elif type == TILDE:
            path = tilde_part + path[len(home_part):]
        yield path




