#!/usr/bin/env python
# Copyright (c) 2004 Sophos.
# See the file LICENSE.txt for licensing information.
#
# Contributors:
#   Shane Caraveo (ShaneC@ActiveState.com)
#   Trent Mick (TrentM@ActiveState.com)

"""
    phpcile - a Code Intelligence Language Engine for the PHP language

    Module Usage:
        from phpcile import scan
        mtime = os.stat("foo.php")[stat.ST_MTIME]
        content = open("foo.php", "r").read()
        scan(content, "foo.php")
    
    Command-line Usage:
        phpcile.py [<options>...] [<PHP files>...]

    Options:
        -h, --help          dump this help and exit
        -V, --version       dump this script's version and exit
        -v, --verbose       verbose output
        -f, --filename <path>   specify the filename of the file content
                            passed in on stdin, this is used for the "path"
                            attribute of the emitted <file> tag.
        --md5=<string>      md5 hash for the input
        --mtime=<secs>      modification time for output info, in #secs since
                            1/1/70.

    Debugging or Unused Options:
        -t, --text          plain text output mode
        -c, --clock         print time info for the scan (for debugging)
        -r, --recursive     parse included files
        -i, --include_path  path to resolve included files

    One or more PHP files can be specified as arguments or content can be
    passed in on stdin.

    This is a Language Engine for the Code Intelligence (codeintel) system.
    Code Intelligence XML format. See:
        http://specs.activestate.com/Komodo_3.0/func/code_intelligence.html
    
    The command-line interface will return non-zero iff the scan failed.
"""

import os
import sys
import getopt
import md5
import re
import logging
import string
import time
import glob
import stat
import codecs

#---- exceptions

class PHPCILEError(Exception):
    pass

def _tryEncoding(buffer, encoding):
    """ buffer, encoding -> encoding_buffer

        Attempts to encode the buffer using the specified encoding

        Returns None on failure, a Unicode version of the buffer on success.
    """
    
    try:
        secret_decoder_ring = codecs.lookup(encoding)[1]
    except LookupError, e:
        # the encoding name doesn't exist, likely a pep263 failure
        # an example is using windows-1250 as the name
        return None
    try:
        (outdata,len) = secret_decoder_ring(buffer)
        return outdata
    except Exception, e: # Figure out the real exception types
        return None

def _getEncodedBuffer(buffer):
    decodedBuffer = _tryEncoding(buffer,'utf-8')
    if decodedBuffer is not None:
        return (decodedBuffer, 'utf-8', '')
    return (_tryEncoding(buffer, 'iso8859-1'), 'iso8859-1', '')             
    
#---- global data

_version_ = (0, 2, 0)
log = logging.getLogger("phpcile")
#log.setLevel(logging.DEBUG)


#---- internal routines and classes

def _getAttrStr(attrs):
    """Construct an XML-safe attribute string from the given attributes
    
        "attrs" is a dictionary of attributes
    
    The returned attribute string includes a leading space, if necessary,
    so it is safe to use the string right after a tag name.
    """
    from xml.sax.saxutils import quoteattr
    s = ''
    for attr, value in attrs.items():
        s += ' %s=%s' % (attr, quoteattr(str(value)))
    return s


# match 0x00-0x1f except TAB(0x09), LF(0x0A), and CR(0x0D)
_encre = re.compile('([\x00-\x08\x0b\x0c\x0e-\x1f])')
if sys.version_info >= (2, 3):
    charrefreplace = 'xmlcharrefreplace'
else:
    # Python 2.2 doesn't have 'xmlcharrefreplace'. Fallback to a
    # literal '?' -- this is better than failing outright.
    charrefreplace = 'replace'

def _xmlencode(s):
    """Encode the given string for inclusion in a UTF-8 XML document.
    
    Specifically, illegal or unpresentable characters are encoded as
    XML character entities.
    """
    # As defined in the XML spec some of the character from 0x00 to 0x19
    # are not allowed in well-formed XML. We replace those with entity
    # references here.
    #   http://www.w3.org/TR/2000/REC-xml-20001006#charsets
    # (XXX It would be nice if Python has a codec for this. Perhaps we
    # should write one.)
    return _encre.sub(
               # replace with XML decimal char entity, e.g. '&#7;'
               lambda m: '&#%d;'%ord(m.group(1)),
               s.encode('utf-8', charrefreplace))

_javadoc1 = re.compile(r'\s*\/\*(.*)\*\/', re.S)
_javadoc2 = re.compile(r'^(\s*\*)', re.M)
_linedoc = re.compile(r'^(\s*#|\s*\/\/)', re.M)
_indent = re.compile(r'^([ \t]*)', re.M)
_param = re.compile(r'^\s*@param\s+(?P<type>\w+)\s+\$?(?P<name>\w+)(?:\s+?(?P<doc>.*?))?', re.M)
_return = re.compile(r'^\s*@return\s+(?P<type>\w+)(?:\s+(?P<doc>.*))?', re.M)

def uncommentDocString(doc):
    # remove block style leading and end comments
    d = '\n'.join(re.findall(_javadoc1, doc))
    if d:
        # remove starting * if javadoc style
        d = re.sub(_javadoc2,'',d)
    else:
        d = doc
        # remove line style comments
        d = re.sub(_linedoc,'',d)


    # trim preceeding blank lines.  we dont want to trim the first non-blank line
    lines = d.split('\n')
    while len(lines) and not lines[0].strip():
        lines.pop(0)
    d = '\n'.join(lines)
    
    # trip any blank end lines
    d = d.rstrip()
    
    # guess the indent size    
    spaces = re.findall(_indent, d)
    indent = spaces[0]
    for s in spaces:
        if len(s) and len(s) < indent:
            indent = len(s)

    # dedent the block
    if not indent:
        return d
    dedent = re.compile(r'^([ \t]{%d})' % indent, re.M)
    d = re.sub(dedent, '', d)
    return d

def parseDocString(doc):
    d = uncommentDocString(doc)
    params = re.findall(_param, d)
    result = re.findall(_return, d)
    if result:
        result = result[0]
    return (d, params, result)


class PHPcile:
    def __init__(self):
        # filesparsed contains all files parsed
        self.filesparsed={}
        # needfile contains a list of files included by the file that is the key
        self.needfile={}
        # infile contains a list of files that the key file is included in 
        self.infile={}
        # classindex tells us what file a class definition is contained in
        self.classindex={}
        # functionindex tells us what file a function is defined in
        self.functionindex={}
        # interfaceindex tells us what file an interface definition is contained in
        self.interfaceindex={}

    def _clearindex(self, filename, index):
        tmp = [k for k in index if index[k] == filename]
        for k in tmp:
            del index[k]
        
    def clear(self, filename):
        # clear include links from the cache
        if filename not in self.filesparsed:
            return
        del self.filesparsed[filename]
        
        if filename in self.needfile:
            for f in self.needfile[filename]:
                i = self.infile[f].index(filename)
                del self.infile[f][i]
            del self.needfile[filename]
        
        self._clearindex(filename, self.classindex)
        self._clearindex(filename, self.functionindex)
        self._clearindex(filename, self.interfaceindex)
        
    def __repr__(self):
        r = ''
        for f in self.filesparsed:
            r += repr(self.filesparsed[f])
        return r + '\n'

    def toXML(self, level=0, indent="    "):
        r = '<codeintel version="0.1">\n'
        for f in self.filesparsed.values():
            r += f.toXML(level+1, indent)
        r += '</codeintel>\n'
        data = _getEncodedBuffer(r)
        r = '<?xml version="1.0" encoding="%s"?>\n' % data[1]
        r += data[0].encode(data[1])
        return r

class PHPArg:
    def __init__(self, name, type=None, default=None):
        self.name = name
        self.type = type
        self.default = default
        
    def __repr__(self):
        attrs = ''
        if self.type:
            attrs += '%s ' % self.type
        if self.name:
            attrs += self.name
        if self.default:
            attrs += ' = %s' % self.default
        return attrs
        
    def toXML(self, level=0, indent="    "):
        attrs = {'name': self.name}
        if self.type:
            attrs['type'] = self.type
        if self.default:
            attrs['default'] = self.default
        return '%s<argument%s/>\n' % (indent*level, _getAttrStr(attrs))

class PHPArgs:
    def __init__(self, argline, argtypes={}):
        if argline:
            args = collector.res["args_info"].findall(argline)
        else:
            args = {}
        self.args = {}
        for arg in args:
            self.args[arg[1]] = [arg[0], arg[2]]
        for key, arg in argtypes.items():
            if key in self.args:
                default = self.args[key][1]
            else:
                # nothing right now
                default = argtypes[key][1]
            self.args[key] = [argtypes[key][0], default]
        self.argline = argline
    
    def __repr__(self):
        args = []
        for name, arg in self.args.items():
            args.append(repr(PHPArg(name, arg[0], arg[1])))
        return string.join(args, ', ')

    def toXML(self, level=0, indent="    "):
        out = ''
        for name, arg in self.args.items():
            out += PHPArg(name, arg[0], arg[1]).toXML(level, indent)
        return out

class PHPVariable:
    def __init__(self, name, line, classname='', attributes=''):
        self.name = name
        self.types = [(line, classname)]
        self.line = line
        if attributes:
            self.attributes = ' '.join(attributes.strip().split())
        else:
            self.attributes = None

    def addType(self, line, type):
        self.types.append((line, type))

    def __repr__(self):
        return "var %s line %s type %s attributes %s\n"\
               % (self.name, self.line, self.type, self.attributes)

    def toXML(self, level=0, indent="    "):
        attributes = type = ''
        attrs = {'name': self.name, 'line': self.line}
        if self.attributes:
            attrs['attributes'] = self.attributes
        varel = '%s<variable%s' % (indent*level, _getAttrStr(attrs))
        typeel = ''
        for line, typ in self.types:
            if not typ: continue
            attrs = {'type': typ, 'line': line}
            typeel += '%s<type%s/>\n' % (indent*(level + 1), _getAttrStr(attrs))
        if typeel:
            varel = '%s>\n%s%s</variable>\n' % (varel, typeel, indent*level)
        else:
            varel += '/>\n'
        return varel

class PHPFunction:
    def __init__(self, funcname, args, lineno, depth=0, attributes='',
                 doc=None, classname=''):
        self.name = funcname
        self.argline = args
        self.line = lineno
        self.lineend = 0
        self.depth = depth
        self.variables = {} # all variables used in class
        attributes = attributes.strip().split()
        # build the signature before we add any attributes that are not part
        # of the signature
        if attributes:
            attrs = ' '.join(attributes)
            self.signature = '%s function %s' % (attrs, self.name)
            self.shortSig = '%s %s' % (attrs, self.name)
        else:
            self.signature = 'function %s' % (self.name)
            self.shortSig = self.name
        # both php 4 and 5 constructor methods
        if funcname == '__construct' or (classname and funcname.lower() == classname.lower()):
            attributes += ['__ctor__']
# if we add destructor attributes...
#        elif funcname == '__destruct':
#            attributes += ['__dtor__']
        self.attributes = ' '.join(attributes)
        self.doc = None
        self.returnType = None
        self.argTypes = {}
        if doc:
            docinfo = parseDocString(doc)
            self.doc = docinfo[0]
            if docinfo[1]:
                for arg in docinfo[1]:
                    self.argTypes[arg[1]] = [arg[0], '']
            if docinfo[2]:
                self.returnType = docinfo[2][0]
        if args or self.argTypes:
            self.args = PHPArgs(args, self.argTypes)
        else:
            self.args = PHPArgs('')
    
    def __str__(self):
        if self.doc:
            return "%s(%s)\n%s" % (self.shortSig, self.argline, self.doc)
        return "%s(%s)" % (self.shortSig, self.argline)

    def __repr__(self):
        return self.signature

    def toXML(self, level=0, indent="    "):
        args = doc = ''
        attrs = {'name': self.name, 'line': self.line}
        signature = self.signature
        if self.attributes:
            attrs['attributes'] = self.attributes
        if self.lineend:
            attrs['lineend'] = self.lineend
        
        # XXX The return type is here!
        #if self.returnType:
        #    attrs['returns'] = self.returnType
        
        if self.doc:
            # XXX parse the docstring
            doc = '%s<doc><![CDATA[%s]]></doc>\n'\
                  % (indent*(level+1), self.doc)
        signature += "("
        if self.args:
            args = self.args.toXML(level+1, indent)
            signature += ' '.join(self.argline.split())
        signature += ")"
        attrs["signature"] = signature
        
        varsout = ''
        # XXX for variables inside functions
        #for v in self.variables.values():
        #    varsout += v.toXML(level+1, indent)

        children = doc + args + varsout
        if children:
            return '%s<function%s>\n%s%s</function>\n'\
                   % (indent*level, _getAttrStr(attrs), children,
                      indent*level)
        else:
            return '%s<function%s/>\n' % (indent*level, _getAttrStr(attrs))

class PHPInterface:
    def __init__(self, name, extends, lineno, depth, doc=None):
        self.name = name
        self.extends = extends
        self.linestart = lineno
        self.lineend = 0
        self.depth = depth
        self.members = {} # declared class variables
        self.variables = {} # all variables used in class
        self.functions = {}
        self.doc = None
        if doc:
            self.doc = uncommentDocString(doc)

    def __repr__(self):
        # dump our contents to human readable form
        r = "INTERFACE %s" % self.name
        if self.extends:
            r += " EXTENDS %s" % self.extends
        r += '\n'

        if self.members:
            r += "Members:\n"
            for m in self.members:
                r += "    var %s line %s\n"  % (m, self.members[m])

        if self.functions:            
            r += "functions:\n"
            for f in self.functions.values():
                r += "    %r" % f

        if self.variables:
            r += "variables:\n"
            for v in self.variables.values():
                r += "    %r" % v
            
        return r + '\n'

    def toXML(self, level=0, indent="    "):
        attrs = {'name': self.name,
                 'line': self.linestart,
                 'lineend': self.lineend}
        signature = "interface "+self.name
        if self.extends:
            signature += " extends "+self.extends
        attrs["signature"] = signature
        out = '%s<interface%s>\n' % (indent*level, _getAttrStr(attrs))

        if self.doc:
            # XXX parse the docstring
            out += '%s<doc><![CDATA[%s]]></doc>\n'\
                   % (indent*(level+1), self.doc)

        if self.extends:
            out += '%s<interfaceref%s/>\n'\
                   % (indent*(level+1), _getAttrStr({'name': self.extends}))
        
        for f in self.functions.values():
            out += f.toXML(level+1, indent)

        for m in self.members.values():
            out += m.toXML(level+1, indent)

        for v in self.variables.values():
            out += v.toXML(level+1, indent)
        out += '%s</interface>\n' % (indent*level)
        return out
    
class PHPClass:
    def __init__(self, name, extends, lineno, depth, attributes=None,
                 interfaces=None, doc=None):
        self.name = name
        self.extends = extends
        self.linestart = lineno
        self.lineend = 0
        self.depth = depth
        self.members = {} # declared class variables
        self.variables = {} # all variables used in class
        self.functions = {}
        if interfaces:
            self.interfaces = interfaces.split(',')
        else:
            self.interfaces = []
        if attributes:
            self.attributes = ' '.join(attributes.strip().split())
        else:
            self.attributes = None
        self.doc = None
        if doc:
            self.doc = uncommentDocString(doc)

    def __repr__(self):
        # dump our contents to human readable form
        r = "CLASS %s" % self.name
        if self.extends:
            r += " EXTENDS %s" % self.extends
        r += '\n'

        if self.members:
            r += "Members:\n"
            for m in self.members:
                r += "    var %s line %s\n"  % (m, self.members[m])

        if self.functions:            
            r += "functions:\n"
            for f in self.functions.values():
                r += "    %r" % f

        if self.variables:
            r += "variables:\n"
            for v in self.variables.values():
                r += "    %r" % v
            
        return r + '\n'

    def toXML(self, level=0, indent="    "):
        attrs = {'name': self.name,
                 'line': self.linestart,
                 'lineend': self.lineend}
        signature = "class "+self.name
        if self.extends:
            signature += " extends "+self.extends
        if self.interfaces:
            signature += " implements "
            interfaces = [i.strip() for i in self.interfaces]
            signature += ", ".join(interfaces)
        attrs["signature"] = signature
        out = '%s<class%s>\n' % (indent*level, _getAttrStr(attrs))
        
        if self.doc:
            # XXX parse the docstring
            out += '%s<doc><![CDATA[%s]]></doc>\n'\
                   % (indent*(level+1), self.doc)

        if self.extends:
            out += '%s<classref%s/>\n'\
                   % (indent*(level+1), _getAttrStr({'name': self.extends}))
        
        for i in self.interfaces:
            out += '%s<interfaceref%s/>\n'\
                   % (indent*(level+1), _getAttrStr({'name': i.strip()}))

        for f in self.functions.values():
            out += f.toXML(level+1, indent)

        for m in self.members.values():
            out += m.toXML(level+1, indent)

        for v in self.variables.values():
            out += v.toXML(level+1, indent)
        out += '%s</class>\n' % (indent*level)
        return out


class PHPFile:
    """CIX specifies that a <file> tag have zero or more <module> children.
    In PHP this is a one-to-one relationship, so this class represents both
    (and emits the XML tags for both).
    """
    def __init__(self, filename, content=None, md5sum=None, mtime=None,
                 include_path=None):
        self.filename = filename
        self.content = content
        self.md5sum = md5sum
        self.mtime = mtime
        self.basedir = os.curdir
        self.error = None
        
        if content is None:
            self.content = self.read()
        if content:
            # something always gets returned
            content = _getEncodedBuffer(content)[0]
        
        if md5sum is None and self.content:
            self.md5sum = md5.new(self.content).hexdigest()
        if mtime is None:
            self.mtime = int(time.time())

        if self.filename != '_global_':
            self.basedir = os.path.dirname(self.filename) or os.curdir

        self.functions = {} # functions declared in file
        self.classes = {} # classes declared in file
        self.variables = {} # all variables used in file
        self.includes = {} # files included into this file
        self.interfaces = {} # interfaces declared in file

    def read(self):
        if self.filename == '_global_':
            return None
        try:
            f = open(self.filename, "r")
        except Exception, e:
            self.error = str(e)
            return
        try:
            return f.read()
        finally:
            f.close()
    
    def __repr__(self):
        # dump our contents to human readable form
        r = "FILE %s\n" % self.filename

        for f, l in self.includes.items():
            r += "include %s on line %d\n" % (f, l)

        r += "functions:\n"
        for f in self.functions.values():
            r += "    %r" % f

        r += "variables:\n"
        for v in self.variables.values():
            r += "    %r" % v

        r += "classes:\n"
        for c in self.classes.values():
            r += repr(c)

        return r + '\n'
    
    def toXML(self, level=0, indent="    "):
        fileAttrs = {"language": "PHP",
                     "generator": "PHP",
                     "md5": self.md5sum,
                     "mtime": self.mtime,
                     "path": self.filename.replace('\\', '/')}
        if self.error:
            fileAttrs["error"] = self.error
        out = '%s<file%s>\n' % (indent*level, _getAttrStr(fileAttrs))
        if self.error:
            return out+'</file>\n'
        
        moduleAttrs = {"name": os.path.basename(self.filename)}
        out += '%s<module%s>\n'\
               % (indent*(level+1), _getAttrStr(moduleAttrs))
        for fn in self.includes:
            attrs = {'module': fn, 'line': self.includes[fn]}
            out += '%s<import%s/>\n' % (indent*(level+2), _getAttrStr(attrs))

        for f in self.functions.values():
            out += f.toXML(level+2, indent)

        for v in self.variables.values():
            out += v.toXML(level+2, indent)

        for i in self.interfaces.values():
            out += i.toXML(level+2, indent)

        for c in self.classes.values():
            out += c.toXML(level+2, indent)

        out += '%s</module>\n' % (indent*(level+1))
        out += '%s</file>\n' % (indent*level)
        return out



class recollector:
    def __init__(self):
        self.res = {}
        self.regs = {}
        
    def add(self, name, reg, mods=None ):
        self.regs[name] = reg % self.regs
        #print "%s = %s" % (name, self.regs[name])
        if mods:
            self.res[name] = re.compile(self.regs[name], mods) # check that it is valid
        else:
            self.res[name] = re.compile(self.regs[name]) # check that it is valid

        
collector = recollector()
a = collector.add

a("newlines" ,          "[ \t]*(\r\n|\r|\n)")
a("comment_line" ,      "(//.*|#.*)")
a("comment_block" ,     r'(?P<doc>/\*.*?\*/)', re.S|re.M)
if sys.version_info[:2] == (2,0):
    a("string" ,        r'^(?<=(?<!\\)"|\').*?((?<!\\)\1)', re.S|re.M)
else:
    a("string" ,        r'((?<!\\)"|\').*?((?<!\\)\1)', re.M)
a("string_delim" ,      r'((?:(?<!\\)"|\'))')
a("ppps",               r'public\s+|protected\s+|private\s+|static\s+')
a("classtypes",         r'abstract\s+|final\s')
a("vartypes",           r'var\s*|%(ppps)s',re.I)
a("classvar" ,          r'(?:\s*\$(?P<name>[\w_]*?)\s*=\s*&?\s*new\s*(?P<classname>[\w_]*?)(?:\(.*?\)|\s*)\s*);',re.I|re.M)
a("variable" ,          r'(?:\s*\$(?P<name>[\w_]*?)(?:(?:\s*=.*?)|\s*));',re.I|re.M)
a("include" ,           r'\s*(?:require|include|require_once|include_once)\s*[\(|\s]*[\"|\']\s*(.*?)\s*[\"|\'][\)|\s]*;',re.I)
a("block_begin" ,       "{")
a("block_end" ,         "}")
a("classfunctypes",     r'%(classtypes)s|%(ppps)s',re.I)
a("function" ,          r'(?:%(comment_block)s?\s*(?P<attributes>(?:%(classfunctypes)s|\s)*)function\s+(?:[&]*?)\s*(?P<name>[\w_]*?)\s*\((?P<args>.*?)\))',re.I|re.S|re.M)
a("class" ,             r'(?:%(comment_block)s?\s*(?P<attributes>(?:%(classtypes)s)*)(?<![$])class\s+(?P<name>[\w_]*)\s*(?:(?:extends\s+(?P<extends>[\w_]*)\s*)|\s*)\s*(?:(?:implements\s+(?P<interfaces>[\w_,\s]*)\s*)|\s*))',re.I|re.S|re.M)
a("classmember" ,       r'(?:(?P<attributes>(?:%(vartypes)s)*)\$(?P<name>[\w_]*?)\s*[=;])',re.I)
a("interface" ,         r'(?:%(comment_block)s?\s*(?<![$])interface\s+(?P<name>[\w_]*)\s*(?:(?:extends\s+(?P<extends>[\w_]*)\s*)|\s*))',re.I|re.S|re.M)
a("php_start" ,         "<\?php|<\?")
a("php_end" ,           "\?>")
a("label" ,             "([a-zA-Z_][a-zA-Z0-9_]*)")
a("lnum" ,              "([0-9]+)")
a("whitespace" ,        "[ \t]+")
a("any_char" ,          "(.)") # |[\n]
a("args_info",          r'(?P<tip>[\w_]*?)\s*\$(?P<arg>[\w_]*)(?:\s*=\s*(?P<default>[\w|%(string)s]*)\s*(?:,)?)?',re.S|re.M)

SKIPTOK = 0x01 # don't consider this a token that is to be considered a part of the grammar, like '\n'
MAPTOK = 0x02  # use the token associated with the pattern when it matches
EXECFN= 0x04   # execute the function associated with the pattern when it matches
USETXT = 0x08  # if you match a single character and want its ascii value to be the token

# Lexer class borrowed from the PyLRd project,
# http://starship.python.net/crew/scott/PyLR.html
class Lexer:
    eof = -1

    def __init__(self):
        self.tokenmap = {}
        self.prgmap = {}
        self.prglist = []
        self.lasttok = -1
        self.text = ""
        self.textindex = 0
        self.tokennum2name = {}

    def nexttok(self):
        self.lasttok = self.lasttok + 1
        return self.lasttok

    def settext(self, t):
        self.text = t
        self.textindex = 0

    def addmatch(self, prg, func=None, tokname="", attributes=MAPTOK|EXECFN):
        self.prglist.append(prg)
        tok = -2
        if not func:
            attributes = attributes & ~EXECFN
        if not tokname:
            attributes = attributes & ~MAPTOK
        if attributes & MAPTOK:
            self.tokenmap[tokname] = tok = self.nexttok()
        else:
            tok = self.nexttok()
        self.prgmap[prg] = tok, attributes, func
        self.tokennum2name[tok] = tokname

    def scan(self):
        for prg in self.prglist:
            mo = prg.match(self.text, self.textindex)
            if not mo: 
                continue
            self.textindex = self.textindex + len(mo.group(0))
            tmpres = mo.group(0)
            t, attributes, fn = self.prgmap[prg]
            #log.debug("'%s' token: %r", self.tokennum2name[t], tmpres)
            if attributes & EXECFN:
                tmpres = apply(fn, (mo,))
            if attributes & USETXT:
                t = ord(mo.group(0)[0])
            return (t, tmpres)
        if self.textindex >= len(self.text):
            return (self.eof, "")
        raise PHPCILEError("Syntax Error in lexer")


class PHPLexer:
    def __init__(self, filename, content=None, md5sum=None, mtime=None,
                 include=0, include_path="", cache=None):
        # hook up the lexical matches to a function that handles the token
        self.lex_matches = [
            ('newline',         collector.res["newlines"],        self.incline,           EXECFN|SKIPTOK),    
            ('comment_line',    collector.res["comment_line"],    None,                   SKIPTOK),
            ('comment_block',   collector.res["comment_block"],   self.doCommentBlock,    EXECFN|SKIPTOK),
            ('include',         collector.res["include"],         self.include_file,      EXECFN|SKIPTOK),
            ('string',          collector.res["string"],          self.doMultiLineBlock,  EXECFN|SKIPTOK),
            #('string',          collector.res["string_delim,    self.toggleStringState, EXECFN|SKIPTOK),
            ('block_begin',     collector.res["block_begin"],     self.incBlock,          EXECFN|SKIPTOK),    
            ('block_end',       collector.res["block_end"],       self.decBlock,          EXECFN|SKIPTOK),    
            ('classvar',        collector.res["classvar"],        self.addClassVar,       EXECFN|SKIPTOK),
            ('variable',        collector.res["variable"],        self.addVariable,       EXECFN|SKIPTOK),
            ('function',        collector.res["function"],        self.addFunction,       EXECFN|SKIPTOK),
            ('class',           collector.res["class"],           self.addClass,          EXECFN|SKIPTOK),
            ('classmember',     collector.res["classmember"],     self.addClassMember,    EXECFN|SKIPTOK),
            ('interface',       collector.res["interface"],       self.addInterface,      EXECFN|SKIPTOK),
            ('php_end',         collector.res["php_end"],         self.endCodeState,   EXECFN|SKIPTOK),
            ('php_start',       collector.res["php_start"],       self.startCodeState,   EXECFN|SKIPTOK),
            ('label',           collector.res["label"],           None,                   0),    
            ('lnum',            collector.res["lnum"],            None,                   SKIPTOK),    
            ('whitespace',      collector.res["whitespace"],      None,                   SKIPTOK),    
            ('any_char',        collector.res["any_char"],        None,                   SKIPTOK),    
        ]

        # setup lexer and add token matching regexes
        self.l = Lexer()
        for p in self.lex_matches:
            # log.debug("adding %r",p[0])
            attributes = p[3]
            if not attributes: attributes = MAPTOK|EXECFN
            self.l.addmatch(p[1],p[2],p[0],attributes)

        self.lineno = 1
        self.phpcode = 0
        self._stringDelim = ""
        self.depth = 0
        self.classStack = []
        self.currentClass = None
        self.currentInterface = None
        self.currentFunction = None
        self.include_path = include_path
        self.lastComment = None
        
        self.basedir = os.path.dirname(filename)
        self.filename = self.findfile(filename)
        
        if cache:
            self.cile = cache
            self.cile.clear(self.filename)
        
        else:
            self.cile = PHPcile()
        
        self.fileinfo = PHPFile(self.filename, content, md5sum, mtime)
        self.doincludes = include
        if not self.fileinfo.error:
            self.scanData()
        self.cile.filesparsed[self.filename] = self.fileinfo

    def idfunc(self, m):
        if not self.phpcode: return
        log.info("ID: %r",m.group(0))
        return m.group(0)

    def incline(self, m):
        self.lineno = self.lineno + 1
        #log.debug("line: %d",self.lineno)
        return ""

    def toggleStringState(self, m):
        if self.phpcode:
            log.info("toggleStringState exiting code fragment at line %d", self.lineno)
            self.phpcode = 0
            self._stringDelim = m.group(0)
        elif self._stringDelim == m.group(0):
            log.info("toggleStringState entering code fragment at line %d", self.lineno)
            self.phpcode = 1
        return ""
        
    def startCodeState(self, m):
        if not self.phpcode:
            log.info("startCodeState entering code fragment at line %d", self.lineno)
            self.phpcode = 1
        return ""

    def endCodeState(self, m):
        if self.phpcode:
            log.info("endCodeState exiting code fragment at line %d", self.lineno)
            self.phpcode = 0
        return ""

    # parses included files
    def include_file(self, m):
        if not self.phpcode: return
        filename = self.findfile(m.group(1))

        # XXX Very simple prevention of include looping.  Really should
        # recurse the indices to make sure we are not creating a loop
        if self.filename == filename: return ""
        
        # parse the included file
        if self.doincludes and filename not in self.cile.filesparsed:
            # build new include path
            if self.include_path:
                path = string.join([self.basedir, self.include_path], os.pathsep)
            else:
                path = self.basedir
            
            log.info("PHPLexer(%s, %d, %r)", filename, self.doincludes, path)
            PHPLexer(filename, include=self.doincludes, include_path=path, cache=self.cile)
            if filename not in self.cile.filesparsed:
                return ""
        
        # add the included file to our list of included files
        if filename not in self.fileinfo.includes:
            self.fileinfo.includes[filename] = self.lineno

        # add the included file to our list of included files
        if self.filename not in self.cile.needfile:
            self.cile.needfile[self.filename] = []
        try:
            self.cile.needfile[self.filename].index(filename)
        except ValueError, e:
            self.cile.needfile[self.filename].append(filename)
            
        # add this file to the infile list
        if filename not in self.cile.infile:
            self.cile.infile[filename] = []
        try:
            self.cile.infile[filename].index(self.filename)
        except ValueError, e:
            self.cile.infile[filename].append(self.filename)
            
        return ""

    def doCommentBlock(self, m):
        self.lastComment = [m, self.lineno, 0]
        self.doMultiLineBlock(m)
        self.lastComment[2] = self.lineno
        log.info("doCommentBlock %r", self.lastComment)

    def getLastComment(self):
        doc = None
        if self.lastComment and self.lastComment[2] > self.lineno - 5:
            doc = self.lastComment[0].group('doc')
        return doc
        
    # takes a block, and figures out how many lines it spans, to
    # keep lineno correct
    def doMultiLineBlock(self, m):
        if not self.phpcode: return
        nl = collector.res["newlines"].findall(m.group(0))
        #log.debug("block start at lineno %d",self.lineno)
        blockLen = len(nl)
        self.lineno = self.lineno + blockLen
        #log.debug("block had %d lines, lineno is %d", blockLen, self.lineno)

    def incBlock(self, m):
        if not self.phpcode: return
        self.depth = self.depth+1
        # log.debug("depth at %d", self.depth)
        return ""

    def decBlock(self, m):
        if not self.phpcode: return
        self.depth = self.depth-1
        # log.debug("depth at %d", self.depth)
        if self.currentClass and self.currentClass.depth == self.depth:
            # log.debug("done with class %s at depth %d", self.currentClass.name, self.depth)
            self.currentClass.lineend = self.lineno
            self.currentClass = self.classStack.pop()
        if self.currentInterface and self.currentInterface.depth == self.depth:
            # log.debug("done with interface %s at depth %d", self.currentClass.name, self.depth)
            self.currentInterface.lineend = self.lineno
            self.currentInterface = None
        elif self.currentFunction and self.currentFunction.depth == self.depth:
            self.currentFunction.lineend = self.lineno
            # XXX stacked functions used to work in php, need verify still is
            self.currentFunction = None
        self.lastComment = None
        return ""

    def addFunction(self, m):
        if not self.phpcode: return
        log.info("FUNC: %s(%s) on line %d", m.group('name'), m.group('args'), self.lineno)
        self.doMultiLineBlock(m)
        doc = m.group('doc')
        if not doc:
            doc = self.getLastComment()
        classname=''
        if self.currentClass:
            classname = self.currentClass.name
        elif self.currentInterface:
            classname = self.currentInterface.name
        attributes=m.groupdict().get('attributes')
        self.currentFunction = PHPFunction(m.group('name'),
                                           m.group('args'),
                                           self.lineno,
                                           self.depth,
                                           attributes=attributes,
                                           doc=doc,
                                           classname=classname)
        if self.currentClass:
            self.currentClass.functions[self.currentFunction.name] = self.currentFunction
        elif self.currentInterface:
            self.currentInterface.functions[self.currentFunction.name] = self.currentFunction
        else:
            self.fileinfo.functions[self.currentFunction.name] = self.currentFunction
            self.cile.functionindex[self.currentFunction.name] = self.fileinfo.filename
        if self.currentInterface or attributes.find('abstract') >= 0:
            self.currentFunction.lineend = self.lineno
            self.currentFunction = None
        return ""

    def addClass(self, m):
        if not self.phpcode: return
        if m.group('name') not in self.fileinfo.classes:
            doc = m.group('doc')
            if not doc:
                doc = self.getLastComment()
            # push the current class onto the class stack
            self.classStack.append(self.currentClass)
            # make this class the current class
            self.currentClass = PHPClass(m.group('name'),
                                         m.group('extends'),
                                         self.lineno,
                                         self.depth,
                                         attributes=m.group('attributes'),
                                         interfaces=m.group('interfaces'),
                                         doc=doc)
            self.fileinfo.classes[self.currentClass.name]=self.currentClass
            # log.debug("adding classindex[%s]=%s", m.group('name'), self.fileinfo.filename)
            self.cile.classindex[self.currentClass.name]=self.fileinfo.filename
            log.info("CLASS: %s extends %s interfaces %s attributes %s on line %d in %s at depth %d\nDOCS: %s",
                     self.currentClass.name, self.currentClass.extends, 
                     self.currentClass.interfaces, self.currentClass.attributes,
                     self.currentClass.linestart, self.filename, self.depth,
                     self.currentClass.doc)
        else:
            # shouldn't ever get here
            pass
        self.doMultiLineBlock(m)
        return ""
    
    def addClassMember(self, m):
        if not self.phpcode: return
        if self.currentFunction:
            if m.group('name') not in self.currentFunction.variables:
                self.currentFunction.variables[m.group('name')] = PHPVariable(m.group('name'), self.lineno)
        elif self.currentClass:
            if m.group('name') not in self.currentClass.members:
                self.currentClass.members[m.group('name')] =\
                    PHPVariable(m.group('name'),
                                self.lineno,
                                attributes=m.groupdict().get('attributes'))
            log.info("CLASSMBR: %r", m.group('name'))
        self.doMultiLineBlock(m)
        return ""
    
    def addClassVar(self, m):
        if not self.phpcode: return
        log.info("CLASSVAR: %s %s on line %d", m.group('name'), m.group('classname'), self.lineno)
        if m.group('name') not in self.fileinfo.variables:
            self.fileinfo.variables[m.group('name')] = PHPVariable(m.group('name'), self.lineno, m.group('classname'))
        else:
            self.fileinfo.variables[m.group('name')].addType(self.lineno, m.group('classname'))
        self.doMultiLineBlock(m)
        return ""

    def addInterface(self, m):
        if not self.phpcode: return
        if m.group('name') not in self.fileinfo.classes:
            # push the current class onto the class stack
            self.classStack.append(self.currentClass)
            # make this class the current class
            self.currentInterface = PHPInterface(m.group('name'),m.group('extends'), self.lineno, self.depth)
            self.fileinfo.interfaces[m.group('name')]=self.currentInterface
            # log.debug("adding classindex[%s]=%s", m.group('name'), self.fileinfo.filename)
            self.cile.interfaceindex[m.group('name')]=self.fileinfo.filename
            log.info("INTERFACE: %s extends %s on line %d in %s at depth %d",
                     m.group('name'), m.group('extends'),
                     self.lineno, self.filename, self.depth)
        else:
            # shouldn't ever get here
            pass
        self.doMultiLineBlock(m)
        return ""

    def addVariable(self, m):
        if not self.phpcode: return
        log.info("VAR: %s on line %d", m.group('name'), self.lineno)
        if self.currentFunction:
            if m.group('name') not in self.currentFunction.variables:
                self.currentFunction.variables[m.group('name')] = PHPVariable(m.group('name'), self.lineno)
        elif self.currentClass:
            pass
            # XXX this variable is local to a class method, what to do with it?
            #if m.group('name') not in self.currentClass.variables:
            #    self.currentClass.variables[m.group('name')] =\
            #        PHPVariable(m.group('name'), self.lineno)
        else:
            if m.group('name') not in self.fileinfo.variables:
                self.fileinfo.variables[m.group('name')] = PHPVariable(m.group('name'), self.lineno)
        self.doMultiLineBlock(m)
        return ""

    def findfile(self, filename):
        if not os.path.exists(filename):
            # try the basedir for a relative include
            if self.basedir:
                fn = os.path.join(self.basedir, filename)
                fn = os.path.normpath(fn)
                if os.path.exists(fn):
                    return fn
            
            # try the include paths
            if self.include_path:
                paths = self.include_path.split(os.pathsep)
                for path in paths:
                    fn = os.path.join(path, filename)
                    fn = os.path.normpath(fn)
                    if os.path.exists(fn):
                        return fn
                
        return filename
    
    def scanData(self):
        # log.debug("doing: %s",data)
        self.l.settext(self.fileinfo.content)
            
        #res = []
        while 1:
            t, v = self.l.scan()
            # log.debug("lex symbol: %r %r", t, v)
            if t == self.l.eof:
                break
            #res.append((t, v))



#---- public module interface

def scan(content, filename, md5sum=None, mtime=None, includePath=None,
         useRepr=0, recursive=0):
    """Scan the given PHP content and return Code Intelligence data
    conforming the the Code Intelligence XML format.
    
        "content" is the PHP content to scan
        "filename" is the source of the PHP content (used in the
            generated output).
        "md5sum" (optional) if the MD5 hexdigest has already been calculated
            for the content, it can be passed in here. Otherwise this
            is calculated.
        "mtime" (optional) is a modified time for the file (in seconds since
            the "epoch"). If it is not specified the _current_ time is used.
            Note that the default is not to stat() the file and use that
            because the given content might not reflect the saved file state.
        XXX includePath, useRepr, recursive
    
    This will raise a PHPCILEError if there was an error processing.
    """
    lex = PHPLexer(filename, content, md5sum, mtime, recursive, includePath)
    if useRepr:
        return repr(lex.cile)
    return lex.cile.toXML()


#---- mainline

def main(argv):
    logging.basicConfig()

    # Parse options.
    useRepr = 0
    includePath = None
    recursive = 0
    doTiming = 0
    stdinFilename = None
    md5sum = None
    mtime = None
    try:
        opts, args = getopt.getopt(argv[1:], "Vvhf:ti:rc",
            ["version", "verbose", "help", "filename=", "md5=", "mtime=",
             "text", "include_path", "recursive", "clock"])
    except getopt.GetoptError, ex:
        log.error(str(ex))
        log.error("Try `phpcile --help'.")
        return 1
    for opt, optarg in opts:
        if opt in ("-h", "--help"):
            sys.stdout.write(__doc__)
            return
        elif opt in ("-V", "--version"):
            ver = '.'.join([str(part) for part in _version_])
            print "phpcile %s" % ver
            return
        elif opt in ("-v", "--verbose"):
            log.setLevel(logging.DEBUG)
        elif opt in ("-f", "--filename"):
            stdinFilename = optarg
        elif opt in ("--md5",):
            md5sum = optarg
        elif opt in ("--mtime",):
            mtime = optarg
        elif opt in ("-t", "--text"):
            useRepr = 1
        elif opt in ("-i", "--include_path"):
            recursive = optarg
        elif opt in ("-r", "--recursive"):
            doInclude = 1
        elif opt in ("-c", "--clock"):
            doTiming = 1

    if len(args) == 0:
        contentOnStdin = 1
        filenames = [stdinFilename or "<stdin>"]
    else:
        contentOnStdin = 0
        filenames = []
        for arg in args:
            filenames += glob.glob(arg)

    try:
        for filename in filenames:
            if contentOnStdin:
                log.debug("reading content from stdin")
                content = sys.stdin.read()
                log.debug("finished reading content from stdin")
                if mtime is None:
                    mtime = int(time.time())
            else:
                if mtime is None:
                    mtime = int(os.stat(filename)[stat.ST_MTIME])
                fin = open(filename, 'r')
                try:
                    content = fin.read()
                finally:
                    fin.close()

            if doTiming: start = time.clock()
            data = scan(content, filename, md5sum=md5sum, mtime=mtime,
                        useRepr=useRepr, recursive=recursive,
                        includePath=includePath)
            if doTiming: end = time.clock()
            if data:
                sys.stdout.write(data)
            if doTiming:
                print "time to parse '%s': %fs" % (filename, end-start)
    except PHPCILEError, ex:
        log.error(str(ex))
        if log.isEnabledFor(logging.DEBUG):
            print
            import traceback
            traceback.print_exception(*sys.exc_info())
        return 1
    except KeyboardInterrupt:
        log.debug("user abort")
        return 1


if __name__ == "__main__":
    sys.exit( main(sys.argv) )

