# Copyright (c) 2005 ActiveState Corp, a division of Sophos Plc.
#
# Contributors:
#   Eric Promislow (EricP@ActiveState.com)

"""
Ruby lexing support for codeintel/rubycile.py

A two-layer approach for getting lexed tokens from SilverCity.
The worker (Producer) thread receives tokens from SilverCity's lexer and
puts them in a queue.  The main thread, working in the Supplier class,
pulls items off the queue and returns them  to the caller.

Usage:
import lex_wrapper
lexer = lex_wrapper.Lexer(code)
while 1:
    tok = lexer.get_next_token()
    if tok[0] == EOF_STYLE:
        break;
    # tok is an array of (style, text, start-col, start-line, end-col, end-line)
    # column and line numbers are all zero-based.
"""

import threading
import Queue
import re
import sys
import string

import SilverCity
from SilverCity import Ruby, ScintillaConstants


MAX_TOKEN_Q_SIZE=100

class Token:
    def __init__(self, style, text="", start_column=None, start_line=None, end_column=None, end_line=None):
        self.style = style
        self.text = text
        self.start_column = start_column
        self.start_line = start_line
        self.end_column = end_column
        self.end_line = end_line
        
    # Hardwired cloner for convenience
    def clone(tok, text):
        return Token(tok.style, text, tok.start_column, tok.start_line, tok.end_column, tok.end_line)
        
    def dump(self):
        s = "[" + str(self.style) + ", " + repr(self.text)
        for attr in ['start_column', 'start_line', 'end_column', 'end_line']:
            if hasattr(self, attr):
                val = getattr(self, attr)
                s += ", " + attr + "=" + str(val)
        s += "]"
        print s

EOF_STYLE = -1
EOF_TOKEN = Token(EOF_STYLE)

class Producer(threading.Thread):
    def __init__(self, q, code):
        self.q = q
        self.code = code
        threading.Thread.__init__(self)
        
    
        
    def make_callback(self):
        def adapt_line(line_num):
            if line_num is None:
                return None
            return line_num + 1
        def add_token(style, text, start_column, start_line, **other_args):
            self.q.put(Token(style, text, start_column, adapt_line(start_line), other_args.get('end_column',None), adapt_line(other_args.get('end_line', None))))
        return add_token
        
    def run(self):
        Ruby.RubyLexer().tokenize_by_style(self.code, self.make_callback())
        self.q.put(EOF_TOKEN)

def build_dict(ws_sep_str):
    the_dict = {}
    for the_key in ws_sep_str.split():
        the_dict[the_key] = 1
    return the_dict

ws_re = re.compile("^[" + string.whitespace + "\\\\" + "]*$")
trailing_spaces_re = re.compile("\n([ \t]*)$")

def contains_nl(str2):
    return str2.find("\n") >= 0

class Signature:
    def __init__(self):
        self._gathering = False
        self._text = ""

    def open(self):
        self._gathering = True
        self._text = ""
        
    def close(self):
        self._gathering = False

    def text(self):
        return self._text

    def append(self, text):
        self._text += text
        
    def replace(self, text):
        self._text = text

    def is_gathering(self):
        return self._gathering

string_types = [ScintillaConstants.SCE_RB_STRING,
                ScintillaConstants.SCE_RB_CHARACTER,
                ScintillaConstants.SCE_RB_STRING_Q,
                ScintillaConstants.SCE_RB_STRING_QQ,
                ScintillaConstants.SCE_RB_STRING_QX,
                ScintillaConstants.SCE_RB_STRING_QR,
                ScintillaConstants.SCE_RB_STRING_QW
                ]

def is_string_token(tok):
    return tok.style in string_types

class Lexer:
    def __init__(self, code):
        self.q = Queue.Queue(MAX_TOKEN_Q_SIZE)
        self.token_producer = Producer(self.q, code)
        self.token_producer.start()
        self.gen = self._get_next_token()
        self.pending_tokens = []
        self.multi_char_ops = build_dict('!= !~ && ** :: <= << == => =~ >> ||')
        self.curr_indentation = 0
        self.curr_comments = []
        self.finished_comment = True
        self.use_leading_spaces = True
        self.signature = Signature()
    
    def __del__(self):
        # Wrapped in case the object wasn't completely constructed.
        try:
            self.token_producer.join(1.0) # seconds
        except:
            pass
        
    def _get_next_token(self):
        while 1:
            yield self.q.get()
    
    def _get_eof_token(self):
        while 1:
            yield EOF_TOKEN
            
    def matches_whitespace(self, text):
        return ws_re.match(text)

    def get_curr_indentation(self):
        return self.curr_indentation

    def curr_comment(self, destroy=1):
        hold = self.curr_comments
        self.curr_comments = []
        return hold

    def clear_comments(self):
        self.curr_comments = []

    def has_comment(self):
        return len(self.curr_comments) > 0

    # Methods for manipulating signatures

    def start_sig(self):
        self.signature.open()

    def stop_sig(self):
        self.signature.close()

    def get_sig(self):
        return self.signature.text()
            
    def get_next_token(self, skip_ws=1):
        if len(self.pending_tokens) > 0:
            tok = self.pending_tokens.pop()
            gather = self.signature.is_gathering() and not hasattr(tok, "generated")
        else:
            gather = self.signature.is_gathering()
            tok = self.gen.next()
        if tok.style == EOF_STYLE:
            # Stop leaning on the queue, just return an eof_token
            self.gen = self._get_eof_token()
            self.curr_indentation = 0
            
        elif tok.style == ScintillaConstants.SCE_RB_COMMENTLINE:
            if self.finished_comment:
                self.curr_comments = []
                self.finished_comment = False
            self.curr_comments.append(tok.text)
            self.use_leading_spaces = False
            if skip_ws:
                return self.get_next_token(1)
            else:
                return tok
            
        elif tok.style == ScintillaConstants.SCE_RB_DEFAULT and self.matches_whitespace(tok.text):
            if gather:
                self.signature.append(tok.text)
            has_nl = contains_nl(tok.text)
            if has_nl or self.use_leading_spaces:
                # Update this line's indentation only if we're at the start
                if has_nl:
                    try:
                        self.curr_indentation = len(trailing_spaces_re.findall(tok.text)[-1])
                    except:
                        self.curr_indentation = 0
                else:
                    self.curr_indentation = len(tok.text)
                # Do we still need to count white-space in subsequent tokens?
                self.use_leading_spaces = (self.curr_indentation == 0)
            if skip_ws:
                return self.get_next_token(1)
            else:
                return tok
        else:
            # At this point we're done with comments and leading white-space
            if gather:
                self.signature.append(tok.text)
            self.finished_comment = True
            self.use_leading_spaces = False

            if tok.style == ScintillaConstants.SCE_RB_OPERATOR and len(tok.text) > 1:
                # Python/Ruby lexer seems sloppy, returning things like (SCE_RB_OPERATOR, "),")
                if not self.multi_char_ops.has_key(tok.text):
                    first_char, rest_str = tok.text[0], tok.text[1:]
                    new_tok = tok.clone(rest_str)
                    tok.text = first_char
                    tok.end_column = tok.start_column
                    new_tok.start_column = tok.end_column + 1
                    self.put_back(new_tok)
            
        return tok
    
    def put_back(self, tok):
        if self.signature.is_gathering():
            sig = self.signature.text()
            last_text = tok.text
            if sig[-len(last_text):] == last_text:
                sig = sig[0:-len(last_text)]
                self.signature.replace(sig)
        self.pending_tokens.append(tok)

def provide_sample_code():
    return r"""require 'rdoc/parsers/parse_rb.rb'

# full-line comment
# comment at start of line
 # comment at col 1
  # comment at col 2

module Generators
  class XMLGenerator < HTMLGenerator # a comment here
    def our_generate(file_info)
      @info       = info
      @files      = []
      @classes    = []
      @hyperlinks = {}
    end
    # comment on what test_fn does
    # more...

    def test_fn(a, b='val1', c=f(3), *d, &e)
       print "nothing\n"   # end-of-line comment
       print %Q(percent string\n)
    end

    def no_paren a, b, \
           c
       print "blah"
    end
  end
end
"""

def read_and_detab(fs, closefd=False, tabwidth=8):
    sample_code = ""
    try:
        # Decompress tabs so our indentation calculations work
        lines = []
        for line in fs:
            lines.append(line.expandtabs(tabwidth))
        sample_code = "".join(lines)
    finally:
        if closefd:
            fs.close()
    return sample_code

if __name__ == "__main__":
    if len(sys.argv) == 1:
        sample_code = provide_sample_code();
        fs = None
    elif sys.argv[1] == "-":
        fs = sys.stdin
    else:
        fs = open(sys.argv[1], "r")
    if fs is not None:
        sample_code = read_and_detab(fs)
        # fs comes back closed
    
    try:
        lexer_wrapper = Lexer(sample_code)
        last_line = -1
        while 1:
            tok = lexer_wrapper.get_next_token(1)
            if tok.style == EOF_STYLE:
                break
            if last_line != tok.start_line:
                print "[%d:%d] " % (tok.start_line, lexer_wrapper.curr_indentation),
                last_line = tok.start_line
                if lexer_wrapper.has_comment():
                    comments = lexer_wrapper.curr_comment(1)
                    print comments
            tok.dump()
    except:
        while lexer_wrapper.get_next_token().style != EOF_STYLE:
            pass
        raise # and reraise
