# Copyright (c) 2005 ActiveState Corp, a division of Sophos Plc.
#
# Contributors:
#   Eric Promislow (EricP@ActiveState.com)

"""Ruby parsing support for codeintel/rubycile.py"""

import string
import sys
import re
import textwrap

from SilverCity import ScintillaConstants
from codeintel import ruby_lexer


# Parse Ruby code

NODE_FILE = 0
NODE_MODULE = 0
NODE_CLASS = 0
NODE_METHOD = 0
NODE_ARGUMENT = 0

NameTokenTypes = [ScintillaConstants.SCE_RB_CLASSNAME,
                  ScintillaConstants.SCE_RB_DEFNAME,
                  ScintillaConstants.SCE_RB_MODULE_NAME,
                  ScintillaConstants.SCE_RB_IDENTIFIER,
                  ScintillaConstants.SCE_RB_WORD,
                  ScintillaConstants.SCE_RB_WORD_DEMOTED]

leading_hash_re = re.compile(r'^\s*\#+\s*')

class Name_LineNum:
    def __init__(self, name, line_num, type=None):
        self.line_num = line_num
        self.name = name
        self.type = type

class VarInfo:
    def __init__(self, line_num, type=None):
        self.line_num = line_num
        self.type = type

def update_collection(coll, name, line_num, type=None):
    if not coll.has_key(name):
        coll[name] = VarInfo(line_num, type)

class Node:
    def __init__(self, line_num, class_name=None):
        self.children = []
        self.line_num = line_num
        self.indentation = 0
        self.doc_lines = []
        self.imports = [] # require and load stmts
        self.includes = []  # include stmts

        self.class_vars = {}
        self.instance_vars = {}
        self.local_vars = {}
        if class_name: self.class_name = class_name
        
    def append_node(self, new_node):
        self.children.append(new_node)

    def set_line_end_num(self, line_end_num):
        self.lineend = line_end_num        
    
    def dump_kids(self, indent_level):
        for c in self.children:
            c.dump(indent_level + 2)
            
    def dump2(self, node_type_name, indent_level, *call_backs):
        print "%s %s %s - line %d" % (" " * indent_level, node_type_name, self.name, self.line_num)
        if len(self.doc_lines) > 0:
            print "%s Documentation: %s" % (" " * (indent_level + 2), "\n".join(self.doc_lines))
        if len(self.imports) > 0:
            for m in self.imports:
                print "%s Import: %s" % (" " * (indent_level + 2), m.name)
        if len(self.includes) > 0:
            for m in self.includes:
                print "%s Include: %s" % (" " * (indent_level + 2), m.name)
        self.dump_collection('Globals', 'global_vars', indent_level + 2)
        self.dump_collection('Locals', 'local_vars', indent_level + 2)
        self.dump_collection('Instance vars', 'instance_vars', indent_level + 2)
        self.dump_collection('Class vars', 'class_vars', indent_level + 2)
        for cb in call_backs:
            cb()
        self.dump_kids(indent_level)
        
    def dump_collection(self, label, attr, indent_level):
        if hasattr(self, attr):
            collection = getattr(self, attr)
            if len(collection) > 0:
                print "%s %s: " % (" " * indent_level, label)
                for name, varInfo in collection.items():
                    line, type = varInfo.line_num, varInfo.type
                    if type:
                        typeString = " [" + type + "]"
                    else:
                        typeString = ""
                    print "%s %s - line %s%s" % (" " * (indent_level + 2), name, line, typeString)
            
    
class ClassNode(Node):
    def __init__(self, the_name, line_num):
        self.name = the_name
        self.classrefs = []
        Node.__init__(self, line_num, "Class")
        
    def add_classrefs(self, class_ref_name, line_num, classref_type=None):
        if class_ref_name not in self.classrefs:
            self.classrefs.append([class_ref_name, line_num, classref_type])
        
    def dump(self, indent_level):
        def cb():
            classrefs = self.classrefs
            if len(classrefs) > 0:
                for ref in classrefs:
                    print "%sClassref %s - line %s" % (" " * (indent_level + 2),
                                                       ref[0], ref[1])
        self.dump2("Class", indent_level, cb)
    
class FileNode(Node):
    def __init__(self):
        Node.__init__(self, 0)
        self.global_vars = {}  # Easy to get at
        
    def dump(self):
        self.name = ""
        self.dump2("file", 0)
        return
        indent_level = 0
        if len(self.imports) > 0:
            for m in self.imports:
                print "%s Import: %s" % (" " * (indent_level + 2), m.name)
        if len(self.includes) > 0:
            for m in self.includes:
                print "%s Include: %s" % (" " * (indent_level + 2), m.name)
        self.dump_kids(0)
        
class ArgNode:
    def __init__(self, name, extra_info):
        self.name = name
        self.extra_info = extra_info
    def get_full_name(self):
        if self.extra_info:
            return self.extra_info + self.name
        return self.name

class MethodNode(Node):
    def __init__(self, the_name, line_num):
        self.name = the_name
        self.is_constructor = (the_name == "initialize")
        self.signature = ""
        self.args = []
        self.is_classmethod = False
        Node.__init__(self, line_num, "Method")
        
    def add_arg(self, name, extra_info):
        self.args.append(ArgNode(name, extra_info))
        
    def dump(self, indent_level):
        def cb():
            args = self.args
            for arg in args:
                print "%sArg %s" % (" " * (indent_level + 2), arg.get_full_name())
        self.dump2("Method", indent_level, cb)
        if len(self.signature) > 0:
            print "%sSignature %s" % (" " * (indent_level + 2), self.signature)

class ModuleNode(Node):
    def __init__(self, the_name, line_num):
        self.name = the_name
        Node.__init__(self, line_num, "Module")
        
    def dump(self, indent_level):
        self.dump2("Module", indent_level)

class VariableNode(Node):
    def __init__(self, the_name, line_num):
        self.name = the_name
        Node.__init__(self, line_num, "Variable")
        
    def dump(self, indent_level):
        self.dump2("Module", indent_level)

class BlockNode(Node):
    def __init__(self, the_name, line_num):
        self.name = the_name
        Node.__init__(self, line_num, "Block")
        
    def dump(self, indent_level):
        self.dump2("Module", indent_level)

# Unbound helper functions   

def is_identifier(tok, allow_keywords):
    return (tok.style == ScintillaConstants.SCE_RB_IDENTIFIER or
            (allow_keywords and
             tok.style in [ScintillaConstants.SCE_RB_WORD,
                           ScintillaConstants.SCE_RB_WORD_DEMOTED]))

#!!! So far this only deals with 'end', but watch out for
# the demoted keywords.

def is_keyword(tok, target):
    return tok.style == ScintillaConstants.SCE_RB_WORD and tok.text == target

def is_operator(tok, target):
    return tok.style == ScintillaConstants.SCE_RB_OPERATOR and tok.text == target

def remove_hashes(lines):
    return map(lambda s: leading_hash_re.sub("", s), lines)

class Parser:
    def __init__(self, tokenizer):
        self.tokenizer = tokenizer
        self.block_stack = []
        self.tree = FileNode()
        self.curr_node = self.tree
        self.bracket_matchers = {"[":"]", "{":"}", "(":")"}
        self.containers = {ScintillaConstants.SCE_RB_GLOBAL : [self.tree.global_vars],
                           ScintillaConstants.SCE_RB_CLASSNAME : [], #classes
                           ScintillaConstants.SCE_RB_CLASS_VAR : [],
                           ScintillaConstants.SCE_RB_INSTANCE_VAR : [],
                           ScintillaConstants.SCE_RB_IDENTIFIER : [self.tree.local_vars]} #locals
        

    def class_has_method(self, curr_node, the_text):
        try:
            class_node = self.containers[ScintillaConstants.SCE_RB_CLASSNAME][-1]
            for c in class_node.children:
                if isinstance(c, MethodNode) and c.name == the_text:
                    return True
        except:
            pass
        return False


    def curr_block_start_indentation(self):
        try:
            ind = self.block_stack[-1].indentation
        except:
            ind = 0
        return ind

    # This routine compares the current line indentation
    # with the one at the start of the block
    #
    # Neg value: we've moved to far back
    # 0: same ind
    # Pos value: curr indent'n is greater than where we started
    
    def compare_curr_ind(self):
        start_ind = self.curr_block_start_indentation()
        curr_indent = self.tokenizer.get_curr_indentation() # 
        return curr_indent - start_ind
    
    def dump(self):
        self.tree.dump()
        
    def parse(self):
        try:
            self.parse_aux(self.tree)
        except:
            # Swallow the rest of the tokens so the lexer thread finishes.
            # Otherwise Python will hang on an exception
            while self.tokenizer.get_next_token().style != ruby_lexer.EOF_STYLE:
                pass
            raise # and reraise
        return self.tree
        
    def get_parsing_objects(self, kwd):
        return {
            "module": [ModuleNode, self.parse_aux],
            "class" : [ClassNode, self.parse_aux],
            "def" : [MethodNode, self.parse_method]
        }.get(kwd, [None, None])

    def parse_open_parens(self):
        paren_count = 0
        while True:
            tok = self.tokenizer.get_next_token()
            if is_operator(tok, "("):
                paren_count += 1
            else:
                return (tok, paren_count)

    def parse_attr_stmt_capture_info(self, collection, attr_tok, tok, curr_node):
        base_name = tok.text[1:]
        update_collection(collection,
                          '@' + base_name, tok.start_line)
        if isinstance(curr_node, ClassNode):
            if attr_tok != 'attr_writer':
                new_node = MethodNode(base_name, tok.start_line)
                new_node.set_line_end_num(tok.start_line)
                curr_node.append_node(new_node)
            if attr_tok in ['attr_writer', 'attr_accessor']:
                new_node = MethodNode(base_name + '=', tok.start_line)
                new_node.set_line_end_num(tok.start_line)
                new_node.add_arg(base_name, "")
                curr_node.append_node(new_node)

    def parse_attr_stmts(self, attr_tok, curr_node):
        """
        attr_tok is one of ['attr', 'attr_reader', 'attr_writer', 'attr_accessor']
        if curr_node is a ClassNode then set up a bunch of methods as well
        """
        try:
            collection = self.containers[ScintillaConstants.SCE_RB_INSTANCE_VAR][-1]
            if collection is None:
                return
        except:
            return
        (tok, paren_count) = self.parse_open_parens()
        if tok.style == ScintillaConstants.SCE_RB_SYMBOL:
            self.parse_attr_stmt_capture_info(collection, attr_tok, tok, curr_node)                
        # Allow for more, but special case attr's
        while True:
            tok = self.tokenizer.get_next_token()
            if is_operator(tok, ","):
                tok = self.tokenizer.get_next_token()
                # Intern unless attr
                if tok.style == ScintillaConstants.SCE_RB_SYMBOL and attr_tok != 'attr':
                    self.parse_attr_stmt_capture_info(collection, attr_tok, tok, curr_node)
            else:
                self.tokenizer.put_back(tok)
                break
        # Might as well consume the closing parens
        while paren_count > 0:
            tok = self.tokenizer.get_next_token()
            if is_operator(tok, ")"):
                paren_count -= 1
            else:
                self.tokenizer.put_back(tok)
                break;

    def parse_assignment(self, collectionA, tok_text, start_line):
        if len(collectionA) == 0 or collectionA[-1] is None:
            return
        
        tok = self.tokenizer.get_next_token()
        # We don't need to Watch out for calls like abc=(1)
        # because that's always an assignment to a local.
        # self.abc=(1) is different
        
        if tok.style == ScintillaConstants.SCE_RB_OPERATOR:
            if tok.text == "=":
                self._finishVarAssignment(collectionA, tok_text, start_line)
                # Look to see if we have an instance of a class
                # This is 
                update_collection(collectionA[-1], tok_text, start_line)
                return
            elif tok.text in ["::", "."]:
                # Don't store fully-qualified names, but do consume them
                # Keep name for debugging purposes.
                rest_of_name = self.get_fully_qualified_name()
                return
        self.tokenizer.put_back(tok)

    def _is_number(self, tok):
        return tok.style == ScintillaConstants.SCE_RB_NUMBER
    
    def _is_string(self, tok):
        return tok.style in [ScintillaConstants.SCE_RB_STRING,
                             ScintillaConstants.SCE_RB_CHARACTER,
                             ScintillaConstants.SCE_RB_HERE_Q,
                             ScintillaConstants.SCE_RB_HERE_QQ,
                             ScintillaConstants.SCE_RB_STRING_Q,
                             ScintillaConstants.SCE_RB_STRING_QQ,
                             ScintillaConstants.SCE_RB_STRING_QX
                             ]
    
    def _get_builtin_type(self, tok):
        if self._is_number(tok):
            numval = tok.text
            if numval.find(".") >= 0:
                return "Float"
            else:
                return "Fixnum" 
        elif self._is_string(tok):
            return "String"
        elif tok.style == ScintillaConstants.SCE_RB_STRING_QR:
            return "Regexp"
        elif tok.style == ScintillaConstants.SCE_RB_STRING_QW:
            return "Array"
        return None
    
    def _at_end_expression(self, tok):
        if tok.style != ScintillaConstants.SCE_RB_OPERATOR:
            return True
        return tok.text in [",", "}"]
    
    def skip_to_op(self, opText, nestedLevel=1):
        skipped_toks = []
        while 1:
            tok = self.tokenizer.get_next_token()
            skipped_toks.append(tok)
            if tok.style == ruby_lexer.EOF_STYLE:
                return skipped_toks
            elif tok.style == ScintillaConstants.SCE_RB_OPERATOR:
                if tok.text in ["(", "{", "["]:
                    nestedLevel += 1
                elif tok.text in ["]", "}", ")"]:
                    nestedLevel -= 1
                    if nestedLevel <= 0:
                        return skipped_toks
                # don't care about other ops
            # don't care about other token types
    
    def _finishVarAssignment(self, collectionA, var_name, start_line):
        # See doc for finishVarAssignment in src/perlcile/ParseModule.pm
        # for comments.
        # 1: a = b.c.d.e -- don't bother
        # 2.1: a = <number> <no op>
        # 2.2: a = <string> <no op>
        # 3. a = <classname>.New [( ... )]<no op>

        tok = self.tokenizer.get_next_token()
        builtin_type = self._get_builtin_type(tok)
        if builtin_type:
            type1, locn = tok.style, tok.start_line
            tok2 = self.tokenizer.get_next_token()
            if self._at_end_expression(tok2):
                if self._is_number(tok):
                    type1 = "Number"
                else:
                    type1 = "String"
                update_collection(collectionA[-1], var_name, start_line, builtin_type)
                toks = [tok2]
            else:
                update_collection(collectionA[-1], var_name, start_line)
                toks = [tok2, tok]
        elif tok.style == ScintillaConstants.SCE_RB_IDENTIFIER and tok.text[0].isupper():
            class_name = tok.text
            tok2 = self.tokenizer.get_next_token()
            if self._at_end_expression(tok2):
                update_collection(collectionA[-1], var_name, start_line, "Class")
                toks = [tok2]
            elif is_operator(tok2, "."):
                tok3 = self.tokenizer.get_next_token()
                toks = [tok3, tok2]
                if tok3.style == ScintillaConstants.SCE_RB_IDENTIFIER and tok3.text == "new":
                    tok4 = self.tokenizer.get_next_token()
                    toks.insert(0, tok4)
                    if is_operator(tok4, "("):
                        skipped_toks = self.skip_to_op(")")
                        skipped_toks.reverse()
                        toks = skipped_toks + toks
                        final_tok = self.tokenizer.get_next_token()
                        toks.insert(0, final_tok)
                    else:
                        final_tok = tok4
                    if not self._at_end_expression(final_tok):
                        class_name = None
                    update_collection(collectionA[-1], var_name, start_line, class_name)
                else:
                    update_collection(collectionA[-1], var_name, start_line)
            else:
                update_collection(collectionA[-1], var_name, start_line)
                toks = [tok2]
        elif is_operator(tok, "["):
            toks = self._finish_list(collectionA, var_name, start_line, tok, "]", "Array")
        elif is_operator(tok, "{"):
            toks = self._finish_list(collectionA, var_name, start_line, tok, "}", "Hash")
        else:
            update_collection(collectionA[-1], var_name, start_line)
            toks = [tok]
        for t in toks:
            self.tokenizer.put_back(t)
            
    def _finish_list(self, collectionA, var_name, start_line, orig_tok, end_op, class_name):
        skipped_toks = self.skip_to_op(end_op)
        final_tok = self.tokenizer.get_next_token()
        skipped_toks.append(final_tok)
        skipped_toks.reverse()
        skipped_toks.append(orig_tok)
        class_name_final = self._at_end_expression(final_tok) and class_name or None
        update_collection(collectionA[-1], var_name, start_line, class_name_final)
        return skipped_toks

    def _set_basename_method(self, node_class, curr_node, nm_token):
        if node_class != MethodNode: return (nm_token, False)
        method_name = nm_token[0]
        idx1 = method_name.rfind('::')
        if idx1 >= 0: idx1_len = 2
        idx2 = method_name.rfind('.')
        if idx2 >= 0: idx2_len = 1
        if idx1 < 0:
            if idx2 < 0:
                return (nm_token, False)
            idx = idx2
            idx_len = 1
        elif idx2 < idx1:
            # Includes no idx2
            idx = idx1
            idx_len = 2
        else:
            idx = idx2
            idx_len = 1
        first_part = method_name[0:idx]
        if first_part != getattr(curr_node, 'name', '') and first_part != 'self':
            # We're doing something like defining inside a mix-in
            return (nm_token, False)            
            
        basename = method_name[idx + idx_len :]
        if len(nm_token) != 2:
            raise("Expectations dashed!")
        
        return ((basename, nm_token[1]), True)

    def parse_aux(self, curr_node):
        while 1:
            tok = self.tokenizer.get_next_token()
            if tok.style == ruby_lexer.EOF_STYLE:
                break;
            # style, text, start_column, start_line, end_column, end_line = tok
            style, text = tok.style, tok.text
            if style == ScintillaConstants.SCE_RB_WORD:
                if text in ["module", "class", "def"]:
                    # Check to see if we missed some nodes,
                    # using heuristic that the end statements line up
                    # with the start of each block, for definitions

                    if len(self.block_stack) > 0 and self.compare_curr_ind() <= 0:
                        # We missed one or more end's, so put the token
                        # back and try in an outer instance
                        self.tokenizer.put_back(tok)
                        return

                    curr_indent = self.tokenizer.get_curr_indentation() # 
                    node_class, node_parser = self.get_parsing_objects(text)
                    if node_class is None:
                        sys.stderr.write("Couldn't get parsing objects for type %s\n" % text)
                        break
                    elif node_class is MethodNode:
                        self.tokenizer.start_sig()

                    # Get the comments before further parsing.
                    comment_lines = remove_hashes(self.tokenizer.curr_comment())
                    nm_token = self.get_fully_qualified_name()
                    if not nm_token[0]:
                        return
                    elif nm_token[0] == "<<" and node_class == ClassNode:
                        # Singleton classes
                        nm_token = self.get_fully_qualified_name()
                        if not nm_token[0]:
                            return

                    (nm_token, is_class_method) = self._set_basename_method(node_class, curr_node, nm_token)
                    new_node = node_class(nm_token[0], tok.start_line)
                    if is_class_method:
                        new_node.is_classmethod = True
                    elif node_class == MethodNode and (isinstance(curr_node, ClassNode) or isinstance(curr_node, ModuleNode)):
                        # Add info on the 'self' variable
                        update_collection(new_node.local_vars, 'self', tok.start_line, curr_node.name)
                    new_node.doc_lines = comment_lines
                    new_node.indentation = curr_indent
                    self.block_stack.append(new_node)
                    curr_node.append_node(new_node)

                    # Push new containers on the symbol table
                    self.containers[ScintillaConstants.SCE_RB_IDENTIFIER].append(new_node.local_vars)
                    if node_class in [ClassNode, ModuleNode]:
                        self.containers[ScintillaConstants.SCE_RB_CLASSNAME].append(new_node)
                        self.containers[ScintillaConstants.SCE_RB_CLASS_VAR].append(new_node.class_vars)
                        self.containers[ScintillaConstants.SCE_RB_INSTANCE_VAR].append(new_node.instance_vars)

                    # Watch out for class inheritence
                    if node_class == ClassNode:
                        self.parse_classref(new_node)
                    node_parser(new_node)  # Has self bound to it
                    self.block_stack.pop()
                    self.containers[ScintillaConstants.SCE_RB_IDENTIFIER].pop()
                    if node_class in [ClassNode, ModuleNode]:
                        self.containers[ScintillaConstants.SCE_RB_CLASS_VAR].pop()
                        self.containers[ScintillaConstants.SCE_RB_INSTANCE_VAR].pop()
                        self.containers[ScintillaConstants.SCE_RB_CLASSNAME].pop()

                    # Clear any comment that's hanging around
                    self.tokenizer.clear_comments()

                elif text == "end":
                    if len(self.block_stack) > 0:
                        end_position = self.compare_curr_ind()
                        if end_position < 0:
                            # We've gone too far, so put it back and try later
                            curr_node.set_line_end_num(tok.start_line)
                            self.tokenizer.put_back(tok)
                            return
                        elif end_position == 0:
                            curr_node.set_line_end_num(tok.start_line)
                            # the caller will pop the stack
                            return
                        
            elif style == ScintillaConstants.SCE_RB_IDENTIFIER:
                if text == "require" or text == "load":
                    tok = self.tokenizer.get_next_token()
                    if (ruby_lexer.is_string_token(tok)):
                        curr_node.imports.append(Name_LineNum(tok.text.strip('\'\"'), tok.start_line))
                elif text == "include":
                    nm_token = self.get_fully_qualified_name()
                    if nm_token[0]:
                        curr_node.includes.append(Name_LineNum(nm_token[0], nm_token[1]))
                elif tok.start_column == self.tokenizer.get_curr_indentation():
                    # Look at these things only at start of line
                    if text in ['attr', 'attr_reader', 'attr_writer', 'attr_accessor']:
                        self.parse_attr_stmts(text, curr_node)
                    elif not self.class_has_method(curr_node, tok.text + "="):
                        # Make sure it isn't an assignment function
                        self.parse_assignment(self.containers[ScintillaConstants.SCE_RB_IDENTIFIER], text, tok.start_line)

            elif style == ScintillaConstants.SCE_RB_OPERATOR:
                if text == "{":
                    new_node = BlockNode("{", tok.start_line);
                    new_node.indentation = self.tokenizer.get_curr_indentation()
                    #XXX Transfer any children a block defines to the
                    # parent of the block.  Uncommon formulation
                    self.block_stack.append(new_node)
                    self.parse_aux(new_node)
                    self.block_stack.pop()
                elif text == "}":
                    if len(self.block_stack) > 0:
                        end_position = self.compare_curr_ind()
                        if end_position < 0:
                            # We've gone too far, so put it back and try later
                            self.tokenizer.put_back(tok)
                            return
                        elif end_position == 0:
                            # the caller will pop the stack
                            return
            # Check for assignment to some kind of variable
            elif tok.style in [ScintillaConstants.SCE_RB_GLOBAL,
                               ScintillaConstants.SCE_RB_INSTANCE_VAR,
                               ScintillaConstants.SCE_RB_CLASS_VAR]:
                self.parse_assignment(self.containers[tok.style], tok.text, tok.start_line)
                    
            # end if WORD block
            #XXX: process variables as well
        # end while
        return
    # end parse_aux()
    
    """
    Simplified Ruby function-defn grammar
    
method_def ::= kDEF fname f_arglist bodystmt kEND
fname ::= ID
f_arglist ::= "(" f_args opt_nl ")" | f_args term
term ::= "\n" | ";"
f_args ::= f_arg(s /,/)
	 | <nothing>
f_arg ::= f_norm_arg | f_opt | f_rest_arg | f_block_arg
f_norm_arg ::= id
f_opt ::= id '=' arg_value
f_rest_arg ::= "*" id? 
f_block_arg ::= "&" id 

arg_value ::= simple_expn
simple_expn ::= (simple_term | nested_expn)+
simple_term ::= <anything but a ";", "\n", ")", "," or a block-starter or closer
 - the exceptions depend on whether we're in a parenthesized construct or not
simple_inner_expn ::= (simple_term | nested_expn)+
nested_expn ::= "(" simple_expn ")" | "{" simple_expn "}" | "[" simple_expn "]"
"""
        
    def parse_nested_expn(self, block_end_char):
        while 1:
            tok = self.tokenizer.get_next_token()
            if tok.style == ruby_lexer.EOF_STYLE:
                return
            elif tok.style == ScintillaConstants.SCE_RB_OPERATOR:
                # Don't look at commas or semi-colons here, as we might be
                # processing nested comma-sep things or even blocks
                if len(tok.text) == 1:
                    if tok.text == block_end_char:
                        return -1
                    elif "[{(".find(tok.text) >= 0:
                        self.parse_nested_expn(self.bracket_matchers[tok.text])
            elif is_keyword(tok, "end"):
                if len(self.block_stack) == 0:
                    return
                elif self.compare_curr_ind() <= 0:
                    # Did we go too far?
                    self.tokenizer.put_back(tok)
                    return
    # end parse_nested_expn
    
    # This might not always work
    def parse_simple_expn(self, has_paren):
        while 1:
            tok = self.tokenizer.get_next_token()
            if tok.style == ruby_lexer.EOF_STYLE:
                return
            elif tok.style == ScintillaConstants.SCE_RB_OPERATOR:
                if tok.text == ",":
                    self.tokenizer.put_back(tok)
                    return
                elif has_paren and tok.text == ')':
                    self.tokenizer.put_back(tok)
                    return
                elif not has_paren and tok.text == ';':
                    self.tokenizer.put_back(tok)
                    return
                elif len(tok.text) == 1 and "[{(".find(tok.text) >= 0:
                    self.parse_nested_expn(self.bracket_matchers[tok.text])
            elif is_keyword(tok, "end"):
                if len(self.block_stack) == 0:
                    return
                elif self.compare_curr_ind() <= 0:
                    # Did we go too far?
                    self.tokenizer.put_back(tok)
                    return
    # end parse_simple_expn

    def parse_method_args(self, curr_node):
        tok = self.tokenizer.get_next_token()
        if tok.style == ruby_lexer.EOF_STYLE:
            return
        elif is_operator(tok, ";"):
            self.tokenizer.put_back(tok)
            return  # No args
        elif is_operator(tok, "("):
            has_paren = True
        elif tok.start_line > curr_node.line_num:
            # Assume a zero-paren function
            self.tokenizer.put_back(tok)
            return  # No args
        
        else:
            self.tokenizer.put_back(tok)  # simplifies the while loop
            has_paren = False
        # Further simplification: look for the "," at the start of the loop
        comma_token = ruby_lexer.Token(style=ScintillaConstants.SCE_RB_OPERATOR, text=",")
        comma_token.generated = 1
        self.tokenizer.put_back(comma_token)
        while 1:
            tok = self.tokenizer.get_next_token()
            # First check for white-space, set indent,
            if tok.style == ruby_lexer.EOF_STYLE:
                return
            elif has_paren and is_operator(tok, ")"):
                return
            elif is_keyword(tok, 'end') and self.compare_curr_ind() <= 0:
                # Did we go too far?
                self.tokenizer.put_back(tok)
                return

            # Note -- silvercity swallows \-line-continuations,
            # so we don't need to handle them

            if is_operator(tok, ","):
                tok = self.tokenizer.get_next_token()
            else:
                self.tokenizer.put_back(tok)
                return
            
            # then check for arg-list termination,
            # and then process the arg...
            extra_info = ""
            if (tok.style == ScintillaConstants.SCE_RB_OPERATOR and
                len(tok.text) == 1 and "*&".find(tok.text) >= 0):
                extra_info = tok.text
                tok = self.tokenizer.get_next_token()
            if is_identifier(tok, 1):
                curr_node.add_arg(tok.text, extra_info)
            else:
                # give up -- expected to see an arg, didn't
                return
            # check for an '=' sign
            tok = self.tokenizer.get_next_token()
            if is_operator(tok, '='):
                self.parse_simple_expn(has_paren)
            else:
                self.tokenizer.put_back(tok)
        # end while
    # end parse_method_args
    
    def parse_method(self, curr_node):
        # Get the args before dispatching back to the main routine

        self.parse_method_args(curr_node)
        self.tokenizer.stop_sig()

        #XXX Remove newlines and \-prefixed newlines, with white-space
        curr_node.signature = self.tokenizer.get_sig().strip()
        
        # Now look for the 'end' on the same line as a special case
        # The start line is contained in curr_node.line_num        
        while 1:
            tok = self.tokenizer.get_next_token()
            if tok.style == ruby_lexer.EOF_STYLE:
                return;
            # style, text, start_column, start_line, end_column, end_line = tok
            if tok.end_line > self.block_stack[-1].line_num:
                self.tokenizer.put_back(tok)
                break
            #*** Heuristic: if we find an 'end' followed by a newline, end the sub here
            elif is_keyword(tok, "end"):
                # Assume that's the end
                # Let the caller pop the block_stack
                return
        # end while
        self.parse_aux(curr_node)
    # end parse_method
    
    def parse_classref(self, node):
        toks = []
        tok = self.tokenizer.get_next_token(1)
        fqname = classref_type = None
        if is_operator(tok, "<"):
            fqname, line_start = self.get_fully_qualified_name()
            if fqname == "DelegateClass":
                tok = self.tokenizer.get_next_token(1)
                if is_operator(tok, "("):
                    # No putback once we enter this point
                    inner_name, junk = self.get_fully_qualified_name()
                    tok = self.tokenizer.get_next_token()
                    if is_operator(tok, ")"):
                        fqname = "%s(%s)" % (fqname, inner_name)
                        classref_type = inner_name
                else:
                    toks.insert(0, tok)
        else:
            toks.insert(0, tok)
        if fqname is not None:
            node.add_classrefs(fqname, line_start, classref_type)
        else:
            for t in toks:
                self.tokenizer.put_back(t)

    def get_fully_qualified_name(self):
        tok = self.tokenizer.get_next_token()
        if tok.style == ruby_lexer.EOF_STYLE:
            return (None, None)
        name_start = tok.text
        line_start = tok.start_line
        # Watch out if it starts with a "::"
        if name_start == "::":
            tok = self.tokenizer.get_next_token()
            if tok.style != ScintillaConstants.SCE_RB_IDENTIFIER:
                self.tokenizer.put_back(tok)
                return (name_start, line_start)
            name_start += tok.text
                
        while 1:
            # Collect operator-type methods
            if tok.style == ScintillaConstants.SCE_RB_OPERATOR:
                while 1:
                    tok = self.tokenizer.get_next_token()
                    if (tok.style == ScintillaConstants.SCE_RB_OPERATOR
                        and tok.text != "("):
                        name_start += tok.text
                    else:
                        self.tokenizer.put_back(tok)
                        break
                # And it will be the end of the line
                return (name_start, line_start)
            tok = self.tokenizer.get_next_token()
            if tok.style != ScintillaConstants.SCE_RB_OPERATOR:
                self.tokenizer.put_back(tok)
                break
            if not tok.text in ["::", "."]:
                self.tokenizer.put_back(tok)
                break
            tok2 = self.tokenizer.get_next_token()
            if tok2.style != ScintillaConstants.SCE_RB_IDENTIFIER:
                self.tokenizer.put_back(tok2)
                self.tokenizer.put_back(tok)
                break
            name_start += tok.text + tok2.text
        return (name_start, line_start)  
            
        

# end class Parser
        
                    
if __name__ == "__main__":
    if len(sys.argv) == 1:
        sample_code = ruby_lexer.provide_sample_code();
        fs = None
    elif sys.argv[1] == "-":
        fs = sys.stdin
        closefs = False
    else:
        fs = open(sys.argv[1], "r")
        closefs = True
    if fs is not None:
        sample_code = ruby_lexer.read_and_detab(fs, closefs)
        # fs comes back closed
    tokenizer = ruby_lexer.Lexer(sample_code)
    parser = Parser(tokenizer)
    tree = parser.parse()
    print "Analyze the parse tree"
    tree.dump()
