#!python
# Copyright (c) 2004 Sophos Inc.
# See the file LICENSE.txt for licensing information.

"""One-writer-multiple-readers lock class."""
# Dev Notes:
# - Should implement RWRLock (a re-entrant version). That could get
#   complicated though: e.g. how to handle thread X requesting a write lock
#   when it already holds a read lock.

import sys
import threading


DEBUG = False
if DEBUG:
    def log(s):
        sys.stdout.write("[%s] %s\n" % (threading.currentThread().getName(), s))
        sys.stdout.flush()


class RWLock:
    """Create a one-writer-multiple-readers lock.
    
    This lock implements one-writer-multiple-readers thread locking
    semantics. I.e. the only possible thread access states for the guarded
    resource(s) will be:
    - one writer and no readers
    - no writers and zero or more readers
    
    If a request is made for a write-lock when readers are active, subsequent
    read-lock requests will block until all current readers finish and the
    writer starts and finishes. In other words, writers cannot be starved
    by continual overlapped read-lock requests.
    """
    # Dev Notes:
    # - could pool events

    def __init__(self):
        self.mutex = threading.Lock()
        self.mode = None # None: no acquirees, 'r': reading, 'w': writing
        self.rcount = 0  # number of readers if self.mode=='r'
        self.blocked = [] # FIFO of blocked acquire() calls
        self.held = {} # mapping of thread name to held lock type (used for error trapping)

    def acquire(self, purpose):
        """acquire a lock for reading or writing

            "purpose" must be one of 'r' or 'w'.
        """
        assert purpose in ('r', 'w')
        event = None
        self.mutex.acquire()
        if DEBUG: log("acquiring %s-lock: acquiring..." % purpose)
        try:
            tid = threading.currentThread().getName()
            if tid in self.held:
                from warnings import warn
                msg = "Thread %r tried to acquire a '%s'-lock when "\
                      "already holds a '%s'-lock."\
                      % (tid, purpose, self.held[tid])
                warn(msg, stacklevel=2)

            if purpose == 'r':
                if self.mode is None:   # No one reading or writing: go ahead.
                    self.mode = 'r'
                    self.rcount = 1
                    self.held[tid] = 'r'
                elif self.mode == 'r':
                    if not self.blocked:# Reading and no blocked writers: go ahead.
                        self.rcount += 1
                        self.held[tid] = 'r'
                    else:               # Reading, but a writer is blocked: block.
                        event = threading.Event()
                        self.blocked.append( (purpose, event, tid) )
                elif self.mode == 'w':  # Writing: block.
                    event = threading.Event()
                    self.blocked.append( (purpose, event, tid) )
            elif purpose == 'w':
                if self.mode is None:   # No one reading or writing: go ahead.
                    self.mode = 'w'
                    self.held[tid] = 'w'
                else:                   # Someone is reading or writing: block.
                    event = threading.Event()
                    self.blocked.append( (purpose, event, tid) )
        finally:
            self.mutex.release()
        if event is not None:
            if DEBUG: log("acquiring %s-lock: block" % purpose)
            event.wait()
            if DEBUG: log("acquiring %s-lock: ...acquired (after block)" % purpose)
        else:
            if DEBUG: log("acquiring %s-lock: ...acquired" % purpose)

    def release(self, purpose):
        """release an acquired lock for reading or writing

            "purpose" must match the value ('r' or 'w') for the associated
                acquire() call.
        """
        #XXX Could track the kind of lock acquire per thread and not
        #    require "purpose" on release.
        assert purpose in ('r', 'w')
        self.mutex.acquire()
        if DEBUG: log("releasing %s-lock: releasing..." % purpose)
        try:
            tid = threading.currentThread().getName()
            assert purpose == self.mode
            if purpose == 'r':
                self.rcount -= 1
                del self.held[tid]
                if self.rcount == 0:
                    self.mode = None
                    if self.blocked:
                        # writer blocked at head of list: unblock it
                        bpurpose, bevent, btid = self.blocked.pop(0)
                        if DEBUG: log("releasing %s-lock: unblocking writer %r" % (purpose, btid))
                        assert bpurpose == 'w'
                        self.mode = bpurpose
                        self.held[btid] = bpurpose
                        bevent.set()
            elif purpose == 'w':
                self.mode = None
                del self.held[tid]
                if self.blocked:
                    # free writer or continguous set of readers at head of list
                    if self.blocked[0][0] == 'w':
                        bpurpose, bevent, btid = self.blocked.pop(0)
                        if DEBUG: log("releasing %s-lock: unblocking writer %r" % (purpose, btid))
                        self.held[btid] = self.mode = bpurpose
                        bevent.set()
                    else:
                        while self.blocked and self.blocked[0][0] == 'r':
                            bpurpose, bevent, btid = self.blocked.pop(0)
                            if DEBUG: log("releasing %s-lock: unblocking reader %r" % (purpose, btid))
                            self.rcount += 1
                            self.held[btid] = self.mode = bpurpose
                            bevent.set()
        finally:
            self.mutex.release()
        if DEBUG: log("releasing %s-lock: ...released" % purpose)
